/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture;

import java.util.ArrayList;
import java.util.List;
import org.json.JSONArray;
import org.json.JSONObject;
import at.jku.iic.microfluidics.droplet.basic.architecture.discrete.DiscreteBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.discrete.DiscreteEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructureBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructureEdge;

/**
 * A container holding different representations of one and the same biochip.
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public class BiochipContainer {

    /**
     * The biochip in the structural representation
     */
    private StructureBiochip structureBiochip;

    /**
     * The experiments with edges in the structural representation
     */
    private List<Path<StructureEdge>> structureExperiments;

    /**
     * The biochip in the physical representation
     */
    private PhysicalBiochip physicalBiochip;

    /**
     * The experiments with edges in the physical representation
     */
    private List<Path<PhysicalEdge>> physicalExperiments;

    /**
     * The biochip in the discrete representation
     */
    private DiscreteBiochip discreteBiochip;

    /**
     * The experiments with edges in the discrete representation
     */
    private List<Path<DiscreteEdge>> discreteExperiments;

    
    public StructureBiochip getStructureBiochip() {
    
        return structureBiochip;
    }

    
    public List<Path<StructureEdge>> getStructureExperiments() {
    
        return structureExperiments;
    }

    
    public PhysicalBiochip getPhysicalBiochip() {
    
        return physicalBiochip;
    }

    
    public List<Path<PhysicalEdge>> getPhysicalExperiments() {
    
        return physicalExperiments;
    }

    
    public DiscreteBiochip getDiscreteBiochip() {
    
        return discreteBiochip;
    }

    
    public List<Path<DiscreteEdge>> getDiscreteExperiments() {
    
        return discreteExperiments;
    }


    
    public void setStructureBiochip(StructureBiochip structureBiochip, List<Path<StructureEdge>> structureExperiments) {
    
        this.structureBiochip = structureBiochip;
        this.structureExperiments = structureExperiments;
    }


    
    public void setPhysicalBiochip(PhysicalBiochip physicalBiochip, List<Path<PhysicalEdge>> physicalExperiments) {
    
        this.physicalBiochip = physicalBiochip;
        this.physicalExperiments = physicalExperiments;
    }


    
    public void setDiscreteBiochip(DiscreteBiochip discreteBiochip, List<Path<DiscreteEdge>> discreteExperiments) {
    
        this.discreteBiochip = discreteBiochip;
        this.discreteExperiments = discreteExperiments;
    }

    public JSONObject toJSON(){
        JSONObject json = new JSONObject();

        //structureBiochip
        json.put("structureBiochip", structureBiochip.toJSON());

        //structureExperiments
        JSONArray jsonStructureExperiments = new JSONArray();
        json.put("structureExperiments", jsonStructureExperiments);
        for(Path<StructureEdge> path : structureExperiments){
            jsonStructureExperiments.put(path.toJSON());
        }

        //physicalBiochip
        json.put("physicalBiochip", physicalBiochip.toJSON());

        //physicalExperiments
        JSONArray jsonPhysicalExperiments = new JSONArray();
        json.put("physicalExperiments", jsonPhysicalExperiments);
        for(Path<PhysicalEdge> path : physicalExperiments){
            jsonPhysicalExperiments.put(path.toJSON());
        }

        //discreteBiochip
        json.put("discreteBiochip", discreteBiochip.toJSON());

        //discreteExperiments
        JSONArray jsonDiscreteExperiments = new JSONArray();
        json.put("discreteExperiments", jsonDiscreteExperiments);
        for(Path<DiscreteEdge> path : discreteExperiments){
            jsonDiscreteExperiments.put(path.toJSON());
        }

        return json;
    }

    public static BiochipContainer fromJSON(JSONObject json){
        BiochipContainer chip = new BiochipContainer();
        chip.structureExperiments = new ArrayList<>();
        chip.physicalExperiments = new ArrayList<>();
        chip.discreteExperiments = new ArrayList<>();
        // chip.structureExperiments = new ArrayList<Path<StructureEdge>>();
        // chip.physicalExperiments = new ArrayList<Path<PhysicalEdge>>();
        // chip.discreteExperiments = new ArrayList<Path<DiscreteEdge>>();

        //structure
        chip.structureBiochip = StructureBiochip.fromJSON(json.getJSONObject("structureBiochip"));

        JSONArray jsonStructureExperiments = json.getJSONArray("structureExperiments");
        for (int i = 0; i < jsonStructureExperiments.length(); i++) {
            JSONArray jsonStructureExperiment = jsonStructureExperiments.getJSONArray(i);
            List<StructureEdge> path = new ArrayList<>();
            for (int j = 0; j < jsonStructureExperiment.length(); j++) {
                JSONObject jsonEdge = jsonStructureExperiment.getJSONObject(j);

                //find corresponding edge in the structureBiochip
                for(StructureEdge edge: chip.structureBiochip.edges){
                    if(jsonEdge.getInt("srcId") == edge.src.id && jsonEdge.getInt("dstId") == edge.dst.id && jsonEdge.getString("name").equals(edge.name)) {
                        path.add(edge);
                        break;
                    }
                }
            }
            chip.structureExperiments.add(new Path<>(path));
        }

        //physical
        chip.physicalBiochip = PhysicalBiochip.fromJSON(json.getJSONObject("physicalBiochip"));

        JSONArray jsonPhysicalExperiments = json.getJSONArray("physicalExperiments");
        for (int i = 0; i < jsonPhysicalExperiments.length(); i++) {
            JSONArray jsonPhysicalExperiment = jsonPhysicalExperiments.getJSONArray(i);
            List<PhysicalEdge> path = new ArrayList<>();
            for (int j = 0; j < jsonPhysicalExperiment.length(); j++) {
                JSONObject jsonEdge = jsonPhysicalExperiment.getJSONObject(j);

                //find corresponding edge in the physicalBiochip
                for(PhysicalEdge edge: chip.physicalBiochip.edges){
                    if(jsonEdge.getInt("srcId") == edge.src.id && jsonEdge.getInt("dstId") == edge.dst.id && jsonEdge.getString("name").equals(edge.name)) {
                        path.add(edge);
                        break;
                    }
                }
            }
            chip.physicalExperiments.add(new Path<>(path));
        }

        //discrete
        chip.discreteBiochip = DiscreteBiochip.fromJSON(json.getJSONObject("discreteBiochip"));

        JSONArray jsonDiscreteExperiments = json.getJSONArray("discreteExperiments");
        for (int i = 0; i < jsonDiscreteExperiments.length(); i++) {
            JSONArray jsonDiscreteExperiment = jsonDiscreteExperiments.getJSONArray(i);
            List<DiscreteEdge> path = new ArrayList<>();
            for (int j = 0; j < jsonDiscreteExperiment.length(); j++) {
                JSONObject jsonEdge = jsonDiscreteExperiment.getJSONObject(j);

                //find corresponding edge in the discreteBiochip
                for(DiscreteEdge edge: chip.discreteBiochip.edges){
                    if(jsonEdge.getInt("srcId") == edge.src.id && jsonEdge.getInt("dstId") == edge.dst.id && jsonEdge.getString("name").equals(edge.name)) {
                        path.add(edge);
                        break;
                    }
                }
            }
            chip.discreteExperiments.add(new Path<>(path));
        }
        
        return chip;
    }
}
