/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture;

import java.util.List;
import java.util.Objects;

import at.jku.iic.microfluidics.droplet.basic.architecture.converter.GraphToStructureBiochipConverter;

/**
 * Represents a physical component in which the continuous phase flows.
 * 
 * @author Andreas
 *
 */
public abstract class Edge<K extends Edge<K>> {

    public final EndPoint<K> src;
    public final EndPoint<K> dst;

    public final String name;

    @SuppressWarnings("unchecked")
    public Edge(EndPoint<K> src, EndPoint<K> dst, String name) {
        super();
        src = Objects.requireNonNull(src);
        dst = Objects.requireNonNull(dst);

        this.src = src;
        this.dst = dst;

        // Safe cast
        this.src.addSuccessor((K) this);

        this.name = name;
    }

    @Override
    public String toString() {

        return name + '(' + src.toString() + ", " + dst.toString() + ')';
    }

    public boolean isBypassChannel() {

        performTypeCheck();
        return this instanceof BypassChannel;
    }

    public boolean isChannel() {

        performTypeCheck();
        return this instanceof Channel;
    }

    public boolean isForwardingChannel() {

        performTypeCheck();
        return this instanceof Channel && this.src.isLeeSorter;
    }

    public boolean isModule() {

        performTypeCheck();
        return this instanceof Module;
    }

    private void performTypeCheck() {

        if (!(this instanceof ComponentType)) throw new RuntimeException(
                "Class " + this.getClass().getName() + " does not implement interface " + ComponentType.class.getName());
    }

    public int getEndPointHashs() {

        return (int) (0.5 * (src.id + dst.id) * (src.id + dst.id + 1) + dst.id);
    }

    /**
     * Returns a list with all successors which can be entered by header droplets.
     */
    public abstract List<K> getHeaderSuccessorEdges();

    /**
     * Returns a list with all successors which can be entered by payload droplets.
     */
    public abstract List<K> getPayloadSuccessorEdges();

    protected boolean isFirstPartOfModule() {

        return this instanceof Module && dst.getSuccessors().size() == 1 && dst.getSuccessors().get(0).isModule()
                && name.endsWith(GraphToStructureBiochipConverter.MODULE_SUFFIX_1);
    }
}
