/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import org.json.JSONObject;

/**
 * A point in the NLoC where an <code>Edge</code> terminates or starts.
 * 
 * @author Andreas Grimmer
 *
 */
public class EndPoint<K extends Edge<K>> {

    public final int id;

    private final List<K> predecessors;
    private final List<K> successors;

    public final boolean isLeeSorter;
    
    /**
     * The length of the Lee-Sorting element. This element sorts droplets depending on their size.
     */
    private double leeSortingLength;

    /**
     * The width of the Lee-Sorting element. This element sorts droplets depending on their size.
     */
    private double leeSortingWidth;

    /**
     * Creates a new end point, which is not a sorting element.
     * 
     * @param id
     *            A unique ID.
     */
    public EndPoint(int id) {
        super();
        this.id = id;
        this.isLeeSorter = false;

        this.predecessors = new LinkedList<>();
        this.successors = new LinkedList<>();
    }
    
    /**
     * Creates a new end point.
     * 
     * @param id
     *            A unique ID.
     * @param isLeeSorter
     *            Specifies whether the point is a Lee-sorter. Important: The Lee sorter has to be located
     *            at the destination of the directed edge!
     */
    public EndPoint(int id, double leeSortingLength, double leeSortingWidth) {
        super();
        this.id = id;
        this.isLeeSorter = true;
        this.leeSortingLength = leeSortingLength;
        this.leeSortingWidth = leeSortingWidth;

        this.predecessors = new LinkedList<>();
        this.successors = new LinkedList<>();
    }
    
    public double getLeeSortingLength() {
        
        return leeSortingLength;
    }
    
    
    public double getLeeSortingWidth() {
        
        return leeSortingWidth;
    }

    public void addSuccessor(K e) {

        successors.add(e);
        e.dst.predecessors.add(e);
    }

    public List<K> getPredecessors() {

        if(isLeeSorter) {
            assert predecessors.size() <= 1;
        }
        
        return Collections.unmodifiableList(predecessors);
    }

    public List<K> getSuccessors() {

        return Collections.unmodifiableList(successors);
    }

    public void clearSuccessors() {

        successors.clear();
    }

    public void clearPredecessors() {

        predecessors.clear();
    }

    /**
     * Checks if this <code>EndPoint</code> is connected to two successor channels.
     * 
     * @return
     */
    public boolean isBifurcation() {

        boolean multipleOutgoingChannels = successors.stream().filter(e -> e.isChannel()).count() >= 2;

        if (multipleOutgoingChannels) {
            // Additionally perform a check, if this bifurcation is bypassed
            // FIXME: This check only works for two-way bifurcations

            K c1 = successors.get(0);
            K c2 = successors.get(1);

            assert c1.dst.successors.stream().filter(x -> x.isBypassChannel()).count()
                    + c2.dst.successors.stream().filter(x -> x.isBypassChannel()).count() == 1;

            Optional<K> firstC1 = c1.dst.successors.stream().filter(x -> x.isBypassChannel()).findFirst();
            Optional<K> firstC2 = c2.dst.successors.stream().filter(x -> x.isBypassChannel()).findFirst();
            // Check if there is a bypass channel between both successor channels
            if (!(firstC1.isPresent() && !firstC2.isPresent() && c2.dst.predecessors.stream().anyMatch(x -> x == firstC1.get())
                    || firstC2.isPresent() && !firstC1.isPresent() && c1.dst.predecessors.stream().anyMatch(x -> x == firstC2.get()))) {
                throw new RuntimeException("Bifurcation is not valid bypassed! Check this case manually. Successors="
                        + successors.stream().map(Object::toString).collect(Collectors.joining(", ")));
            }
        }
        return multipleOutgoingChannels;
    }

    /**
     * Checks if two or more channels merge.
     * 
     * @return
     */
    public boolean isMerge() {

        return predecessors.stream().filter(e -> e.isChannel()).count() > 1;
    }

    @Override
    public String toString() {

        return Integer.toString(id);
    }

    public JSONObject toJSON() {
        JSONObject json = new JSONObject();
        json.put("id", id);
        json.put("isLeeSorter", isLeeSorter);
        if(Double.isNaN(leeSortingLength)){
            json.put("leeSortingLength", 0);
        } else {
            json.put("leeSortingLength", leeSortingLength);
        }
        if(Double.isNaN(leeSortingWidth)){
            json.put("leeSortingWidth", 0);
        } else {
            json.put("leeSortingWidth", leeSortingWidth);
        }

        return json;
    }
}
