/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.stream.Collectors;
import org.json.JSONArray;
import org.json.JSONObject;

/**
 * A path through the architecture consists of an ordered list of edges. Therefore, this class simply
 * represents an container for a list of edges.<br>
 * A path through the architecture can represent an biological experiment.
 * 
 * @author Andreas
 *
 * @param <K>
 */
public class Path<K extends Edge<K>> implements Iterable<K> {

    private final List<K> edges;

    public Path(List<K> edges) {
        super();
        // Wrap in ArrayList in order to guarantee a fast
        // index-based access.
        this.edges = new ArrayList<>(edges);
    }

    public List<K> getEdges() {

        return Collections.unmodifiableList(edges);
    }

    /**
     * Checks if this path contains an edge. Note: If this method causes performance problems, the contains
     * query should be made on an IdentitySet.
     * 
     * @param edge
     * @return
     */
    public boolean containsEdge(K edge) {

        return edges.contains(edge);
    }

    @Override
    public Iterator<K> iterator() {

        return edges.iterator();
    }

    @Override
    public String toString() {

        return edges.stream().map(t -> t.name).collect(Collectors.joining(", "));
    }
    
    @Override
    public boolean equals(Object obj) {
    
        if (this == obj) return true;
        if (obj == null) return false;
        if (getClass() != obj.getClass()) return false;
        Path<?> other = (Path<?>) obj;
        return edges.equals(other.edges);
    }
    
    public <V extends Edge<V>> boolean comparePathsByEndPointIds(Path<V> p) {
            
        Iterator<K> it1 = iterator();
        Iterator<V> it2 = p.iterator();
        
        while(it1.hasNext() && it2.hasNext()) {
            K e1 = it1.next();
            V e2 = it2.next();
            if(e1.src.id != e2.src.id || e1.dst.id != e2.dst.id) return false;
        }       
        
        return !it1.hasNext() && !it2.hasNext();
    }

    /**
     * Utility method which determines for a given path the driving pump (i.e. the pump connected to the first
     * edge of the path).
     * 
     * @param biochip The biochip containing the path.
     * @param path The path for which the pump has to be determined.
     * @return
     */
    public <V extends Pump<K>> V getDrivingPump(Biochip<K, V> biochip) {

        EndPoint<K> input = edges.get(0).src;
        List<V> drivingPumps = biochip.getPumps().stream().filter(p -> p.src == input).collect(Collectors.toList());
        if(drivingPumps.size() != 1) {
            throw new RuntimeException("No or multiple driving pumps!");
        }
        return drivingPumps.get(0);
    }

    public JSONArray toJSON() {
        JSONArray jsonEdges = new JSONArray();
        for (K edge : getEdges()) {
            JSONObject jsonEdge = new JSONObject();
            jsonEdge.put("type", edge.getClass().getSimpleName());
            jsonEdge.put("name", edge.name);
            jsonEdge.put("srcId", edge.src.id);
            jsonEdge.put("dstId", edge.dst.id);

            jsonEdges.put(jsonEdge);
        }

        return jsonEdges;
    }
}
