/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.converter;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

import at.jku.iic.microfluidics.droplet.basic.architecture.Biochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.Edge;
import at.jku.iic.microfluidics.droplet.basic.architecture.Module;
import at.jku.iic.microfluidics.droplet.basic.architecture.Path;
import at.jku.iic.microfluidics.droplet.basic.architecture.Pump;
import at.jku.iic.microfluidics.droplet.basic.graph.utility.GraphUtility;

public class ExperimentToPathMapper<K extends Edge<K>, V extends Pump<K>> {

    private final Biochip<K, V> biochip;

    public ExperimentToPathMapper(Biochip<K, V> biochip) {

        this.biochip = biochip;
    }

    /**
     * Maps the experiments (given as sequences of characters) to a path.
     * 
     * @param experiments
     * @return
     */
    public List<Path<K>> mapExperiments(List<String> experiments) {

        Set<String> remainingExperiments = new HashSet<>(experiments);
        List<Path<K>> mappedExperiments = new LinkedList<>();

        for (Entry<V, List<Path<K>>> pathsOfPump : GraphUtility.findAllDropletPaths(biochip).entrySet()) {

            for (Path<K> path : pathsOfPump.getValue()) {
                // Only when the path does not contain a forwarding channel, it is a valid path for a payload droplet.
                if (!path.getEdges().stream().anyMatch(e -> e.isForwardingChannel())) {
                    
                    StringBuilder experiment = new StringBuilder();
                    for (K edge : path) {
                        if (edge.isModule() && ((Module) edge).isFirstPartOfModule()) {
                            experiment.append(((Module) edge).getOperation());
                        }
                    }
                    if (remainingExperiments.contains(experiment.toString())) {
                        remainingExperiments.remove(experiment.toString());
                        mappedExperiments.add(path);
                    }
                }
            }
        }

        if (!remainingExperiments.isEmpty()) {
            throw new RuntimeException("Some  experiments have no mapping ");
        }

        return mappedExperiments;
    }

    /**
     * Takes a list of paths and maps it to edges given as input parameter.
     * 
     * @param originalPaths
     * @param edges
     * @return
     */
    public static <K extends Edge<K>, T extends Edge<T>> List<Path<K>> convertPaths(List<Path<T>> originalPaths, List<K> edges) {

        Map<Integer, K> mapping = edges.stream().collect(Collectors.toMap(K::getEndPointHashs, Function.identity()));
        List<Path<K>> convertedPaths = new LinkedList<>();

        for (Path<T> p : originalPaths) {
            convertedPaths.add(convertPath(mapping, p));
        }

        return convertedPaths;
    }

    /**
     * Takes a path and maps it to edges given as input parameter. The mapping is based on the endpoint IDs.
     * 
     * @param originalPath
     * @param edges
     * @return
     */
    public static <K extends Edge<K>, T extends Edge<T>> Path<K> convertPath(Path<T> originalPath, List<K> edges) {

        Map<Integer, K> mapping = edges.stream().collect(Collectors.toMap(K::getEndPointHashs, Function.identity()));
        return convertPath(mapping, originalPath);
    }

    private static <K extends Edge<K>, T extends Edge<T>> Path<K> convertPath(Map<Integer, K> mapping, Path<T> originalPath) {

        List<K> path = new LinkedList<>();
        for (T t : originalPath) {
            path.add(mapping.get(t.getEndPointHashs()));
        }
        return new Path<K>(path);
    }

}
