/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.converter;

import java.util.Arrays;
import java.util.Deque;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang3.tuple.ImmutablePair;

import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructureBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructureBypassChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructureChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructureEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructureModule;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructurePump;

public class GraphToStructureBiochipConverter {

    public static final String MODULE_SUFFIX_2 = "_2";
    public static final String MODULE_SUFFIX_1 = "_1";
    private static final String BP_PREXFIX = "bp";
    private static final String CHANNEL_PREFIX = "c";

    /**
     * FIXME: Move this method in test package upon the fix is done.
     * 
     * @return
     */
    public static StructureBiochip unnecessaryBifurcationTestCase() {

        boolean[][] connectionMatrix = { { false, true, false, false, false }, { false, false, true, true, true },
                { false, false, false, true, true }, { false, false, false, false, true }, { true, false, false, false, false } };

        boolean[] isPump = new boolean[connectionMatrix.length];
        isPump[0] = true;
        boolean singleOutlet = true;

        List<ImmutablePair<String, Character>> names = Arrays.asList(new ImmutablePair<>("p", 'p'), new ImmutablePair<>("m", 'm'), new ImmutablePair<>("h", 'h'),
                new ImmutablePair<>("i", 'i'), new ImmutablePair<>("d", 'd'));

        return convert(connectionMatrix, isPump, singleOutlet, names);
    }

    /**
     * FIXME: Note that this method probably does not produce the simplest channel network and causes
     * unnecessary bifurcations: At some points in the network, the channels could be merged and so
     * bifurcations could be saved/reused. Example: see <code>unnecessaryBifurcationTestCase</code>. Potential
     * solution: Post analyze the architecture and remove paths which are not required by any experiment.
     * 
     * @param connectionMatrix
     *            Represents a dense representation of the connected modules. The index equals zero thereby
     *            represents the connections to the single pump.
     * @param moduleNames
     * @return
     */
    public static StructureBiochip convert(boolean[][] connectionMatrix, boolean[] isPump, boolean singleOutlet,
            List<ImmutablePair<String, Character>> moduleNames) {

        assert connectionMatrix.length == moduleNames.size();
        assert connectionMatrix.length == isPump.length;

        List<StructureEdge> edges = new LinkedList<>();

        // Important to start the IDs with 1 for using it in Matlab
        int[] endPointIdCounter = { 1 };
        int[] channelCounter = { 0 };
        int[] bypassChannelCounter = { 0 };

        List<StructurePump> pumps = new LinkedList<>();

        ImmutablePair<EndPoint<StructureEdge>, EndPoint<StructureEdge>>[] components = createComponents(connectionMatrix, isPump, singleOutlet,
                moduleNames, endPointIdCounter, channelCounter, edges, pumps);

        for (int i = 0; i < connectionMatrix.length; i++) {
            if (components[i] != null) {

                Set<ImmutablePair<Integer, Boolean>> destinations = getOutgoingDestinations(connectionMatrix, isPump, i);
                assert destinations.size() >= 1;

                if (destinations.size() == 1 || destinations.stream().map(t -> t.right).allMatch(t -> t)) {
                    // Single successor or all successors are connected to the pump

                    Iterator<EndPoint<StructureEdge>> distinctIt = destinations.stream().map(t -> components[t.left].left).distinct()
                            .iterator();
                    while (distinctIt.hasNext()) {
                        edges.add(new StructureChannel(components[i].right, distinctIt.next(), CHANNEL_PREFIX + channelCounter[0]++));
                    }

                } else if (destinations.stream().map(t -> t.right).anyMatch(t -> t)) {
                    throw new UnsupportedOperationException(
                            "The destinations of the successors are pumps and modules. This is unsupported!");

                } else {
                    assert destinations.size() > 1;

                    Iterator<ImmutablePair<Integer, Boolean>> destIt = destinations.iterator();

                    // Channel to bifurcation
                    StructureChannel cIn = new StructureChannel(components[i].right, new EndPoint<StructureEdge>(endPointIdCounter[0]++),
                            CHANNEL_PREFIX + channelCounter[0]++);
                    edges.add(cIn);

                    // Realize multi-way bifurcations as two-way bifurcations

                    // A queue containing the endpoints where bifurcation channels can be connected.
                    Deque<EndPoint<StructureEdge>> queue = new LinkedList<EndPoint<StructureEdge>>();
                    queue.addLast(cIn.dst);

                    // Do until there are enough connectors
                    for (int availableConnectors = 2; availableConnectors < destinations.size(); availableConnectors++) {

                        StructureChannel c = new StructureChannel(queue.peekFirst(), new EndPoint<StructureEdge>(endPointIdCounter[0]++),
                                CHANNEL_PREFIX + channelCounter[0]++);
                        edges.add(c);

                        // Add a further channel to the ends of the bifurcation-successors
                        StructureChannel cSucc = new StructureChannel(c.dst, new EndPoint<StructureEdge>(endPointIdCounter[0]++),
                                CHANNEL_PREFIX + channelCounter[0]++);
                        edges.add(cSucc);
                        queue.addLast(cSucc.dst);

                        // Pop connector in case this connector already contains two successors
                        if (queue.peekFirst().getSuccessors().size() == 2) {
                            EndPoint<StructureEdge> endPoint = queue.removeFirst();

                            // Add bypass channels
                            EndPoint<StructureEdge> dst1 = endPoint.getSuccessors().get(0).dst;
                            EndPoint<StructureEdge> dst2 = endPoint.getSuccessors().get(1).dst;
                            StructureBypassChannel bypassChannel = new StructureBypassChannel(dst1, dst2,
                                    BP_PREXFIX + bypassChannelCounter[0]++);
                            edges.add(bypassChannel);
                        }
                    }

                    // Connect successor modules to endpoints
                    while (!queue.isEmpty()) {
                        EndPoint<StructureEdge> endPoint = queue.removeFirst();
                        StructureEdge c1;

                        if (endPoint.getSuccessors().size() == 0) {
                            c1 = new StructureChannel(endPoint, new EndPoint<StructureEdge>(endPointIdCounter[0]++),
                                    CHANNEL_PREFIX + channelCounter[0]++);
                            edges.add(c1);
                            // Add channel connecting the module
                            edges.add(new StructureChannel(c1.dst, components[destIt.next().left].left,
                                    CHANNEL_PREFIX + channelCounter[0]++));
                        } else {
                            assert endPoint.getSuccessors().size() == 1;
                            c1 = endPoint.getSuccessors().get(0);
                        }
                        StructureChannel c2 = new StructureChannel(endPoint, new EndPoint<StructureEdge>(endPointIdCounter[0]++),
                                CHANNEL_PREFIX + channelCounter[0]++);
                        edges.add(c2);
                        // Add channel connecting the module
                        edges.add(new StructureChannel(c2.dst, components[destIt.next().left].left, CHANNEL_PREFIX + channelCounter[0]++));

                        // Add a bypass to the destinations of the two channels
                        StructureBypassChannel bypassChannel = new StructureBypassChannel(c1.dst, c2.dst,
                                BP_PREXFIX + bypassChannelCounter[0]++);
                        edges.add(bypassChannel);
                    }
                    assert !destIt.hasNext();
                }
            }
        }

        return new StructureBiochip(edges, pumps);
    }

    /**
     * Creates the used components (i.e. pumps and modules). In case there are multiple incoming edges to a
     * component (i.e. multiple channels merge), this method additionally creates a proper input channel.
     * 
     * @param isPump
     * @param singleOutlet
     * 
     * @param sol
     * @param nloc
     * @param graphResult
     * @param endPointIdCounter
     * @return Returns the source and destination endpoints of the component or channel-component-combination.
     */
    private static ImmutablePair<EndPoint<StructureEdge>, EndPoint<StructureEdge>>[] createComponents(boolean[][] connectionMatrix,
            boolean[] isPump, boolean singleOutlet, List<ImmutablePair<String, Character>> moduleNames, int[] endPointIdCounter,
            int[] channelCounter, List<StructureEdge> edges, List<StructurePump> pumps) {

        @SuppressWarnings("unchecked")
        ImmutablePair<EndPoint<StructureEdge>, EndPoint<StructureEdge>>[] components = (ImmutablePair<EndPoint<StructureEdge>, EndPoint<StructureEdge>>[]) new ImmutablePair[connectionMatrix.length];

        EndPoint<StructureEdge> singleOutletEP = null;

        // Create all modules
        for (int i = 0; i < connectionMatrix.length; i++) {

            // Check if the module is used
            if (hasModuleOutgoingEdge(connectionMatrix, i)) {

                EndPoint<StructureEdge> src;
                EndPoint<StructureEdge> dst;

                // The component is either the pump or a module
                if (isPump[i]) {

                    EndPoint<StructureEdge> pressureSrc = new EndPoint<StructureEdge>(endPointIdCounter[0]++);
                    EndPoint<StructureEdge> pressureDst;
                    if (singleOutlet) {
                        if (singleOutletEP == null) singleOutletEP = new EndPoint<StructureEdge>(endPointIdCounter[0]++);
                        pressureDst = singleOutletEP;
                    } else {
                        pressureDst = new EndPoint<StructureEdge>(endPointIdCounter[0]++);
                    }

                    ImmutablePair<String, Character> pumpId = moduleNames.get(i);
                    StructurePump pump = new StructurePump(pressureSrc, pressureDst, pumpId.left);
                    pumps.add(pump);

                    // Correct, i.e. the source is the destination of the pressure.
                    src = pump.sink;
                    dst = pump.src;

                } else {
                    // Note: A module consists of two edges. This is an requirement for the Matlab-Simulator
                    // (i.e. otherwise the graph would contain double edges; the module and the forwarding
                    // channel would represent a double edge)

                    ImmutablePair<String, Character> moduleId = moduleNames.get(i);
                    StructureModule m1 = new StructureModule(new EndPoint<StructureEdge>(endPointIdCounter[0]++, Double.NaN, Double.NaN),
                            new EndPoint<StructureEdge>(endPointIdCounter[0]++), moduleId.left + MODULE_SUFFIX_1, moduleId.right);
                    edges.add(m1);
                    StructureModule m2 = new StructureModule(m1.dst, new EndPoint<StructureEdge>(endPointIdCounter[0]++),
                            moduleId.left + MODULE_SUFFIX_2, moduleId.right);
                    edges.add(m2);

                    // Add forwarding channel
                    StructureChannel fc = new StructureChannel(m1.src, m2.dst, "fc_" + moduleId.left);
                    edges.add(fc);

                    src = m1.src;
                    dst = m2.dst;

                    if (hasModuleMultipleIncomingEdges(connectionMatrix, i)) {
                        StructureChannel inputChannel = new StructureChannel(new EndPoint<StructureEdge>(endPointIdCounter[0]++),
                                src, CHANNEL_PREFIX + channelCounter[0]++);
                        edges.add(inputChannel);
                        src = inputChannel.src;
                    }
                }

                components[i] = new ImmutablePair<>(src, dst);
            }
        }
        return components;
    }

    /**
     * Checks if a module has an outgoing edge. Then the module is used.
     * 
     * @param connectionMatrix
     * @param moduleIdx
     * @return
     */
    private static boolean hasModuleOutgoingEdge(boolean[][] connectionMatrix, int moduleIdx) {

        for (int j = 0; j < connectionMatrix[moduleIdx].length; j++) {
            if (connectionMatrix[moduleIdx][j]) return true;
        }
        return false;
    }

    /**
     * Returns all successor module indices (i.e. the destination module indices).
     * 
     * @param connectionMatrix
     * @param moduleIdx
     * @return
     */
    private static Set<ImmutablePair<Integer, Boolean>> getOutgoingDestinations(boolean[][] connectionMatrix, boolean[] isPump, int moduleIdx) {

        Set<ImmutablePair<Integer, Boolean>> destinations = new HashSet<>();
        for (int j = 0; j < connectionMatrix[moduleIdx].length; j++) {
            if (connectionMatrix[moduleIdx][j]) {
                destinations.add(new ImmutablePair<>(j, isPump[j]));
            }
        }
        return destinations;
    }

    /**
     * Checks if a module has multiple incoming edges. Then this module is connected to a switch.
     * 
     * @param connectionMatrix
     * @param moduleIdx
     * @return
     */
    private static boolean hasModuleMultipleIncomingEdges(boolean[][] connectionMatrix, int moduleIdx) {

        int cnt = 0;
        for (int i = 0; i < connectionMatrix.length; i++) {
            if (connectionMatrix[i][moduleIdx]) {
                cnt++;
                if (cnt > 1) return true;
            }
        }
        return false;
    }

    private GraphToStructureBiochipConverter() {

    }
}
