/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.converter;

import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.architecture.Path;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructureEdge;

public class NumberOfSubsequentBifsStrategy implements DefaultSuccessorSelectionStrategy {

    private final List<Path<StructureEdge>> experiments;

    public NumberOfSubsequentBifsStrategy(List<Path<StructureEdge>> experiments) {

        super();
        this.experiments = experiments;
    }

    @Override
    public StructureEdge getDefaultSuccessor(EndPoint<StructureEdge> bifEndPoint) {

        assert bifEndPoint.isBifurcation();
        long[] subsequentBifs = getSubsequentBifs(bifEndPoint);
        long max = Arrays.stream(subsequentBifs).max().getAsLong();
        for (int i = 0; i < subsequentBifs.length; i++) {
            if (subsequentBifs[i] == max) {
                return bifEndPoint.getSuccessors().get(i);
            }
        }
        return null;
    }

    @Override
    public StructureEdge getNonDefaultSuccessor(EndPoint<StructureEdge> bifEndPoint) {

        assert bifEndPoint.isBifurcation();
        long[] subsequentBifs = getSubsequentBifs(bifEndPoint);
        long min = Arrays.stream(subsequentBifs).min().getAsLong();
        for (int i = subsequentBifs.length - 1; i >= 0; i--) {
            if (subsequentBifs[i] == min) {
                return bifEndPoint.getSuccessors().get(i);
            }
        }
        return null;
    }

    private long[] getSubsequentBifs(EndPoint<StructureEdge> bifEndPoint) {

        long[] subseqentBifs = new long[bifEndPoint.getSuccessors().size()];
        int idx = 0;

        for (StructureEdge succ : bifEndPoint.getSuccessors()) {

            List<Path<StructureEdge>> experimentsUsingSucc = experiments.stream()
                    .filter(exp -> exp.getEdges().stream().anyMatch(e -> e == succ)).collect(Collectors.toList());

            int maxSuccessorBifs = 0;
            for (Path<StructureEdge> experimentUsingSucc : experimentsUsingSucc) {
                
                int counter = 0;
                boolean edgeHit = false;
                for (StructureEdge edge : experimentUsingSucc.getEdges()) {
                    if (edge == succ) {
                        edgeHit = true;
                    }
                    if(edgeHit && edge.dst.isBifurcation()) {
                        counter++;
                    }
                }
                maxSuccessorBifs = Integer.max(maxSuccessorBifs, counter);
            }
            subseqentBifs[idx] = maxSuccessorBifs;
            idx++;
        }

        return subseqentBifs;
    }
}
