/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.converter;

import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.architecture.Module;
import at.jku.iic.microfluidics.droplet.basic.architecture.discrete.DiscreteBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.discrete.DiscreteBypassChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.discrete.DiscreteChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.discrete.DiscreteEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.discrete.DiscreteModule;
import at.jku.iic.microfluidics.droplet.basic.architecture.discrete.DiscretePump;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalPump;

/**
 * A converter which takes a fully specified physical biochip (i.e. given as
 * resistance model) and converts it to the discrete model.<br>
 * For details we refer to our DAC'17 work.
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public class PhysicalToDiscreteBiochipConverter {

	/**
	 * Converts a fully specified physical biochip to a discrete biochip.
	 * 
	 * @param chip The fully specified biochip in the resistance model.
	 * @return The biochip in the discrete model.
	 */
	public static DiscreteBiochip convert(PhysicalBiochip chip, double resolutionFactor) {

		Map<Integer, EndPoint<DiscreteEdge>> endPoints = new HashMap<>();
		List<DiscreteEdge> edges = new LinkedList<>();

		// Create new physical EndPoints
		for (PhysicalEdge e : chip) {
			for (EndPoint<PhysicalEdge> ep : Arrays.asList(e.src, e.dst)) {
				if (!endPoints.containsKey(ep.id)) {
					if (ep.isLeeSorter) {
						endPoints.put(ep.id, new EndPoint<DiscreteEdge>(ep.id, Double.NaN, Double.NaN));
					} else {
						endPoints.put(ep.id, new EndPoint<DiscreteEdge>(ep.id));
					}
				}
			}
		}

		TimesForDiscreteModelCaller caller = new TimesForDiscreteModelCaller();
		double[][] times = caller.determineTimes(chip, false);
		int i = 0;

		// Iterate over all edges and create discrete equivalent edges.
		for (PhysicalEdge e : chip) {

			if (e.isBypassChannel()) {
				edges.add(new DiscreteBypassChannel(endPoints.get(e.src.id), endPoints.get(e.dst.id), e.name));
			} else if (e.isChannel()) {
				edges.add(new DiscreteChannel(endPoints.get(e.src.id), endPoints.get(e.dst.id), e.name, times[i][0],
						times[i][1], times[i][2], resolutionFactor));
			} else if (e.isModule()) {
				edges.add(new DiscreteModule(endPoints.get(e.src.id), endPoints.get(e.dst.id), e.name, times[i][0],
						times[i][1], times[i][2], resolutionFactor, ((Module) e).getOperation()));
			}
			i++;
		}

		List<DiscretePump> physicalPump = new LinkedList<>();
		for (PhysicalPump p : chip.getPumps()) {
			physicalPump.add(new DiscretePump(endPoints.get(p.src.id), endPoints.get(p.sink.id), p.name));
		}
		return new DiscreteBiochip(edges, physicalPump);
	}

	private PhysicalToDiscreteBiochipConverter() {

	}
}
