/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.converter;

import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.architecture.converter.dimensionator.ChannelDimensionator;
import at.jku.iic.microfluidics.droplet.basic.architecture.converter.dimensionator.ChannelSectionDimensionator;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.DefaultPhysicalModule;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBypassChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalPump;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.VolumetricFlowRatePump;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructureBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructureEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructureModule;
import at.jku.iic.microfluidics.droplet.basic.architecture.structure.StructurePump;
import at.jku.iic.microfluidics.droplet.basic.physics.DimensioningProperties;
import at.jku.iic.microfluidics.droplet.basic.physics.Dimensions;
import at.jku.iic.microfluidics.droplet.basic.physics.FluidProperties;

public class StructureToPhysicalBiochipConverter {

    private static final Logger LOGGER = Logger.getLogger(StructureToPhysicalBiochipConverter.class.getSimpleName());

    /**
     * Converts a structure-only biochip into a physical-biochip. Therefore, standard resistance values for
     * modules, forwarding-channels, bypass-channels and bifurcation-channels are used. The dimensions of the
     * remaining channels are not fully specified.
     * 
     * @param chip
     *            The structure of the biochip.
     * @return The physical-equivalent biochip.
     */
    public static PhysicalBiochip convert(StructureBiochip chip, ChannelSectionDimensionator channelDimensionator, Dimensions dim,
            FluidProperties fluidProperties, DimensioningProperties dimensioningProperties, double headerDropletVolume,
            double payloadDropletVolume, DefaultSuccessorSelectionStrategy selectionStrategy) {

        Map<Integer, EndPoint<PhysicalEdge>> endPoints = new HashMap<>();
        List<PhysicalEdge> edges = new LinkedList<>();

        // Create new physical EndPoints
        for (StructureEdge e : chip) {
            for (EndPoint<StructureEdge> ep : Arrays.asList(e.src, e.dst)) {
                if (!endPoints.containsKey(ep.id)) {
                    if (ep.isLeeSorter) {
                        endPoints.put(ep.id, new EndPoint<PhysicalEdge>(ep.id, dim.leeSortingLength, dim.leeSortingWidth));
                    } else {
                        endPoints.put(ep.id, new EndPoint<PhysicalEdge>(ep.id));
                    }
                }
            }
        }

        // Iterate over all edges and create physical equivalent edges.
        for (StructureEdge e : chip) {

            if (e.isBypassChannel()) {
                edges.add(new PhysicalBypassChannel(endPoints.get(e.src.id), endPoints.get(e.dst.id), e.name));
            } else if (e.isForwardingChannel()) {
                // This is a forwarding channel
                edges.add(new PhysicalChannel(endPoints.get(e.src.id), endPoints.get(e.dst.id), e.name, dim.forwardingChannelWidth,
                        dim.forwardingChannelHeight, dim.forwardingChennelLength));
            } else if (e.isChannel()) {

                if (!e.src.isBifurcation()) {
                    // A channel which is not a successor channel of a bifurcation
                    if (channelDimensionator instanceof ChannelDimensionator) {
                        edges.add(new PhysicalChannel(endPoints.get(e.src.id), endPoints.get(e.dst.id), e.name,
                                channelDimensionator.getWidth(e.name), channelDimensionator.getHeight(e.name),
                                ((ChannelDimensionator) channelDimensionator).getLength(e.name)));
                    } else {
                        edges.add(new PhysicalChannel(endPoints.get(e.src.id), endPoints.get(e.dst.id), e.name,
                                channelDimensionator.getWidth(e.name), channelDimensionator.getHeight(e.name)));
                    }
                }

                // Check if the destination of the channel is connected to a bifurcation
                if (e.dst.isBifurcation()) {
                    // If so, create physical successor channels
                    StructureEdge defaultSuccessor = selectionStrategy.getDefaultSuccessor(e.dst);
                    StructureEdge nonDefaultSuccessor = selectionStrategy.getNonDefaultSuccessor(e.dst);
                    assert defaultSuccessor != nonDefaultSuccessor;
                    
                    PhysicalChannel pSucc1 = new PhysicalChannel(endPoints.get(defaultSuccessor.src.id),
                            endPoints.get(defaultSuccessor.dst.id), defaultSuccessor.name, dim.channelWidth, dim.channelHeight,
                            dim.bifurcationDefaultSuccessorLength);
                    PhysicalChannel pSucc2 = new PhysicalChannel(endPoints.get(nonDefaultSuccessor.src.id),
                            endPoints.get(nonDefaultSuccessor.dst.id), nonDefaultSuccessor.name, dim.channelWidth, dim.channelHeight,
                            dim.bifurcationNonDefaultSuccessorLength);

                    edges.add(pSucc1);
                    edges.add(pSucc2);
                }
            } else if (e.isModule()) {
                // Note: A module consists of two edges. This is a requirement for the Matlab-Simulator
                // (i.e. otherwise the graph would contain double edges which are not supported; the module
                // and the forwarding channel would represent a double edge)

                if (e.dst.getSuccessors().size() == 1 && e.dst.getSuccessors().get(0).isModule()) {
                    // First part of the module
                    edges.add(new DefaultPhysicalModule(endPoints.get(e.src.id), endPoints.get(e.dst.id), e.name,
                            ((StructureModule) e).getOperation(), dim.moduleWidth1, dim.moduleHeight1, dim.moduleLength1));
                } else {
                    // Second part of the module
                    edges.add(new DefaultPhysicalModule(endPoints.get(e.src.id), endPoints.get(e.dst.id), e.name,
                            ((StructureModule) e).getOperation(), dim.moduleWidth2, dim.moduleHeight2, dim.moduleLength2));
                }
            }
        }
        List<PhysicalPump> physicalPumps = new LinkedList<>();
        for (StructurePump p : chip.getPumps()) {
            LOGGER.info("By default a pump producing a volumetric flow rate is used.");
            physicalPumps.add(new VolumetricFlowRatePump(endPoints.get(p.src.id), endPoints.get(p.sink.id), p.name,
                    dimensioningProperties.inputFlowRate));
        }
        return new PhysicalBiochip(edges, physicalPumps, fluidProperties, headerDropletVolume, payloadDropletVolume);
    }

    private StructureToPhysicalBiochipConverter() {

    }
}
