/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.converter;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import Jama.Matrix;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.CloggablePhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PressurePump;
import at.jku.iic.microfluidics.droplet.basic.calculator.FlowCalculator;

public class TimesForDiscreteModelCaller {

	public TimesForDiscreteModelCaller() {
		super();
	}

	/**
	 * Wrapper for getTimesForDiscreteModel (see below)
	 * @param chip
	 * @param debugResOutput
	 * @return
	 */
	public double[][] determineTimes(PhysicalBiochip chip, boolean debugResOutput) {
		Matrix res = getTimesForDiscreteModel(chip, Collections.emptySet());
		return res.getArray();
	}

	/**
	 * Calculates times for the chip/channels with and w/o header/payload droplets (formerly getTimesForDiscreteModel.m)
	 * @param chip
	 * @param cloggedChannels
	 * @return Matrix T of size: n_channels x 3
	 * 				0th column: for each channel time value without droplet
	 * 				1st column: for each channel time value with header droplet (additional resistance)
	 * 				2nd column: for each channel time value with payload droplet (additional resistance)
	 */
	public Matrix getTimesForDiscreteModel(PhysicalBiochip chip, Set<CloggablePhysicalChannel> cloggedChannels) {

		//TODO: Change interface to return a map edge->time instead of a matrix with unknown ordering?
		//TODO: Find out what f_c (10e-11 in unit test) did ... works fine without.
		// The following values could also be passed as function arguments
		double factor = 0.001;
		double mu_c = chip.fluidProperties.muC;
		double mu_d = chip.fluidProperties.muD;
		double header_volume = chip.headerDropletVolume;
		double payload_volume = chip.payloadDropletVolume;
		
		
		FlowCalculator fc = new FlowCalculator(chip, true);
		
		// Handle mapping of FlowCalculator's row indices and indices here
		// (might have a different number of permutated rows):
		
		// FlowCalculator's internal indexMap
		Map<Object, Integer> fcMap = fc.getIndexMap();
		
		// Indices same as fcMap but may have less entries
		Map<Object, Integer> solverMap = new IdentityHashMap<>();
		
		// Local rowIndex-to-edge map
		Map<Integer, Object> localMap = new IdentityHashMap<>();
		
		// Hold widths/heights/length temporarily (order of localMap is imposed)
		List<Double> wList = new ArrayList<Double>();
		List<Double> hList = new ArrayList<Double>();
		List<Double> lList = new ArrayList<Double>();
		
		// Gather physical edges (widths/heights/lengths) and set up indices
		int k = 0;
		//TODO: Ordering affects results
		for(Object edge : fc.getGraph().edgeSet()) {
			if(edge instanceof PhysicalEdge) {
				if (!cloggedChannels.contains(edge)) {
					solverMap.put(edge, fcMap.get(edge));
					localMap.put(k, edge);
					wList.add(k, ((PhysicalEdge) edge).getWidth());
					hList.add(k, ((PhysicalEdge) edge).getHeight());
					lList.add(k, ((PhysicalEdge) edge).getLength());
		        }
			} else if(edge instanceof PressurePump) {
				// Taken from SimulatorAccessor.java (deleted), might need to be adapted
				solverMap.put(edge, fcMap.get(edge));
				localMap.put(k, edge);
				wList.add(k, ((PhysicalEdge) edge).getWidth());
				hList.add(k, ((PhysicalEdge) edge).getHeight());
				lList.add(k, ((PhysicalEdge) edge).getLength());
			}
			k++;
		}
		
		// Transitive result of solverMap(localMap(i)), 
		// i.e. indices of local rows here mapping to row indices of FlowCalculator
		Map<Integer, Integer> remap = new HashMap<Integer, Integer>();
		for(Map.Entry<Integer, Object> e : localMap.entrySet()) {
			remap.put(e.getKey(), fcMap.get(e.getValue()));
		}
		
		Matrix W = new Matrix(wList.stream().mapToDouble(o -> o).toArray(), 1);
		Matrix H = new Matrix(hList.stream().mapToDouble(o -> o).toArray(), 1);
		Matrix L = new Matrix(lList.stream().mapToDouble(o -> o).toArray(), 1);

		// Result matrix
		Matrix T = new Matrix(L.getColumnDimension(), 3);

		// Call FlowCalculator's results
		Matrix x = fc.solve(new HashMap<PhysicalEdge, Double>(), factor);
		Matrix Q = remapColumnVector(x, remap);

		// Calculates times without droplets and store to T (0th column)
		Matrix WH = W.arrayTimes(H);
		Matrix speed = Q.transpose();
		speed.arrayRightDivideEquals(WH);
		Matrix tmp = L.arrayRightDivide(speed);
		T.setMatrix(0, T.getRowDimension() - 1, 0, 0, tmp.transpose());

		// For each edge, calculate the times when a header droplet is in the channel (1st column); 
		// and when a payload droplet is inside the channel (2nd column)
		for(Map.Entry<Integer, Object> entry : localMap.entrySet()) {
			if(!(entry.getValue() instanceof PhysicalEdge) || !((PhysicalEdge) entry.getValue()).isBypassChannel()) {
				// It is not a bypass

				// Header
				int i = entry.getKey();

				// Calculate the actual length of the header droplet in the
				// current channel i
				double header_length = header_volume / WH.get(0, i);
				double R_drop = L_to_R(W.get(0, i), H.get(0, i), header_length, mu_c, mu_d) * factor;
				
				// Q contains all the flow rates of the channels (generators included)
				Map<PhysicalEdge, Double> resistances = new HashMap<PhysicalEdge, Double>();
				if(entry.getValue() instanceof PhysicalEdge) {
					resistances.put((PhysicalEdge) entry.getValue(), R_drop);
				}
				x = fc.solve(resistances, factor);
				Q = remapColumnVector(x, remap);

				// Initialization of the matrix that will contain the speed of each branch at
				// each event
				speed = Q.transpose();
				speed.arrayRightDivideEquals(WH);

				tmp = L.arrayRightDivide(speed);
				T.set(i, 1, tmp.get(0, i));

				// Calculate payload time
				// starting channel for the header

				// Calculate the actual length of the payload droplet in the
				// current channel c_d
				double payload_length = payload_volume / WH.get(0, i);
				R_drop = L_to_R(W.get(0, i), H.get(0, i), payload_length, mu_c, mu_d) * factor;

				// Q contains all the flow rates of the channels (generators included)
				resistances.clear();
				resistances = new HashMap<PhysicalEdge, Double>();
				if(entry.getValue() instanceof PhysicalEdge) {
					resistances.put((PhysicalEdge) entry.getValue(), R_drop);
				}
				x = fc.solve(resistances, factor);
				Q = remapColumnVector(x, remap);

				// Initialization of the matrix that will contain the speed of each branch at
				// each event
				speed = Q.transpose();
				speed.arrayRightDivideEquals(WH);

				tmp = L.arrayRightDivide(speed);
				T.set(i, 2, tmp.get(0, i));
			}
		}

		return T;
	}

	/**
	 * Calculates the additional resistance due to a droplet in the channel
	 * @param w width
	 * @param h height
	 * @param l length
	 * @param mu_c mu of continuous phase
	 * @param mu_d mu of droplet
	 * @return resistance
	 */
	public double L_to_R(double w, double h, double l, double mu_c, double mu_d) {
		// Hagen-Poiseuille geometric parameter for each channel
		double a = 12 * Math.pow(1 - 192 * h / (Math.pow(Math.PI, 5) * w) * Math.tanh(Math.PI * w / (2 * h)), -1);

		// the additional resistance due to the droplet in that channel
		return l * a * (mu_d - mu_c) / (w * Math.pow(h, 3));
	}

	/**
	 * Calculates the resistance of a channel without a droplet
	 * @param w width
	 * @param h height
	 * @param l length
	 * @param mu_c mu of continuous phase
	 * @return resistance
	 */
	public double L_to_R(double w, double h, double l, double mu_c) {
		// Hagen-Poiseuille geometric parameter for each channel
		double a = 12 * Math.pow(1 - 192 * h / (Math.pow(Math.PI, 5) * w) * Math.tanh(Math.PI * w / (2 * h)), -1);

		// if the number of input parameters is 4 it means that the microfluidic
		// channel contains just continuous phase
		return l * a * (mu_c) / (w * Math.pow(h, 3));
	}
	
	/**
	 * Remaps the result vector of FlowCalculator::solve to the needed order in here
	 * @param v input vector, i.e. (1,2,3,4)
	 * @param map mapping from TimesForDiscrete index to FlowCalculator index (kind of the wrong way around, for technical simplicity!), i.e. (0 -> 3, 3 -> 1, 2 -> 2, 1 -> 0)
	 * @return remapped vector, i.e. (4,1,3,0)
	 */
	public Matrix remapColumnVector(Matrix v, Map<Integer, Integer> map) {
		Matrix r = new Matrix(map.size(), 1);
		
		for(Map.Entry<Integer, Integer> e : map.entrySet()) {
			r.setMatrix(e.getKey(), e.getKey(), 0, 0, v.getMatrix(e.getValue(), e.getValue(), 0, 0));
		}
		
		return r;
	}

}
