/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.discrete;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import org.json.JSONArray;
import org.json.JSONObject;
import at.jku.iic.microfluidics.droplet.basic.architecture.Biochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.architecture.Path;
import at.jku.iic.microfluidics.droplet.basic.graph.utility.GraphUtility;

import java.util.Set;

public class DiscreteBiochip extends Biochip<DiscreteEdge, DiscretePump> {

    public DiscreteBiochip(List<DiscreteEdge> edges, List<DiscretePump> pump) {
        super(edges, pump);
        
        // Check if there is a unique ordering of the successor channels
        // of a bifurcation.
        for(DiscreteEdge edge : edges) {
            if(edge.src.isBifurcation()) {
                Set<Integer> staticTimes = new HashSet<>();
                for(DiscreteEdge succ : edge.src.getSuccessors()) {
                    if(!staticTimes.add(succ.staticTimeSteps)) {
                        throw new RuntimeException("Too small resolution factor! Please choose a larger resolution factor.");
                    }
                }
            }
        }
    }

    public int getMaxPath() {

        int max = -1;
        for (Entry<DiscretePump, List<Path<DiscreteEdge>>> pathsOfPump : GraphUtility.findAllDropletPaths(this).entrySet()) {

            max = Integer.max(max, pathsOfPump.getValue().stream().mapToInt(e -> getStaticDuration(e)).max().getAsInt());
        }
        return max;
    }

    public static int getStaticDuration(Path<DiscreteEdge> path) {

        return path.getEdges().stream()
                .mapToInt(r -> r.timeStepsForPayload > r.timeStepsForHeader ? r.timeStepsForPayload : r.timeStepsForHeader).sum();
    }

    public JSONObject toJSON(){
        JSONObject json = new JSONObject();

        //endPoints
        //all relevant endPoints are stored during the loop through the pumps and edges
        Set<EndPoint<DiscreteEdge>> endPointIds = new HashSet<>();

        //pumps
        JSONArray jsonPumps = new JSONArray();
        json.put("pumps", jsonPumps);
        for (DiscretePump pump : getPumps()) {
            endPointIds.add(pump.src);
            endPointIds.add(pump.sink);
            jsonPumps.put(pump.toJSON());
        }

        //edges
        JSONArray jsonEdges = new JSONArray();
        json.put("edges", jsonEdges);
        for (DiscreteEdge edge : getEdges()) {
            endPointIds.add(edge.src);
            endPointIds.add(edge.dst);
            jsonEdges.put(edge.toJSON());
        }

        //write stored endPoints
        JSONArray jsonEndPoints = new JSONArray();
        json.put("endPoints", jsonEndPoints);
        for (EndPoint<DiscreteEdge> endPointId : endPointIds) {
            jsonEndPoints.put(endPointId.toJSON());
        }

        return json;
    }

    public static DiscreteBiochip fromJSON(JSONObject json) {
        //endPoints
        JSONArray jsonEndPoints = json.getJSONArray("endPoints");
        Map<Integer, EndPoint<DiscreteEdge>> endPoints = new HashMap<>();
        for (int i = 0; i < jsonEndPoints.length(); i++) {
            JSONObject jsonEndPoint = jsonEndPoints.getJSONObject(i);
            int id = jsonEndPoint.getInt("id");
            boolean isLeeSorter = jsonEndPoint.getBoolean("isLeeSorter");
            double leeSortingLength = jsonEndPoint.getDouble("leeSortingLength");
            double leeSortingWidth = jsonEndPoint.getDouble("leeSortingWidth");

            EndPoint<DiscreteEdge> endPoint;
            if (isLeeSorter) {
                endPoint = new EndPoint<>(id, leeSortingLength, leeSortingWidth);
            } else{
                endPoint = new EndPoint<>(id);
            }
            
            endPoints.put(id, endPoint);
        }

        //pumps
        JSONArray jsonPumps = json.getJSONArray("pumps");
        List<DiscretePump> pumps = new ArrayList<>();
        for (int i = 0; i < jsonPumps.length(); i++) {
            JSONObject jsonPump = jsonPumps.getJSONObject(i);
            // String type = jsonPump.getString("type");
            String name = jsonPump.getString("name");
            int srcId = jsonPump.getInt("srcId");
            int dstId = jsonPump.getInt("dstId");

            pumps.add(new DiscretePump(endPoints.get(srcId), endPoints.get(dstId), name));
        }

        //edges
        JSONArray jsonEdges = json.getJSONArray("edges");
        List<DiscreteEdge> edges = new ArrayList<>();
        for (int i = 0; i < jsonEdges.length(); i++) {
            JSONObject jsonEdge = jsonEdges.getJSONObject(i);
            String type = jsonEdge.getString("type");
            String name = jsonEdge.getString("name");
            int srcId = jsonEdge.getInt("srcId");
            int dstId = jsonEdge.getInt("dstId");
            int staticTimeSteps = jsonEdge.getInt("staticTimeSteps");
            int timeStepsForHeader = jsonEdge.getInt("timeStepsForHeader");
            int timeStepsForPayload = jsonEdge.getInt("timeStepsForPayload");

            if (type.equals(DiscreteChannel.class.getSimpleName())) {
                edges.add(new DiscreteChannel(endPoints.get(srcId), endPoints.get(dstId), name, staticTimeSteps, timeStepsForHeader, timeStepsForPayload, 1));
            } else if (type.equals(DiscreteBypassChannel.class.getSimpleName())) {
                edges.add(new DiscreteBypassChannel(endPoints.get(srcId), endPoints.get(dstId), name));
            } else if (type.equals(DiscreteModule.class.getSimpleName())) {
                char operation = jsonEdge.getString("operation").charAt(0);
                edges.add(new DiscreteModule(endPoints.get(srcId), endPoints.get(dstId), name, staticTimeSteps, timeStepsForHeader, timeStepsForPayload, 1, operation));
            }
        }

        return new DiscreteBiochip(edges, pumps);
    }
}