/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.discrete;

import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import org.json.JSONObject;
import at.jku.iic.microfluidics.droplet.basic.architecture.Edge;
import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;

/**
 * Represents an edge in the discrete model. For the definition of the discrete model, we refer to our DAC'17
 * work.<br>
 * In the discrete model each edge has the following three different kinds of time steps:
 * <li><code>staticTimeSteps</code>: Represents the amount of time steps which a particle of the continuous
 * flow would require to pass an edge in a droplet-free NLoC. This is only used to determine the default
 * direction!
 * <li><code>timeStepsForHeader</code>: Represents the amount of time steps which a header droplet requires to
 * pass this edge.
 * <li><code>timeStepsForPayload</code>: Represents the amount of time steps which a payload droplet requires
 * to pass this edge.
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public abstract class DiscreteEdge extends Edge<DiscreteEdge> {

    /**
     * Represents the amount of time steps which a particle of the continuous flow would require to pass an
     * edge in a droplet-free NLoC.
     */
    public final int staticTimeSteps;

    /**
     * Represents the amount of time steps which a header droplet requires to pass this edge.
     */
    public final int timeStepsForHeader;

    /**
     * Represents the amount of time steps which a payload droplet requires to pass this edge.
     */
    public final int timeStepsForPayload;

    public DiscreteEdge(EndPoint<DiscreteEdge> src, EndPoint<DiscreteEdge> dst, String name, double staticTime, double timeOfHeader,
            double timeOfPayload, double resolutionFactor) {
        super(src, dst, name);
        this.staticTimeSteps = (int) Math.round(staticTime * resolutionFactor);
        this.timeStepsForHeader = (int) Math.round(timeOfHeader * resolutionFactor);
        this.timeStepsForPayload = (int) Math.round(timeOfPayload * resolutionFactor);

        if (!isBypassChannel()) {
            if (this.staticTimeSteps == 0 || this.timeStepsForHeader == 0 || this.timeStepsForPayload == 0) {
                // When it is not a bypass channel, the time steps have to be greater than zero
                throw new RuntimeException("Too small resolution factor! Please choose a larger resolution factor.");
            }
        }
    }

    /**
     * Returns a sorted list with all successors which can be entered by header droplets. The sorting is based
     * on a natural ordering of the <code>staticTimeSteps</code>.
     */
    @Override
    public List<DiscreteEdge> getHeaderSuccessorEdges() {

        Stream<DiscreteEdge> stream = dst.getSuccessors().stream().filter(e -> !e.isBypassChannel());
        if (dst.isLeeSorter) {
            // In this case, make a discrete decision where the header flows into.
            stream = stream.filter(e -> e.isChannel());
        }
        return stream.sorted(new StaticTimesComparator()).collect(Collectors.toList());
    }

    /**
     * Returns a sorted set with all successors which can be entered by payload droplets. The sorting is based
     * on a natural ordering of the <code>staticTimeSteps</code>.
     */
    @Override
    public List<DiscreteEdge> getPayloadSuccessorEdges() {

        Stream<DiscreteEdge> stream = dst.getSuccessors().stream().filter(e -> !e.isBypassChannel());
        if (dst.isLeeSorter) {
            // In this case, make a discrete decision where the header flows into.
            stream = stream.filter(e -> e.isModule());
        }
        return stream.sorted(new StaticTimesComparator()).collect(Collectors.toList());
    }

    private class StaticTimesComparator implements Comparator<DiscreteEdge> {

        @Override
        public int compare(DiscreteEdge o1, DiscreteEdge o2) {

            return Integer.compare(o1.staticTimeSteps, o2.staticTimeSteps);
        }
    }

    abstract public JSONObject toJSON();
}
