/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.physical;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;

import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.calculator.PressureCalculator;

public class LaplacePressureParameter {

    private final EndPoint<PhysicalEdge> measuredPoint;

    private final EndPoint<PhysicalEdge> referencePoint;

    private final double gapWidth;

    private final double dropletRadius;

    private final List<PhysicalEdge> emptyEdges;

    /**
     * 
     * @param pressureGradient
     * @param gapWidth
     * @param dropletRadius
     * @param emptyEdges
     *            If one of these edges contains a droplet, a Laplace violation is not reported. This is
     *            required when merging of droplets is allowed as e.g. in trapping wells.
     */
    public LaplacePressureParameter(EndPoint<PhysicalEdge> measuredPoint, EndPoint<PhysicalEdge> referencePoint, double gapWidth,
            double dropletRadius, List<PhysicalEdge> emptyEdges) {
        super();
        this.measuredPoint = measuredPoint;
        this.referencePoint = referencePoint;

        this.gapWidth = gapWidth;
        this.dropletRadius = dropletRadius;
        this.emptyEdges = emptyEdges;
    }

    public List<PhysicalEdge> getEmptyEdges() {

        return Collections.unmodifiableList(emptyEdges);
    }

    public boolean isPressureExceeding(Map<PhysicalEdge, Double> flowRates, Set<CloggablePhysicalChannel> cloggedChannels, PhysicalBiochip chip) {

        Map<EndPoint<PhysicalEdge>, Double> pressureGradients = PressureCalculator.calculatePressureGradients(flowRates, cloggedChannels,
                referencePoint, chip);
        return pressureGradients.get(measuredPoint) > calculateLaplacePressure(chip.fluidProperties.interfacialTension);
    }

    /**
     * Helper method for calculating the Laplace pressure
     * 
     * @param gapWidth
     * @param dropletRadius
     * @param interfacialTension
     * @return
     */
    public double calculateLaplacePressure(double interfacialTension) {

        return interfacialTension * (2.0 / getGapWidth() - 1.0 / getDropletRadius());
    }

    public double getGapWidth() {

        return gapWidth * getDefectGapWidth();
    }

    
    public EndPoint<PhysicalEdge> getMeasuredPoint() {
    
        return measuredPoint;
    }

    
    public EndPoint<PhysicalEdge> getReferencePoint() {
    
        return referencePoint;
    }

    
    public double getDropletRadius() {
    
        return dropletRadius * getDefectDropletRadius();
    }
    
    protected double getDefectGapWidth() {
    	return 1.0;
    }
    protected double getDefectDropletRadius() {
    	return 1.0;
    }

}
