/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.physical;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.json.JSONArray;
import org.json.JSONObject;

import at.jku.iic.microfluidics.droplet.basic.architecture.Biochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.physics.FluidProperties;

public class PhysicalBiochip extends Biochip<PhysicalEdge, PhysicalPump> {

    public final FluidProperties fluidProperties;
    
    /**
     * The volume of the header droplet
     */
    public final double headerDropletVolume;

    /**
     * The volume of the payload droplet
     */
    public final double payloadDropletVolume;
    
    public PhysicalBiochip(List<PhysicalEdge> edges, List<PhysicalPump> pumps, FluidProperties fluidProperties, double headerDropletVolume,
            double payloadDropletVolume) {
        super(edges, pumps);
        this.fluidProperties = fluidProperties;
        this.headerDropletVolume = headerDropletVolume;
        this.payloadDropletVolume = payloadDropletVolume;
    }

    public List<PressurePump> getPressurePumps() {

        return getPumps().stream().filter(p -> p instanceof PressurePump).map(p -> ((PressurePump) p)).collect(Collectors.toList());
    }

    public List<VolumetricFlowRatePump> getVolumetricFlowRatePumps() {

        return getPumps().stream().filter(p -> p instanceof VolumetricFlowRatePump).map(p -> ((VolumetricFlowRatePump) p))
                .collect(Collectors.toList());
    }

    /**
     * @return Returns an array containing the lengths of the edges. The order is defined by the internal
     *         list.
     */
    public double[] getLengths() {

        return getEdges().stream().map(PhysicalEdge::getLength).mapToDouble(x -> x).toArray();
    }

    /**
     * @return Returns an array containing the widths of the edges. The order is defined by the internal list.
     */
    public double[] getWidths() {

        return getEdges().stream().map(PhysicalEdge::getWidth).mapToDouble(x -> x).toArray();
    }

    /**
     * @return Returns an array containing the heights of the edges. The order is defined by the internal
     *         list.
     */
    public double[] getHeights() {

        return getEdges().stream().map(PhysicalEdge::getHeight).mapToDouble(x -> x).toArray();
    }

    /**
     * Checks if all <code>PhysicalChannel</code> edges are fully specified.
     * 
     * @return Returns <code>true</code> if all channels are fully specified, <code>false</code> otherwise.
     */
    public boolean isFullySpecified() {

        for (PhysicalEdge pc : getEdges()) {
            if (pc instanceof PhysicalChannel && !((PhysicalChannel) pc).isFullySpecified()) {
                return false;
            }
            assert !Double.isNaN(pc.getResistance(fluidProperties.muC));
        }
        return true;
    }

    public String serialize() {

        StringBuilder sb = new StringBuilder();
        ComponentCounter counter = new ComponentCounter();
        for (PlaceAndRouteSerialization p : getPumps()) {
            sb.append(p.getStringRepresentation(counter));
        }
        for (PhysicalEdge e : getEdges()) {
            if (e instanceof PlaceAndRouteSerialization) sb.append(((PlaceAndRouteSerialization) e).getStringRepresentation(counter));
        }
        return sb.toString();
    }

    public JSONObject toJSON(){
        JSONObject json = new JSONObject();
        json.put("fluidProperties", fluidProperties.toJSON());
        json.put("headerDropletVolume", headerDropletVolume);
        json.put("payloadDropletVolume", payloadDropletVolume);

        //endPoints
        //all relevant endPoints are stored during the loop through the pumps and edges
        Set<EndPoint<PhysicalEdge>> endPointIds = new HashSet<>();

        //pumps
        JSONArray jsonPumps = new JSONArray();
        json.put("pumps", jsonPumps);
        for (PhysicalPump pump : getPumps()) {
            endPointIds.add(pump.src);
            endPointIds.add(pump.sink);
            jsonPumps.put(pump.toJSON());
        }

        //edges
        JSONArray jsonEdges = new JSONArray();
        json.put("edges", jsonEdges);
        for (PhysicalEdge edge : getEdges()) {
            endPointIds.add(edge.src);
            endPointIds.add(edge.dst);
            jsonEdges.put(edge.toJSON());
        }

        //write stored endPoints
        JSONArray jsonEndPoints = new JSONArray();
        json.put("endPoints", jsonEndPoints);
        for (EndPoint<PhysicalEdge> endPointId : endPointIds) {
            jsonEndPoints.put(endPointId.toJSON());
        }

        return json;
    }

    public static PhysicalBiochip fromJSON(String jsonString) {
        JSONObject json = new JSONObject(jsonString);
        return fromJSON(json);
    }

    public static PhysicalBiochip fromJSON(JSONObject json) {
        FluidProperties fluidProperties = FluidProperties.fromJSON(json.getJSONObject("fluidProperties"));
        double headerDropletVolume = json.getDouble("headerDropletVolume");
        double payloadDropletVolume = json.getDouble("payloadDropletVolume");

        //endPoints
        JSONArray jsonEndPoints = json.getJSONArray("endPoints");
        Map<Integer, EndPoint<PhysicalEdge>> endPoints = new HashMap<>();
        for (int i = 0; i < jsonEndPoints.length(); i++) {
            JSONObject jsonEndPoint = jsonEndPoints.getJSONObject(i);
            int id = jsonEndPoint.getInt("id");
            boolean isLeeSorter = jsonEndPoint.getBoolean("isLeeSorter");
            double leeSortingLength = jsonEndPoint.getDouble("leeSortingLength");
            double leeSortingWidth = jsonEndPoint.getDouble("leeSortingWidth");

            EndPoint<PhysicalEdge> endPoint;
            if (isLeeSorter) {
                endPoint = new EndPoint<>(id, leeSortingLength, leeSortingWidth);
            } else{
                endPoint = new EndPoint<>(id);
            }
            
            endPoints.put(id, endPoint);
        }

        //pumps
        JSONArray jsonPumps = json.getJSONArray("pumps");
        List<PhysicalPump> pumps = new ArrayList<>();
        for (int i = 0; i < jsonPumps.length(); i++) {
            JSONObject jsonPump = jsonPumps.getJSONObject(i);
            String type = jsonPump.getString("type");
            String name = jsonPump.getString("name");
            int srcId = jsonPump.getInt("srcId");
            int dstId = jsonPump.getInt("dstId");

            if(type.equals(PressurePump.class.getSimpleName())) {
                double inputPressure = jsonPump.getDouble("inputPressure");
                pumps.add(new PressurePump(endPoints.get(srcId), endPoints.get(dstId), name, inputPressure));
            } else if (type.equals(VolumetricFlowRatePump.class.getSimpleName())) {
                double inputFlowRate = jsonPump.getDouble("inputFlowRate");
                pumps.add(new VolumetricFlowRatePump(endPoints.get(srcId), endPoints.get(dstId), name, inputFlowRate));
            }
        }

        //edges
        JSONArray jsonEdges = json.getJSONArray("edges");
        List<PhysicalEdge> edges = new ArrayList<>();
        for (int i = 0; i < jsonEdges.length(); i++) {
            JSONObject jsonEdge = jsonEdges.getJSONObject(i);
            String type = jsonEdge.getString("type");
            String name = jsonEdge.getString("name");
            int srcId = jsonEdge.getInt("srcId");
            int dstId = jsonEdge.getInt("dstId");
            double height = jsonEdge.getDouble("height");
            double width = jsonEdge.getDouble("width");
            double length = Double.NaN;
            if (jsonEdge.has("length")) {
                length = jsonEdge.getDouble("length");
            }

            if(type.equals(PhysicalChannel.class.getSimpleName())) {
                if(Double.isNaN(length)){
                    edges.add(new PhysicalChannel(endPoints.get(srcId), endPoints.get(dstId), name, width, height));
                } else {
                    edges.add(new PhysicalChannel(endPoints.get(srcId), endPoints.get(dstId), name, width, height, length));
                }
            } else if (type.equals(PhysicalBypassChannel.class.getSimpleName())) {
                edges.add(new PhysicalBypassChannel(endPoints.get(srcId), endPoints.get(dstId), name, width, height, length));
            } else if (type.equals(CloggablePhysicalChannel.class.getSimpleName())) {
                edges.add(new CloggablePhysicalChannel(endPoints.get(srcId), endPoints.get(dstId), name, width, height, length));
            } else if (type.equals(DefaultPhysicalModule.class.getSimpleName())) {
                char operation = jsonEdge.getString("operation").charAt(0);
                edges.add(new DefaultPhysicalModule(endPoints.get(srcId), endPoints.get(dstId), name, operation, width, height, length));
            }
        }

        return new PhysicalBiochip(edges, pumps, fluidProperties, headerDropletVolume, payloadDropletVolume);
    }
}
