/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.physical;

import org.json.JSONObject;

import at.jku.iic.microfluidics.droplet.basic.architecture.Channel;
import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;

/**
 * @author Andreas
 *
 */
public class PhysicalChannel extends PhysicalEdge implements Channel, PlaceAndRouteSerialization  {

    private boolean fullySpecified;

    public PhysicalChannel(EndPoint<PhysicalEdge> src, EndPoint<PhysicalEdge> dst, String name, double width, double height) {

        super(src, dst, name, width, height, Double.NaN);
        this.fullySpecified = false;
    }

    public PhysicalChannel(EndPoint<PhysicalEdge> src, EndPoint<PhysicalEdge> dst, String name, double width, double height,
            double length) {

        super(src, dst, name, width, height, length);
        this.fullySpecified = true;
    }

    public boolean isFullySpecified() {

        return fullySpecified;
    }

    @Override
    public double getResistance(double muC) {

        if (!isFullySpecified()) return Double.NaN;
        return getResistance(muC, width, height, length);
    }

    /**
     * Calculates and sets the channel length by using the provided resistance.
     * 
     * @param resistance
     *            The fluidic resistance of the channel.
     * @return Returns the calculated and set length.
     */
    public double setChannelLength(double resistance, double muC) {

        fullySpecified = true;
        length = resistance / (getAFactor(width, height) * muC / (width * Math.pow(height, 3)));
        return length;
    }

    @Override
    public String getStringRepresentation(ComponentCounter componentCounter) {

        if (src.isBifurcation()) return "";

        StringBuilder sb = new StringBuilder();
        
        if (dst.isBifurcation()) {
            sb.append("bifurcation").append(';');
            sb.append("bif").append(componentCounter.getBifurcationCounter()).append(';');            
            componentCounter.incBifurcationCounter();
            sb.append(dst.id).append(';');
            
            sb.append("(").append(dst.getSuccessors().get(0).dst.id);
            for (int i = 1; i < dst.getSuccessors().size(); i++) {
                sb.append(';');
                sb.append(dst.getSuccessors().get(i).dst.id);
            }
            sb.append(")").append(';');
            
            sb.append((int) (dst.getSuccessors().stream().mapToDouble(e -> e.getWidth()).max().getAsDouble() * 1E06)).append(';');
            
            int length = (int) (dst.getSuccessors().stream().mapToDouble(e -> e.getLength()).sum() * 1E06);
            if(length % 50 != 0) length += length % 50;
            sb.append(length);
            
            sb.append('\n');
        }
        
        sb.append("channel").append(';');
        sb.append(name).append(';');
        sb.append(src.id).append(';');
        sb.append(dst.id).append(';');
        sb.append((int) (width * 1E06));
        if(isFullySpecified()) {
            sb.append(';').append((int) (length * 1E06));
        }        
        sb.append('\n');
        
        return sb.toString();
    }
    
    @Override
    public int getTypeId() {

        return 3;
    }

    @Override
    JSONObject toJSON() {
        JSONObject json = new JSONObject();
        json.put("type", this.getClass().getSimpleName());
        json.put("name", name);
        json.put("srcId", src.id);
        json.put("dstId", dst.id);
        json.put("height", height);
        json.put("width", width);
        if(!Double.isNaN(length)){
            json.put("length", length);
        }

        return json;
    }
}
