/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.structure;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.json.JSONArray;
import org.json.JSONObject;
import at.jku.iic.microfluidics.droplet.basic.architecture.Biochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;

public class StructureBiochip extends Biochip<StructureEdge, StructurePump> {

    public StructureBiochip(List<StructureEdge> edges, List<StructurePump> pump) {
        super(edges, pump);
    }

    public JSONObject toJSON(){
        JSONObject json = new JSONObject();

        //endPoints
        //all relevant endPoints are stored during the loop through the pumps and edges
        Set<EndPoint<StructureEdge>> endPointIds = new HashSet<>();

        //pumps
        JSONArray jsonPumps = new JSONArray();
        json.put("pumps", jsonPumps);
        for (StructurePump pump : getPumps()) {
            endPointIds.add(pump.src);
            endPointIds.add(pump.sink);
            jsonPumps.put(pump.toJSON());
        }

        //edges
        JSONArray jsonEdges = new JSONArray();
        json.put("edges", jsonEdges);
        for (StructureEdge edge : getEdges()) {
            endPointIds.add(edge.src);
            endPointIds.add(edge.dst);
            jsonEdges.put(edge.toJSON());
        }

        //write stored endPoints
        JSONArray jsonEndPoints = new JSONArray();
        json.put("endPoints", jsonEndPoints);
        for (EndPoint<StructureEdge> endPointId : endPointIds) {
            jsonEndPoints.put(endPointId.toJSON());
        }

        return json;
    }

    public static StructureBiochip fromJSON(JSONObject json) {
        //endPoints
        JSONArray jsonEndPoints = json.getJSONArray("endPoints");
        Map<Integer, EndPoint<StructureEdge>> endPoints = new HashMap<>();
        for (int i = 0; i < jsonEndPoints.length(); i++) {
            JSONObject jsonEndPoint = jsonEndPoints.getJSONObject(i);
            int id = jsonEndPoint.getInt("id");
            boolean isLeeSorter = jsonEndPoint.getBoolean("isLeeSorter");
            double leeSortingLength = jsonEndPoint.getDouble("leeSortingLength");
            double leeSortingWidth = jsonEndPoint.getDouble("leeSortingWidth");

            EndPoint<StructureEdge> endPoint;
            if (isLeeSorter) {
                endPoint = new EndPoint<>(id, leeSortingLength, leeSortingWidth);
            } else{
                endPoint = new EndPoint<>(id);
            }
            
            endPoints.put(id, endPoint);
        }

        //pumps
        JSONArray jsonPumps = json.getJSONArray("pumps");
        List<StructurePump> pumps = new ArrayList<>();
        for (int i = 0; i < jsonPumps.length(); i++) {
            JSONObject jsonPump = jsonPumps.getJSONObject(i);
            // String type = jsonPump.getString("type");
            String name = jsonPump.getString("name");
            int srcId = jsonPump.getInt("srcId");
            int dstId = jsonPump.getInt("dstId");

            pumps.add(new StructurePump(endPoints.get(srcId), endPoints.get(dstId), name));
        }

        //edges
        JSONArray jsonEdges = json.getJSONArray("edges");
        List<StructureEdge> edges = new ArrayList<>();
        for (int i = 0; i < jsonEdges.length(); i++) {
            JSONObject jsonEdge = jsonEdges.getJSONObject(i);
            String type = jsonEdge.getString("type");
            String name = jsonEdge.getString("name");
            int srcId = jsonEdge.getInt("srcId");
            int dstId = jsonEdge.getInt("dstId");

            if(type.equals(StructureChannel.class.getSimpleName())) {
                edges.add(new StructureChannel(endPoints.get(srcId), endPoints.get(dstId), name));
            } else if(type.equals(StructureBypassChannel.class.getSimpleName())) {
                edges.add(new StructureBypassChannel(endPoints.get(srcId), endPoints.get(dstId), name));
            } else if (type.equals(StructureModule.class.getSimpleName())) {
                char operation = jsonEdge.getString("operation").charAt(0);
                edges.add(new StructureModule(endPoints.get(srcId), endPoints.get(dstId), name, operation));
            }
        }

        return new StructureBiochip(edges, pumps);
    }
}
