/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.architecture.utility;

import at.jku.iic.microfluidics.droplet.basic.architecture.Biochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.Edge;
import at.jku.iic.microfluidics.droplet.basic.architecture.Pump;

public class ComponentByNameSearcher {

    /**
     * Searches for an <code>Edge</code> by its name.<br>
     * This method throws an <code>RuntimeException<code> in case the biochip does not contain an edge with
     * the specified name or in case the biochip contains multiple edges with the specified name.
     * 
     * @param biochip
     * @param name
     * @return
     */
    public static <K extends Edge<K>, V extends Pump<K>> K getEdgeByName(Biochip<K, V> biochip, String name) {

        K res = null;

        for (K e : biochip) {
            if (e.name.equals(name)) {
                if (res != null) {
                    throw new RuntimeException("Ambiguity: Multiple edges with specired name found.");
                }
                res = e;
            }
        }

        if (res == null) {
            throw new RuntimeException("No edge with specified name found.");
        }
        return res;
    }

    /**
     * Searches for a <code>Pump</code> by its name.<br>
     * This method throws an <code>RuntimeException<code> in case the biochip does not contain a pump with the
     * specified name or in case the biochip contains multiple pumps with the specified name.
     * 
     * @param biochip
     * @param name
     * @return
     */
    public static <K extends Edge<K>, V extends Pump<K>> V getPumpByName(Biochip<K, V> biochip, String name) {

        V res = null;

        for (V pump : biochip.getPumps()) {
            if (pump.name.equals(name)) {
                if (res != null) {
                    throw new RuntimeException("Ambiguity: Multiple edges with specired name found.");
                }
                res = pump;
            }
        }

        if (res == null) {
            throw new RuntimeException("No edge with specified name found.");
        }
        return res;
    }
    
    private ComponentByNameSearcher() {
        
    }
}
