/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.calculator;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jgrapht.Graph;
import org.jgrapht.alg.ConnectivityInspector;
import org.jgrapht.alg.cycle.PatonCycleBase;
import org.jgrapht.graph.SimpleGraph;

import Jama.Matrix;
import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.CloggablePhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.NotFullySpecifiedPhysicalBiochipException;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalPump;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PressurePump;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.VolumetricFlowRatePump;

/**
 * Calculator for the flow rates. Therefore, the Kirchoff's equation system is determined using the Matlab
 * functions. The flow rates are calculated using Jama.
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 * @author Gerold Fink
 *
 */
public class FlowCalculator {

    private final PhysicalBiochip chip;

    private Matrix a_nodes;
    private Matrix a_cycles_index;
    private Matrix b;
    private Graph<EndPoint<PhysicalEdge>, Object> graph;
    private Map<Object, Integer> indexMap;

    private Set<CloggablePhysicalChannel> cloggedChannels;

    private final boolean perfectClogging;
    private final int CLOGGING_RESISTANCE_INCREASE_FACTOR = 25;

    public FlowCalculator(PhysicalBiochip chip, boolean perfectClogging) {

        this.chip = chip;
        this.perfectClogging = perfectClogging;

        cloggedChannels = Collections.emptySet();
        createGraph();
        determineEquationSystem();
    }

    private void createGraph() {

        // Create Graph
        graph = new SimpleGraph<>(Object.class);

        // Add PhysicalEdges to graph
        for (PhysicalEdge edge : chip.getEdges()) {
            // only add vertices(edges) if they are not clogged or perfectClogging is false
            if (!cloggedChannels.contains(edge) || !perfectClogging) {
                graph.addVertex(edge.src);
                graph.addVertex(edge.dst);
                graph.addEdge(edge.src, edge.dst, edge);
            }
        }

        // Add PhysicalPumps to graph
        for (PhysicalPump pump : chip.getPumps()) {
            graph.addVertex(pump.src);
            graph.addVertex(pump.sink);
            graph.addEdge(pump.src, pump.sink, pump);
        }

        // Search and remove not connected vertices(edges)
        // Create ConnectivityInspector
        ConnectivityInspector<EndPoint<PhysicalEdge>, Object> connections = new ConnectivityInspector<>(graph);

        // Get all vertices which have a connection to the input channel
        Set<EndPoint<PhysicalEdge>> notConnectedVertices = new HashSet<>(graph.vertexSet());

        for (PhysicalPump p : chip.getPumps()) {
            Set<EndPoint<PhysicalEdge>> connectedVertices = connections.connectedSetOf(p.src);
            notConnectedVertices.removeAll(connectedVertices);
        }
        // Remove not connected vertices from graph
        graph.removeAllVertices(notConnectedVertices);
    }

    /**
     * Updates the equation system using the edges and pumps from the updated <code>PhysicalBiochip</code>.
     * 
     * @param cloggedChannels
     */
    public final void updateEquationSystem(Set<CloggablePhysicalChannel> cloggedChannels) {

        if (!chip.isFullySpecified()) {
            throw new NotFullySpecifiedPhysicalBiochipException();
        }

        this.cloggedChannels = cloggedChannels;
        if (perfectClogging) {
            createGraph();
            determineEquationSystem();
        }
    }

    private void determineEquationSystem() {

        int amountNodes = graph.vertexSet().size();
        int amountEdges = graph.edgeSet().size();
        int amountCycles = amountEdges - (amountNodes - 1);
        b = new Matrix(amountNodes - 1 + amountCycles, 1);

        // HashMap
        indexMap = new IdentityHashMap<>();
        int i = 0;
        for (Object edge : graph.edgeSet()) {
            indexMap.put(edge, i++);
        }

        // ### Kirchhoff current law ###
        a_nodes = new Matrix(amountNodes - 1, amountEdges);
        int row = 0;
        // Loop over all vertices except one => only amountNodes-1 independent equations
        for (EndPoint<PhysicalEdge> vertex : graph.vertexSet()) {
            if (row == amountNodes - 1) break;
            // Loop over all connected edges of the vertex
            for (Object edge : graph.edgesOf(vertex)) {
                // Check if edge is a PhysicalEdge => adapt matrix "a_nodes"
                if (edge instanceof PhysicalEdge) {
                    // Check if the edge enters or leaves the vertex
                    if (((PhysicalEdge) edge).src == vertex) {
                        a_nodes.set(row, indexMap.get(edge), -1.0); // leaves the vertex
                    } else {
                        a_nodes.set(row, indexMap.get(edge), 1.0); // enters the vertex
                    }
                }
                // Check if edge is a PressurePump => adapt matrix "a_nodes"
                else if (edge instanceof PressurePump) {
                    // Check if the edge enters or leaves the vertex
                    if (((PressurePump) edge).src == vertex) {
                        a_nodes.set(row, indexMap.get(edge), 1.0); // leaves the vertex
                    } else {
                        a_nodes.set(row, indexMap.get(edge), -1.0); // enters the vertex
                    }
                }
                // Check if edge is a VolumetricFlowRatePump => adapt matrix "b"
                else if (edge instanceof VolumetricFlowRatePump) {
                    // Check if the edge enters or leaves the vertex
                    if (((VolumetricFlowRatePump) edge).src == vertex) {
                        b.set(row, 0, b.get(row, 0) - ((VolumetricFlowRatePump) edge).inputFlowRate); // leaves
                                                                                                      // the
                                                                                                      // vertex
                    } else {
                        b.set(row, 0, b.get(row, 0) + ((VolumetricFlowRatePump) edge).inputFlowRate); // enters
                                                                                                      // the
                                                                                                      // vertex
                    }
                }
            }
            row++;
        }

        // ### Kirchhoff voltage law ###
        a_cycles_index = new Matrix(amountCycles, amountEdges);
        row = 0;
        // Find cycles of the network
        PatonCycleBase<EndPoint<PhysicalEdge>, Object> patonCylceBase = new PatonCycleBase<>(graph);
        List<List<EndPoint<PhysicalEdge>>> cycles = patonCylceBase.findCycleBase();

        // Loop over the cycles
        for (List<EndPoint<PhysicalEdge>> cycle : cycles) {
            // Loop over all vertices of the cycle
            for (i = 0; i < cycle.size(); i++) {
                EndPoint<PhysicalEdge> src = cycle.get(i); // Source vertex
                EndPoint<PhysicalEdge> dst; // Destination vertex
                // Check if the end of the loop is reached
                if (i == cycle.size() - 1) {
                    dst = cycle.get(0); // End is reached => destination vertex is first vertex of the cycle
                } else {
                    dst = cycle.get(i + 1);
                }

                // Find edge
                Object edge = graph.getEdge(src, dst);

                // Check if edge is a PhysicalEdge => adapt matrix "a_nodes"
                if (edge instanceof PhysicalEdge) {
                    // Check if the edge enters or leaves the vertex
                    if (((PhysicalEdge) edge).src == src) {
                        a_cycles_index.set(row, indexMap.get(edge), 1.0); // leaves the src vertex
                    } else {
                        a_cycles_index.set(row, indexMap.get(edge), -1.0); // enters the src vertex
                    }
                }
                // Check if edge is a VolumetricFlowRatePump => adapt matrix "a_nodes"
                else if (edge instanceof VolumetricFlowRatePump) {
                    // Check if the edge enters or leaves the vertex
                    if (((VolumetricFlowRatePump) edge).src == src) {
                        a_cycles_index.set(row, indexMap.get(edge), 1.0); // leaves the src vertex
                    } else {
                        a_cycles_index.set(row, indexMap.get(edge), -1.0); // enters the src vertex
                    }
                }
                // Check if edge is a PressurePump => adapt matrix "b"
                else if (edge instanceof PressurePump) {
                    // Check if the edge enters or leaves the vertex
                    if (((PressurePump) edge).src == src) {
                        b.set(row + amountNodes - 1, 0, b.get(row + amountNodes - 1, 0) - ((PressurePump) edge).inputPressure); // leaves
                                                                                                                                // the
                                                                                                                                // src
                                                                                                                                // vertex
                    } else {
                        b.set(row + amountNodes - 1, 0, b.get(row + amountNodes - 1, 0) + ((PressurePump) edge).inputPressure); // enters
                                                                                                                                // the
                                                                                                                                // src
                                                                                                                                // vertex
                    }
                }
            }
            row++;
        }
    }

    /**
     * Calculates the flow rates based on the given droplet states.
     * 
     * @param dropletStates
     * @return
     */
    public final Map<PhysicalEdge, Double> calculateFlowRates(Map<PhysicalEdge, Double> dropletResistances) {

        // Solve equation system
        Matrix x = solve(dropletResistances);

        // Build up map
        Map<PhysicalEdge, Double> flowRateMap = new HashMap<>();
        Set<Object> graphEdges = graph.edgeSet();
        for (PhysicalEdge edge : chip.getEdges()) {
            if (graphEdges.contains(edge)) {
                flowRateMap.put(edge, x.get(indexMap.get(edge), 0));
            } else {
                flowRateMap.put(edge, 0.0);
            }
        }
        return flowRateMap;
    }
    
    public final Matrix solve(Map<PhysicalEdge, Double> dropletResistances) {
    	return solve(dropletResistances, 1);
    }
    
    /**
     * Solves the network and returns flow rates as matrix
     * @param dropletResistances
     * @param resistanceFactor apply scaling factor to resistances (1 = unchanged)
     * @return nx1 vector with flow rates for each edge
     */
    public final Matrix solve(Map<PhysicalEdge, Double> dropletResistances, double resistanceFactor) {

        Set<CloggablePhysicalChannel> imperfectCloggedChannels = perfectClogging ? Collections.emptySet() : cloggedChannels;

        // Create matrix "nets", which contains the resistances of all PhysicalEdges
        Matrix nets = new Matrix(indexMap.size(), indexMap.size());
        // Loop over all edges in the graph
        for (Object edge : graph.edgeSet()) {
            int index = indexMap.get(edge);
            // Check if edge is a PhysicalEdge
            if (edge instanceof PhysicalEdge) {
                double resistance;
                if (imperfectCloggedChannels.contains(edge)) {
                    resistance = ((PhysicalEdge) edge).getResistance(chip.fluidProperties.muC) * resistanceFactor * CLOGGING_RESISTANCE_INCREASE_FACTOR;
                } else {
                    resistance = ((PhysicalEdge) edge).getResistance(chip.fluidProperties.muC) * resistanceFactor;
                }
                nets.set(index, index, resistance);
            }
            // If edge is a PhysicalPump write 1.0 in the diagonal element
            else {
                nets.set(index, index, 1.0);
            }
        }

        // Add resistances caused by droplets
        for (Map.Entry<PhysicalEdge, Double> dropRes : dropletResistances.entrySet()) {
            int index = indexMap.get(dropRes.getKey());
            nets.set(index, index, nets.get(index, index) + dropRes.getValue());
        }

        // Calculate matrix of the cycles
        Matrix a_cycles = a_cycles_index.times(nets);

        // Build whole matrix "a" with "a_nodes" and "a_cycles"
        Matrix a = new Matrix(indexMap.size(), indexMap.size());
        a.setMatrix(0, a_nodes.getRowDimension() - 1, 0, indexMap.size() - 1, a_nodes);
        a.setMatrix(a_nodes.getRowDimension(), indexMap.size() - 1, 0, indexMap.size() - 1, a_cycles);

        // Solve equation system
        Matrix x = a.solve(b);

        return x;
    }
    
    public Matrix getA() {
    	return getA(Collections.emptyMap());
    }
    
    public Matrix getA(Map<PhysicalEdge, Double> dropletResistances) {
    	return getA(dropletResistances, 1);
    }
    
    public Matrix getA(Map<PhysicalEdge, Double> dropletResistances, double resistanceFactor) {
    	Set<CloggablePhysicalChannel> imperfectCloggedChannels = perfectClogging ? Collections.emptySet() : cloggedChannels;

        // Create matrix "nets", which contains the resistances of all PhysicalEdges
        Matrix nets = new Matrix(indexMap.size(), indexMap.size());
        // Loop over all edges in the graph
        for (Object edge : graph.edgeSet()) {
            int index = indexMap.get(edge);
            // Check if edge is a PhysicalEdge
            if (edge instanceof PhysicalEdge) {
                double resistance;
                if (imperfectCloggedChannels.contains(edge)) {
                    resistance = ((PhysicalEdge) edge).getResistance(chip.fluidProperties.muC) * resistanceFactor * CLOGGING_RESISTANCE_INCREASE_FACTOR;
                } else {
                    resistance = ((PhysicalEdge) edge).getResistance(chip.fluidProperties.muC) * resistanceFactor;
                    if(Double.isNaN(resistance)) {
                    	resistance = Double.POSITIVE_INFINITY;
                    }
                }
                
                nets.set(index, index, resistance);
            }
            // If edge is a PhysicalPump write 1.0 in the diagonal element
            else {
                nets.set(index, index, 1.0);
            }
        }

        // Add resistances caused by droplets
        for (Map.Entry<PhysicalEdge, Double> dropRes : dropletResistances.entrySet()) {
            int index = indexMap.get(dropRes.getKey());
            nets.set(index, index, nets.get(index, index) + dropRes.getValue());
        }

        // Calculate matrix of the cycles
        Matrix a_cycles = a_cycles_index.times(nets);
        
    	Matrix a = new Matrix(indexMap.size(), indexMap.size());
        a.setMatrix(0, a_nodes.getRowDimension() - 1, 0, indexMap.size() - 1, a_nodes);
        a.setMatrix(a_nodes.getRowDimension(), indexMap.size() - 1, 0, indexMap.size() - 1, a_cycles);
        for(int i = 0; i < a.getRowDimension(); i++) {
        	for(int j = 0; j < a.getColumnDimension(); j++) {
        		if(Double.isNaN(a.get(i,j))) {
        			a.set(i, j, 0);
        		}
        	}
        	
        }
        return a;
    }
    
    public Matrix getB() {
    	return (Matrix) b.clone();
    }

	public Map<Object, Integer> getIndexMap() {
		return Collections.unmodifiableMap(indexMap);
	}
	
	public Graph<EndPoint<PhysicalEdge>, Object> getGraph() {
		return graph;
	}
}
