/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.calculator;

import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.CloggablePhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PressurePump;

public class PressureCalculator {

    public static final Map<EndPoint<PhysicalEdge>, Double> calculatePressureGradients(Map<PhysicalEdge, Double> flowRates,
            Set<CloggablePhysicalChannel> cloggedChannels, EndPoint<PhysicalEdge> referencePoint, PhysicalBiochip chip) {

        Map<EndPoint<PhysicalEdge>, Double> pressureGradients = new IdentityHashMap<>();
        pressureGradients.put(referencePoint, 0.0);

        Queue<EndPoint<PhysicalEdge>> queue = new LinkedList<>();
        queue.add(referencePoint);

        while (!queue.isEmpty()) {

            EndPoint<PhysicalEdge> cur = queue.remove();
            double dstPressure = pressureGradients.get(cur);

            for (PhysicalEdge pred : cur.getPredecessors()) {
                if (!cloggedChannels.contains(pred)) {
                    if (!pressureGradients.containsKey(pred.src)) {
                        queue.add(pred.src);
                        pressureGradients.put(pred.src, dstPressure + flowRates.get(pred) * pred.getResistance(chip.fluidProperties.muC));
                    }
                }
            }
            for (PhysicalEdge succ : cur.getSuccessors()) {
                if (!cloggedChannels.contains(succ)) {
                    if (!pressureGradients.containsKey(succ.dst)) {
                        queue.add(succ.dst);
                        pressureGradients.put(succ.dst, dstPressure - flowRates.get(succ) * succ.getResistance(chip.fluidProperties.muC));
                    }
                }
            }
            for (PressurePump p : chip.getPressurePumps()) {
                if (p.sink == cur) {
                    if (!pressureGradients.containsKey(p.src)) {
                        queue.add(p.src);
                        pressureGradients.put(p.src, dstPressure + p.inputPressure);
                    }
                }
                if (p.src == cur) {
                    if (!pressureGradients.containsKey(p.sink)) {
                        queue.add(p.sink);
                        pressureGradients.put(p.sink, dstPressure - p.inputPressure);
                    }
                }
            }

        }

        return pressureGradients;
    }

    private PressureCalculator() {

    }
}
