/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.dumper;

import java.awt.Color;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import at.jku.iic.microfluidics.droplet.basic.architecture.Biochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.Edge;
import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.architecture.Pump;
import at.jku.iic.microfluidics.utils.dumper.Dumper;
import at.jku.iic.microfluidics.utils.dumper.GraphDumper;

public abstract class BiochipDumper<K extends Edge<K>, V extends Pump<K>> extends GraphDumper {

    protected final Biochip<K, V> chip;

    protected BiochipDumper(PrintWriter writer, Dumper dumper, Biochip<K, V> chip) {

        super(writer, dumper);
        this.chip = chip;
    }

    public abstract String getEdgeLabel(K edge);

    @Override
    public void dump() {

        assert writer != null;

        // first, dump the header to the stream
        dumper.generateHeader(writer);

        // now, generate unique IDs for every node
        generateNodeNames();

        printSuccessors();

        // write trailer to stream
        dumper.generateTrailer(writer);

        // flush and delete output stream
        writer.flush();
    }

    private void printSuccessors() {

        for (K e : chip) {
            assert e.src != null && e.dst != null;
            String src = Integer.toString(e.src.id);
            String dst = Integer.toString(e.dst.id);

            if (e.isModule()) {
                dumper.generateEdge(writer, src, getEdgeLabel(e), dst, Color.RED, false);
            } else if (e.isBypassChannel()) {
                dumper.generateEdge(writer, src, getEdgeLabel(e), dst, Color.GRAY, false);
            } else {
                dumper.generateEdge(writer, src, getEdgeLabel(e), dst, Color.GREEN, false);
            }
        }
    }

    /**
     * Generates node IDs for all nodes in the provided collection.
     * 
     * @param operations
     * @param maxInstances
     * 
     * @param nodes
     *            collection of nodes
     */
    private void generateNodeNames() {

        Set<Integer> nodeIds = new HashSet<>();

        for (V p : chip.getPumps()) {

            for (EndPoint<K> eP : Arrays.asList(p.sink, p.src)) {
                if (!nodeIds.contains(eP.id)) {
                    nodeIds.add(eP.id);
                    dumper.generateNode(writer, Integer.toString(eP.id), Integer.toString(eP.id), Color.GREEN);
                }
            }
        }

        for (K e : chip) {

            for (EndPoint<K> eP : Arrays.asList(e.src, e.dst)) {
                if (!nodeIds.contains(eP.id)) {
                    nodeIds.add(eP.id);
                    dumper.generateNode(writer, Integer.toString(eP.id), Integer.toString(eP.id), Color.ORANGE);
                }
            }
        }
    }

}
