/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.graph.utility;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import at.jku.iic.microfluidics.droplet.basic.architecture.Biochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.BypassChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.Edge;
import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.architecture.Path;
import at.jku.iic.microfluidics.droplet.basic.architecture.Pump;

public class GraphUtility {

    public static <K extends Edge<K>, V extends Pump<K>> Map<V, List<Path<K>>> findAllDropletPaths(Biochip<K, V> biochip) {

        // A map from a pump to a list of paths
        Map<V, List<Path<K>>> paths = new HashMap<>();

        for (V pump : biochip.getPumps()) {
            List<Path<K>> pathsOfPump = new LinkedList<>();
            List<K> tmp = new LinkedList<>();
            assert pump.src.getSuccessors().size() == 1;
            findAllPathsAt(pump, pump.src.getSuccessors().get(0), pathsOfPump, tmp);
            paths.put(pump, pathsOfPump);
        }
        return paths;
    }

    private static <K extends Edge<K>, V extends Pump<K>> void findAllPathsAt(V pump, K id, List<Path<K>> allPaths, List<K> tmp) {

        tmp.add(id);
        if (id.dst == pump.sink) {
            allPaths.add(new Path<K>(tmp));
            return;
        }

        for (K succ : id.dst.getSuccessors()) {
            if (!(succ instanceof BypassChannel)) {
                List<K> tmp2 = new LinkedList<>(tmp);
                findAllPathsAt(pump, succ, allPaths, tmp2);
            }
        }
    }

    /**
     * Determines all paths form all pumpts to the specified destination edge.
     * @param biochip
     * @param dstEdge The destination edge.
     * @return
     */
    public static <K extends Edge<K>, V extends Pump<K>> Map<V, List<Path<K>>> determineHeaderPathsToEdge(Biochip<K, V> biochip, K dstEdge) {

        Map<V, List<Path<K>>> paths = new HashMap<>();

        for (V pump : biochip.getPumps()) {
            List<Path<K>> pathsOfPump = new LinkedList<>();
            List<K> tmp = new LinkedList<>();
            assert pump.src.getSuccessors().size() == 1;
            findAllHeaderPaths(pump.src.getSuccessors().get(0), pathsOfPump, tmp, dstEdge.dst);
            paths.put(pump, pathsOfPump);
        }
        return paths;
    }

    private static <K extends Edge<K>> void findAllHeaderPaths(K id, List<Path<K>> allPaths, List<K> tmp,
            EndPoint<K> dst) {

        tmp.add(id);
        if (id.dst == dst) {
            allPaths.add(new Path<K>(tmp));
            return;
        }

        if (id.getHeaderSuccessorEdges().size() == 1) {
            // There is no choice where the header goes. Therefore, use the same List again.
            findAllHeaderPaths(id.getHeaderSuccessorEdges().get(0), allPaths, tmp, dst);
        } else {
            for (K succ : id.getHeaderSuccessorEdges()) {
                List<K> tmp2 = new LinkedList<>(tmp);
                findAllHeaderPaths(succ, allPaths, tmp2, dst);
            }
        }
    }
    

    private GraphUtility() {

    }
}
