/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.injection.converter;

import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;

import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalPump;
import at.jku.iic.microfluidics.droplet.basic.injection.discrete.DiscreteDropletInjectionSequence;
import at.jku.iic.microfluidics.droplet.basic.injection.discrete.DiscreteDropletInjectionTime;
import at.jku.iic.microfluidics.droplet.basic.injection.discrete.DiscreteHeaderInjectionTime;
import at.jku.iic.microfluidics.droplet.basic.injection.discrete.DiscretePayloadInjectionTime;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalDropletInjectionSequence;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalDropletInjectionTime;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalHeaderInjectionTime;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalPayloadInjectionTime;

/**
 * Converts a <code>DiscreteDropletInjectionSequence</code> to a <code>PhysicalDropletInjectionSequence</code>
 * .
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public class DiscreteToPhysicalInjectionSequenceConverter {

    private final PhysicalBiochip physicalBiochip;
    private final double realTimeOfTimeStep;

    private final Map<Integer, PhysicalPump> pumpMapping;

    public DiscreteToPhysicalInjectionSequenceConverter(PhysicalBiochip physicalBiochip, double realTimeOfTimeStep) {

        this.physicalBiochip = physicalBiochip;
        this.realTimeOfTimeStep = realTimeOfTimeStep;
        this.pumpMapping = physicalBiochip.getPumps().stream()
                .collect(Collectors.toMap(PhysicalPump::getEndPointHashs, Function.identity()));
    }

    public PhysicalDropletInjectionSequence convert(DiscreteDropletInjectionSequence sequence) {

        return new PhysicalDropletInjectionSequence(sequence.injections.stream().map(inj -> convert(inj)).collect(Collectors.toList()));
    }

    public PhysicalDropletInjectionTime convert(DiscreteDropletInjectionTime inj) {

        PhysicalPump physicalPump = pumpMapping.get(inj.pump.getEndPointHashs());
        if (inj instanceof DiscreteHeaderInjectionTime) {
            return new PhysicalHeaderInjectionTime(physicalPump, inj.timePoint * realTimeOfTimeStep, physicalBiochip.headerDropletVolume);
        } else {
            assert inj instanceof DiscretePayloadInjectionTime;
            return new PhysicalPayloadInjectionTime(physicalPump, inj.timePoint * realTimeOfTimeStep, physicalBiochip.payloadDropletVolume);
        }
    }

}
