/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.injection.physical;

import java.util.LinkedList;
import java.util.List;

import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalPump;

/**
 * Utility methods for generating droplet injection sequences.
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public class PhysicalDropletInjectionGenerator {

    public static List<PhysicalDropletInjectionTime> getPayloadDropletInjections(PhysicalBiochip chip, PhysicalPump pump,
            double continuousFlowRate, double dispersedFlowRate, double xi, double timeFrame) {

        List<PhysicalDropletInjectionTime> dropletInjections = new LinkedList<>();

        PhysicalEdge cIn = pump.src.getSuccessors().get(0);

        // See garstecki2006formation for the derivation
        double ld = cIn.getWidth() * (1 + xi * dispersedFlowRate / continuousFlowRate);
        // See biral2015modeling for the derivation
        double interDropletDistance = ld * continuousFlowRate / dispersedFlowRate;

        double dropletLength = chip.payloadDropletVolume / (cIn.getHeight() * cIn.getWidth());

        double inputVelocity = (continuousFlowRate + dispersedFlowRate) / (cIn.getHeight() * cIn.getWidth());
        double timeBetweenDrops = (interDropletDistance + dropletLength) / inputVelocity;

        for (double acc = 0.0; acc < timeFrame; acc += timeBetweenDrops) {
            dropletInjections.add(new PhysicalPayloadInjectionTime(pump, acc, chip.payloadDropletVolume));
        }

        return dropletInjections;
    }
    
    
    public static List<PhysicalDropletInjectionTime> getPayloadDropletInjections(PhysicalBiochip chip, PhysicalPump pump, double inputVelocity,
            double dropletSpacing, double timeFrame) {
        
        List<PhysicalDropletInjectionTime> dropletInjections = new LinkedList<>();
        
        double timeBetweenDrops = dropletSpacing / inputVelocity;

        for (double acc = 0.0; acc < timeFrame; acc += timeBetweenDrops) {
            dropletInjections.add(new PhysicalPayloadInjectionTime(pump, acc, chip.payloadDropletVolume));
        }

        return dropletInjections;
    }
    

    private PhysicalDropletInjectionGenerator() {

    }

}
