/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.physics;

/**
 * Properties for the Validation and Automated Dimensioning Methods
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public class DimensioningProperties {

    /**
     * The input flow rate applied by the pumps. When a <code>StructureBiochip</code> gets converted to a
     * <code>PhysicalBiochip</code> pumps producing a volumetric flow rate are applied.
     */
    public final double inputFlowRate;

    /**
     * The maximally allowed time a droplet can take to pass a channel/module
     */
    public final double maximalTime;

    /**
     * The minimally allowed flow rate through a channel/module
     */
    public final double minimalFlow;

    /**
     * The maximally allowed Capillary number
     */
    public final double maxAllowedCapillaryNumber;

    /**
     * The maximally allowed Reynolds number
     */
    public final double maxAllowedReynoldsNumber;

    public DimensioningProperties(double inputFlowRate, double maximalTime, double minimalFlow, double maxAllowedCapillaryNumber,
            double maxAllowedReynoldsNumber) {
        super();
        this.inputFlowRate = inputFlowRate;
        this.maximalTime = maximalTime;
        this.minimalFlow = minimalFlow;
        this.maxAllowedCapillaryNumber = maxAllowedCapillaryNumber;
        this.maxAllowedReynoldsNumber = maxAllowedReynoldsNumber;
    }

    private static final double MAXIMAL_TIME = 2;
    private static final double MINIMAL_FLOW = 3E-12;

    private static final double MAX_CAPILLARY_NUMBER = 0.01;
    private static final double MAX_REYNOLDS_NUMBER = 1;

    protected static final double INPUT_FLOW_FOR_ALL_SYNTHETIC_ARCHITECTURES = 50E-12;

    public static DimensioningProperties getDefaultSettings() {

        return new DimensioningProperties(INPUT_FLOW_FOR_ALL_SYNTHETIC_ARCHITECTURES, MAXIMAL_TIME, MINIMAL_FLOW, MAX_CAPILLARY_NUMBER,
                MAX_REYNOLDS_NUMBER);
    }

}
