/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.basic.physics;

import org.json.JSONObject;

public class FluidProperties {

    /**
     * Dynamic viscosity of the continuous phase
     */
    public final double muC;

    /**
     * Dynamic viscosity of the droplet phase
     */
    public final double muD;

    /**
     * The density of the continuous phase in kg/m^3
     */
    public final double densityC;

    /**
     * The interfacial tension (in Nm^-1)
     */
    public final double interfacialTension;


    /**
     * Slip factor of the droplets
     */
    public final double slipFactor;

    public FluidProperties(double muC, double muD, double densityC, double interfacialTension, double slipFactor) {
        super();
        this.muC = muC;
        this.muD = muD;
        this.densityC = densityC;
        this.interfacialTension = interfacialTension;
        this.slipFactor = slipFactor;
    }

    private static final double STD_MUC = 1e-03;
    private static final double STD_MUD = 1.5931e-03;
    private static final double STD_DENSITY_C = 997;
    private static final double STD_INTERFACIAL_TENSION = 0.012;
    
    private static final double STD_SLIP_FACTOR = 1.28;

    public static FluidProperties getOilInWaterDropletPropperties() {

        return new FluidProperties(STD_MUC, STD_MUD, STD_DENSITY_C, STD_INTERFACIAL_TENSION, STD_SLIP_FACTOR);
    }

    public static FluidProperties getWaterInOilDropletProperties() {

        // Continuous phase: Silicon Oil DC 200 with 5cSt (Kinematic viscosity) and density 0.913 g mL^-1
        // Dispersed phase: Water
        return new FluidProperties(4.565E-3, 1E-3, 913, 0.042, STD_SLIP_FACTOR);
    }

    public JSONObject toJSON() {
        JSONObject json = new JSONObject();
        
        json.put("muC", muC);
        json.put("muD", muD);
        json.put("densityC", densityC);
        json.put("interfacialTension", interfacialTension);
        json.put("slipFactor", slipFactor);

        return json;
    }

	public static FluidProperties fromJSON(JSONObject json) {
		double muC = json.getDouble("muC");
		double muD = json.getDouble("muD");
		double densityC = json.getDouble("densityC");
		double interfacialTension = json.getDouble("interfacialTension");
        double slipFactor = json.getDouble("slipFactor");
        
        return new FluidProperties(muC, muD, densityC, interfacialTension, slipFactor);
	}
}
