/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
// import java.util.logging.Logger;
import java.util.stream.Collectors;

import com.google.common.collect.Sets;

import at.jku.iic.microfluidics.droplet.basic.architecture.physical.CloggablePhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.TrappablePhysicalModule;
import at.jku.iic.microfluidics.droplet.basic.calculator.FlowCalculator;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalDropletInjectionSequence;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalDropletInjectionTime;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalHeaderInjectionTime;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalPayloadInjectionTime;
import at.jku.iic.microfluidics.droplet.electric.simulator.event.Event;
import at.jku.iic.microfluidics.droplet.electric.simulator.event.NewDropletInjection;
import at.jku.iic.microfluidics.droplet.electric.simulator.event.NewDropletPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.event.TimeEvent;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletEntireChannelPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletHeadPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletInjectionPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletRangePosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletTailPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletTrappedPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.PhysicalDropletState;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.PhysicalSystemState;

public class PhysicalSimulator {

    // private static final Logger LOGGER = Logger.getLogger(PhysicalSimulator.class.getSimpleName());

    /**
     * The physical biochip on which the droplet sequence is executed.
     */
    private final PhysicalBiochip chip;

    /**
     * A list which is sorted by the injection time
     */
    private final List<PhysicalDropletInjectionTime> dropletInjections;

    /**
     * Counts how many headers are injected
     */
    private int headerCounter;

    /**
     * Counts how many payloads are injected
     */
    private int payloadCounter;

    public PhysicalSimulator(PhysicalBiochip chip, PhysicalDropletInjectionSequence dropletInjectionSeq) {

        this.chip = chip;
        this.dropletInjections = dropletInjectionSeq.injections;
        Collections.sort(this.dropletInjections);
    }

    public List<PhysicalSystemState> simulate(boolean perfectClogging) {

        return simulate(Optional.empty(), perfectClogging);
    }

    public List<PhysicalSystemState> simulate(double timeStepInterval, boolean perfectClogging) {

        return simulate(Optional.of(timeStepInterval), perfectClogging);
    }

    private List<PhysicalSystemState> simulate(Optional<Double> timeStepInterval, boolean perfectClogging) {

        double time = 0;

        List<PhysicalSystemState> systemStates = new LinkedList<>();

        FlowCalculator flowCalculator = new FlowCalculator(chip, perfectClogging);
        Map<PhysicalEdge, Double> flowRates = flowCalculator.calculateFlowRates(Collections.emptyMap());
        PhysicalSystemState current = new PhysicalSystemState(time, Collections.emptyList(), flowRates);
        Set<CloggablePhysicalChannel> lastCloggedChannels = Collections.emptySet();

        double nextTimeStep = timeStepInterval.isPresent() ? timeStepInterval.get() : 0.0;

        // while (true) {
        while (!Thread.currentThread().isInterrupted()) {

            List<Event> events = new LinkedList<>();
            Optional<NewDropletPosition> nextDropPos = current.getNextTriggeringPosition(flowRates, chip.fluidProperties.slipFactor);
            if (nextDropPos.isPresent()) {
                events.add(nextDropPos.get());
            }
            if (!allDropletsInjected()) {
                events.add(new NewDropletInjection(dropletInjections.get(headerCounter + payloadCounter)));
            }
            if (timeStepInterval.isPresent() && events.size() > 0) {
                events.add(new TimeEvent(nextTimeStep));
            }

            Optional<Event> nextEvent = events.stream().min(Event::compareTo);
            if (nextEvent.isPresent()) {
                if (nextEvent.get() instanceof TimeEvent) {
                    nextTimeStep += timeStepInterval.get();
                }

                List<PhysicalDropletState> newDropletStates = moveToNextDropletStates(current, nextEvent.get(), flowRates);

                Set<CloggablePhysicalChannel> cloggedChannels = newDropletStates.stream().map(r -> r.getCloggedChannels())
                        .flatMap(Set::stream).collect(Collectors.toSet());
                if (!cloggedChannels.equals(lastCloggedChannels)) {
                    // The architecture has changed
                    flowCalculator.updateEquationSystem(cloggedChannels);
                    lastCloggedChannels = cloggedChannels;
                }

                flowRates = flowCalculator.calculateFlowRates(Helpers.getDropletResistances(newDropletStates, chip.fluidProperties.muC));
                current = new PhysicalSystemState(nextEvent.get().getNextEventTime(), newDropletStates, flowRates);
                systemStates.add(current);

                TrappablePhysicalModule laplacePressureEx = current.getLaplacePressureEvent(chip);
                if (laplacePressureEx != null) {
                    System.out.println("The droplet was squeezed out of " + laplacePressureEx.name);
                    break;
                }
            } else {
                break;
            }
        }

        // add the first state when systemStates is empty (useful when no droplets are injected but only the flow rates of the networks is important)
        if (systemStates.isEmpty()) {
            systemStates.add(current);
        }

        return systemStates;
    }

    private List<PhysicalDropletState> moveToNextDropletStates(PhysicalSystemState current, Event event,
            Map<PhysicalEdge, Double> flowRates) {

        double eventTime = event.getNextEventTime();
        List<PhysicalDropletState> newDropletStates = new LinkedList<>();

        if (event instanceof NewDropletInjection) {
            NewDropletInjection ndi = (NewDropletInjection) event;
            PhysicalDropletInjectionTime inj = ndi.nextInjection;
            assert inj.pump.src.getSuccessors().size() == 1;
            PhysicalEdge cIn = inj.pump.src.getSuccessors().get(0);

            boolean defaultDirection = inj.pump.src == cIn.src && flowRates.get(cIn) >= 0.0
                    || inj.pump.src == cIn.dst && flowRates.get(cIn) <= 0.0;
            DropletInjectionPosition dropletInjectionPosition = new DropletInjectionPosition(cIn, defaultDirection, inj.dropletVolume, 0.0);
            LinkedList<DropletPosition> pos = new LinkedList<>();
            pos.add(dropletInjectionPosition);

            String name;
            if (inj instanceof PhysicalHeaderInjectionTime) {
                name = "h" + headerCounter;
                headerCounter++;
            } else {
                assert inj instanceof PhysicalPayloadInjectionTime;
                name = "p" + payloadCounter;
                payloadCounter++;
            }

            newDropletStates.add(new PhysicalDropletState(pos, name, Collections.emptySet(), inj));
        }

        double timeDifference = eventTime - current.time;

        assert timeDifference >= 0;

        // Update all positions
        for (PhysicalDropletState state : current.getDropletStates()) {
            LinkedList<DropletPosition> positions = new LinkedList<>();

            Set<CloggablePhysicalChannel> cloggedChannels = Sets.newIdentityHashSet();
            cloggedChannels.addAll(state.getCloggedChannels());

            if (state.getDropletPositions().size() > 0) {
                PhysicalEdge edgeOfTail = state.getDropletPositionOfTail().edge;

                double dropletVelocity;
                // List<PhysicalEdge> dirChanges;
                // if ((dirChanges = getEdgesWhereFlowDirectionChanged(state, flowRates)).size() > 0) {
                // for (PhysicalEdge dirChange : dirChanges) {
                // LOGGER.warning("The flow direction in " + dirChange.name + " where " + state.getName() + " is contained has changed. From now on, correct droplet volumen cannot be guaranteed!");
                // }
                if (getEdgesWhereFlowDirectionChanged(state, flowRates).size() > 0) {

                    double averageFlowRate = getAverageDropletFlowRate(state, flowRates);
                    if (hasFlowDirectionChanged(state, averageFlowRate)) {
                        // LOGGER.warning("The absolute direction of a droplet has changed.");
                        return null;
                    }
                    dropletVelocity = state.getDropletPositionOfTail().getAbsoluteDropletVelocity(averageFlowRate,
                            chip.fluidProperties.slipFactor);

                } else {
                    // Default case: No flow direction has changed.
                    dropletVelocity = state.getDropletPositionOfTail().getAbsoluteDropletVelocity(flowRates,
                            chip.fluidProperties.slipFactor);
                }

                double dropletVolumeShift = dropletVelocity * timeDifference * edgeOfTail.getHeight() * edgeOfTail.getWidth();
                for (DropletPosition pos : state.getDropletPositions()) {
                    positions.addAll(pos.moveDroplet(timeDifference, flowRates, cloggedChannels, event, dropletVolumeShift,
                            chip.fluidProperties.slipFactor));
                }

            }

            // Post-process the droplet states: Handle the case when the last part of the tail is now
            // contained
            // in the next channel.
            if (!positions.isEmpty()) {
                if (positions.getFirst() instanceof DropletHeadPosition) {
                    assert positions.size() == 1;
                    DropletHeadPosition first = (DropletHeadPosition) positions.removeFirst();

                    // Set the head of the droplet equal to the droplet length.
                    positions.add(new DropletRangePosition(first.edge, first.defaultFlowDirection, first.dropletVolume,
                            first.getFullyContainedDropletLength(), 0.0));
                } else if (positions.getFirst() instanceof DropletEntireChannelPosition) {
                    DropletPosition first = (DropletPosition) positions.removeFirst();
                    positions.addFirst(new DropletTailPosition(first.edge, first.defaultFlowDirection, first.dropletVolume, 0.0));
                } else if (positions.getFirst() instanceof DropletTrappedPosition) {
                    assert positions.size() == 1;
                    ((DropletTrappedPosition) positions.getFirst()).setDropletFullyContained(true);
                }
            }

            newDropletStates.add(new PhysicalDropletState(positions, state.getName(), cloggedChannels, state.getDropletInjectionTime()));
        }

        return newDropletStates;
    }

    private List<PhysicalEdge> getEdgesWhereFlowDirectionChanged(PhysicalDropletState state, Map<PhysicalEdge, Double> flowRates) {

        return state.getDropletPositions().stream().filter(e -> e.hasDropletFlowDirectionChanged(flowRates)).map(e -> e.edge)
                .collect(Collectors.toList());
    }

    private double getAverageDropletFlowRate(PhysicalDropletState state, Map<PhysicalEdge, Double> flowRates) {

        return state.getDropletPositions().stream().mapToDouble(s -> flowRates.get(s.edge)).average().getAsDouble();
    }

    private boolean hasFlowDirectionChanged(PhysicalDropletState state, double averageDropletFlowRate) {

        return state.getDropletPositions().stream().anyMatch(e -> e.hasDropletFlowDirectionChanged(averageDropletFlowRate));
    }

    /**
     * Checks whether already all droplets are injected.
     * 
     * @return
     */
    private boolean allDropletsInjected() {

        return headerCounter + payloadCounter == dropletInjections.size();
    }
}
