/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator;

import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

import at.jku.iic.microfluidics.droplet.basic.architecture.Path;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalDropletInjectionTime;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.PhysicalDropletState;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.PhysicalSystemState;

public class PhysicalSystemStatePostProcessor {

    private final List<PhysicalSystemState> states;

    public PhysicalSystemStatePostProcessor(List<PhysicalSystemState> states) {

        this.states = states;
    }

    public Path<PhysicalEdge> getPathOfDroplet(PhysicalDropletInjectionTime inj) {

        return getPathOfDroplet(new DropStateAccessor(inj));
    }

    public Path<PhysicalEdge> getPathOfDroplet(String dropletName) {

        return getPathOfDroplet(new DropStateAccessor(dropletName));
    }

    private Path<PhysicalEdge> getPathOfDroplet(DropStateAccessor t) {

        LinkedList<PhysicalEdge> path = new LinkedList<>();
        for (PhysicalSystemState state : states) {
            Optional<PhysicalDropletState> dropState = t.getDropState(state);
            if (dropState.isPresent() && dropState.get().getDropletPositions().size() > 0) {
                // The droplet with the specified name is in the system.
                PhysicalEdge edge = dropState.get().getDropletPositionOfHead().edge;
                if (path.size() == 0 || path.getLast() != edge) {
                    path.add(edge);
                }
            }
        }

        return new Path<>(path);
    }

    public List<Double> getEnteringTimes(PhysicalDropletInjectionTime inj) {

        return getEnteringTimes(new DropStateAccessor(inj));
    }

    public List<Double> getEnteringTimes(String dropletName) {

        return getEnteringTimes(new DropStateAccessor(dropletName));
    }

    private List<Double> getEnteringTimes(DropStateAccessor t) {

        PhysicalEdge lastChannel = null;
        LinkedList<Double> enteringTime = new LinkedList<>();
        for (PhysicalSystemState state : states) {
            Optional<PhysicalDropletState> dropState = t.getDropState(state);
            if (dropState.isPresent() && dropState.get().getDropletPositions().size() > 0) {
                PhysicalEdge edge = dropState.get().getDropletPositionOfHead().edge;
                if (lastChannel != edge) {
                    enteringTime.add(state.time);
                    lastChannel = edge;
                }
            }
        }

        return enteringTime;
    }
    
    public double getExitingTime(PhysicalDropletInjectionTime inj) {

        return getExitingTime(new DropStateAccessor(inj));
    }

    public double getExitingTime(String dropletName) {

        return getExitingTime(new DropStateAccessor(dropletName));
    }
    
    private double getExitingTime(DropStateAccessor t) {

        PhysicalSystemState lastState = null;
        
        for (PhysicalSystemState state : states) {
            Optional<PhysicalDropletState> dropState = t.getDropState(state);
            if (dropState.isPresent()) {
                lastState = state;
            } else if(lastState != null) {
                return lastState.time;
            }
        }

        return Double.POSITIVE_INFINITY;
    }
    
    private class DropStateAccessor {

        private final String dropletName;
        private final PhysicalDropletInjectionTime inj;

        public DropStateAccessor(String dropletName) {
            super();
            Objects.nonNull(dropletName);
            this.dropletName = dropletName;
            this.inj = null;
        }

        public DropStateAccessor(PhysicalDropletInjectionTime inj) {
            super();
            Objects.nonNull(inj);
            this.dropletName = null;
            this.inj = inj;
        }

        public Optional<PhysicalDropletState> getDropState(PhysicalSystemState state) {

            return inj != null ? state.getDropletStates().stream().filter(e -> e.getDropletInjectionTime() == inj).findAny()
                    : state.getDropletStates().stream().filter(e -> e.getName().equals(dropletName)).findAny();
        }
    }
}
