/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.CloggablePhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBypassChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletTailPosition;

/**
 * Utility method for determining the successor state for a droplet.
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public class SuccessorStateCalculator {

    public static SuccessorStateContainer nextDropletState(DropletPosition dropletPosition, Map<PhysicalEdge, Double> flowRates) {

        List<PhysicalEdge> successorEdges = new LinkedList<>();

        PhysicalEdge currentEdge = dropletPosition.edge;
        EndPoint<PhysicalEdge> currentNode = determineSuccessorEdges(flowRates, successorEdges, currentEdge);

        // Filter the list: A droplet cannot enter a bypass channel.
        List<PhysicalEdge> potentialDropletSuccessors = successorEdges.stream().filter(c -> !(c instanceof PhysicalBypassChannel))
                .collect(Collectors.toList());

        Optional<PhysicalEdge> dropletSuccessor;

        // Checks whether the node where the droplet arrived is a Lee sorter
        // and the droplet flows through the Lee-sorter in the sorting direction.
        if (currentNode.isLeeSorter && currentNode == currentEdge.dst) {
            double l_d = dropletPosition.getDropletLength();
            // Use same naming as in Matlab
            double l = currentNode.getLeeSortingLength();

            double w_v = potentialDropletSuccessors.stream().mapToDouble(m -> m.getWidth()).min().getAsDouble();
            double w_z = currentNode.getLeeSortingWidth();
            PhysicalEdge minWidthChannel = potentialDropletSuccessors.stream().min((i, j) -> Double.compare(i.getWidth(), j.getWidth()))
                    .get();
            double q_v = flowRates.get(minWidthChannel);
            PhysicalEdge maxWidthChannel = potentialDropletSuccessors.stream().max((i, j) -> Double.compare(i.getWidth(), j.getWidth()))
                    .get();
            double q_z = flowRates.get(maxWidthChannel);

            // Horizontal distance from the center of the droplet to the dividing point of the flow Z and the
            // flow V
            double x = (Math.abs(q_z) / Math.abs(q_v) / (1 + Math.abs(q_z) / Math.abs(q_v)) - 0.5) * l;

            // Compute the shear rate ratio
            double srr = Math.abs(q_z) / Math.abs(q_v) * Math.pow(w_v / w_z, 2.0);
            double theta = Math.acos(x / l_d);

            // Compute the projected area ratio
            double par = ((180 - 180 / Math.PI * theta) * Math.PI * Math.pow(l_d, 2) / 180 + Math.pow(x, 2) * Math.tan(theta))
                    / (Math.PI * Math.pow(l_d, 2)
                            - ((180 - 180 / Math.PI * theta) * Math.PI * Math.pow(l_d, 2) / 180 + Math.pow(x, 2) * Math.tan(theta)));
            // compute the shear force ratio using the inverse of the equation in Lee's paper
            double sfr = 1 / (srr * par);
            // If sfr is above 1 I go to channel V and to Z otherwise
            if (sfr > 1.0) {
                dropletSuccessor = Optional.of(minWidthChannel);
            } else {
                // Note, this branch is also entered when theta is an imaginary number (i.e. NaN).
                dropletSuccessor = Optional.of(maxWidthChannel);
            }
        } else {
            dropletSuccessor = potentialDropletSuccessors.stream()
                    .max((p1, p2) -> Double.compare(Math.abs(flowRates.get(p1)), Math.abs(flowRates.get(p2))));
        }

        return new SuccessorStateContainer(dropletSuccessor, getCloggingRelevantChannels(dropletPosition, flowRates));
    }

    private static EndPoint<PhysicalEdge> determineSuccessorEdges(Map<PhysicalEdge, Double> flowRates, List<PhysicalEdge> successorEdges,
            PhysicalEdge currentEdge) {

        EndPoint<PhysicalEdge> presentNode = getPresentNode(flowRates, currentEdge);
        successorEdges.addAll(presentNode.getPredecessors().stream().filter(pre -> pre != currentEdge && flowRates.get(pre) <= 0.0)
                .collect(Collectors.toList()));
        successorEdges.addAll(presentNode.getSuccessors().stream().filter(succ -> succ != currentEdge && flowRates.get(succ) >= 0.0)
                .collect(Collectors.toList()));
        return presentNode;
    }

    private static EndPoint<PhysicalEdge> getPresentNode(Map<PhysicalEdge, Double> flowRates, PhysicalEdge currentEdge) {

        EndPoint<PhysicalEdge> currentNode;
        if (flowRates.get(currentEdge) >= 0.0) {
            currentNode = currentEdge.dst;
        } else {
            currentNode = currentEdge.src;
        }
        return currentNode;
    }

    /**
     * Returns a list of channels which are not clogged anymore.
     * 
     * @param dropletPosition
     * @param flowRates
     * @return
     */
    public static Set<CloggablePhysicalChannel> getUncloggedChannels(DropletTailPosition dropletPosition,
            Map<PhysicalEdge, Double> flowRates) {

        return getCloggingRelevantChannels(dropletPosition, flowRates);
    }

    private static Set<CloggablePhysicalChannel> getCloggingRelevantChannels(DropletPosition dropletPosition,
            Map<PhysicalEdge, Double> flowRates) {

        PhysicalEdge currentEdge = dropletPosition.edge;

        EndPoint<PhysicalEdge> presentNode = getPresentNode(flowRates, currentEdge);
        Set<CloggablePhysicalChannel> collector = new HashSet<>();

        presentNode.getSuccessors().stream().filter(c -> c instanceof CloggablePhysicalChannel).map(c -> ((CloggablePhysicalChannel) c))
                .forEach(c -> collector.add(c));
        presentNode.getPredecessors().stream().filter(c -> c instanceof CloggablePhysicalChannel).map(c -> ((CloggablePhysicalChannel) c))
                .forEach(c -> collector.add(c));

        return collector;
    }

    private SuccessorStateCalculator() {

    }

}
