/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator.gui;

import java.awt.Color;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.electric.simulator.gui.components.ChannelDetailDrawable;
import at.jku.iic.microfluidics.droplet.electric.simulator.gui.components.DrawableAreaList;
import at.jku.iic.microfluidics.droplet.electric.simulator.gui.components.DrawableEdge;
import at.jku.iic.microfluidics.droplet.electric.simulator.gui.components.DropletAreaVisitor;
import at.jku.iic.microfluidics.droplet.electric.simulator.gui.components.SceneHandler;
import at.jku.iic.microfluidics.droplet.electric.simulator.gui.inout.Window;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.PhysicalDropletState;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.PhysicalSystemState;

public abstract class Placer {

    private static final Set<String> relevantFlowRates;

    static {
        relevantFlowRates = new HashSet<>();
        // relevantFlowRates.add("c1");
        // relevantFlowRates.add("c4");
        // relevantFlowRates.add("c5");
        // relevantFlowRates.add("c5");
        //
        // relevantFlowRates.add("c8");
        // relevantFlowRates.add("c9");
        // relevantFlowRates.add("c10");
        // relevantFlowRates.add("c12");
    }

    private static final Color[] dropletColors = { new Color(0, 51, 102), new Color(255, 204, 0),
            /* new Color(138, 5, 9), */ Color.WHITE, Color.DARK_GRAY, Color.RED, Color.GREEN };

    protected final PhysicalBiochip chip;

    public Placer(PhysicalBiochip chip) {

        super();
        this.chip = chip;
    }

    public abstract Map<PhysicalEdge, DrawableEdge> getPlacement(double factor);

    protected int getLength(PhysicalEdge e, double factor) {

        return (int) (e.getLength() * factor);
    }

    /**
     * This creates a graphical output of the states. It is only temporary, until an advanced GUI is
     * developed.
     * 
     * @param factor
     * @param states
     * @throws InterruptedException
     */
    public void printGraphicalOutput(double factor, List<PhysicalSystemState> states, long timeStepFactor, boolean writeScreenshots)
            throws InterruptedException {

        // Create a SceneHandler
        SceneHandler handler = new SceneHandler();

        Map<PhysicalEdge, DrawableEdge> edges = getPlacement(factor);
        Map<PhysicalEdge, ChannelDetailDrawable> detailMap = new IdentityHashMap<>();

        for (Map.Entry<PhysicalEdge, DrawableEdge> entry : edges.entrySet()) {
            handler.addDrawable(entry.getValue());
            if (relevantFlowRates.contains(entry.getKey().name)) {
                ChannelDetailDrawable cdd = new ChannelDetailDrawable(entry.getValue(), 0.0);
                handler.addDrawable(cdd);
                detailMap.put(entry.getKey(), cdd);
            }
        }

        // Open a window
        Window.open(handler, writeScreenshots);
        List<DrawableAreaList> lastIterationsDrawables = new LinkedList<>();

        // while (true) {
        for (int i = 0; i < states.size(); i++) {

            // Update flow rate text
            for (Map.Entry<PhysicalEdge, ChannelDetailDrawable> detail : detailMap.entrySet()) {
                detail.getValue().updateFlowRate(states.get(i).getFlowRate(detail.getKey()));
            }

            lastIterationsDrawables.forEach(e -> handler.removeDrawable(e));
            lastIterationsDrawables.clear();

            for (PhysicalDropletState droplet : states.get(i).getDropletStates()) {
                DropletAreaVisitor visitor = new DropletAreaVisitor(edges, factor);
                droplet.getDropletPositions().forEach(p -> p.accept(visitor));

                int pos = chip.getPumps().indexOf(droplet.getDropletInjectionTime().pump);
                Color c;
                if (pos < dropletColors.length) {
                    c = dropletColors[pos];
                } else {
                    c = Color.RED;
                }
                lastIterationsDrawables.add(new DrawableAreaList(visitor.getAreas(), c));
            }
            lastIterationsDrawables.forEach(e -> handler.addDrawable(e));

            if (i < states.size() - 1) {
                // try {
                // System.in.read();
                // } catch (IOException e) {
                // // TODO Auto-generated catch block
                // e.printStackTrace();
                // }
                // System.out.println(states.get(i).time);
                Thread.sleep((long) ((states.get(i + 1).time - states.get(i).time) * timeStepFactor));
            }
        }
        // }
    }
}
