/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator.gui.components;

import java.awt.geom.Area;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.electric.simulator.gui.inout.Window;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletEntireChannelPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletHeadPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletInjectionPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletPositionVisitor;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletRangePosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletTailPosition;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.DropletTrappedPosition;

public class DropletAreaVisitor implements DropletPositionVisitor {

    private final List<Area> areas;

    private final Map<PhysicalEdge, DrawableEdge> edges;
    private final double factor;

    public DropletAreaVisitor(Map<PhysicalEdge, DrawableEdge> edges, double factor) {
        super();
        this.edges = edges;
        this.factor = factor;
        this.areas = new LinkedList<>();
    }

    public Collection<Area> getAreas() {

        return Collections.unmodifiableCollection(areas);
    }

    @Override
    public void visitHead(DropletHeadPosition pos) {

        if (pos.defaultFlowDirection) {
            showDropHead(pos.headPosition, pos);
        } else {
            showDropTail(pos.edge.getLength() - pos.headPosition, pos);
        }
    }

    @Override
    public void visitInjection(DropletInjectionPosition pos) {

        if (pos.defaultFlowDirection) {
            showDropHead(pos.headPosition, pos);
        } else {
            showDropTail(pos.edge.getLength() - pos.headPosition, pos);
        }
    }

    private void showDropHead(double headPosition, DropletPosition pos) {

        DrawableEdge e = edges.get(pos.edge);
        int x;
        int y;
        int width;
        int height;

        int widthRect;
        int heightRect;

        int dropLength = (int) (pos.getFullyContainedDropletLength() * factor);
        int relativeDropTailPosition = (int) ((pos.getFullyContainedDropletLength() - headPosition) * factor);

        switch (e.rotation) {
        case ZERO:
            x = e.getAbsoluteX() - relativeDropTailPosition;
            y = e.getAbsoluteY();
            width = dropLength;
            height = e.height;

            widthRect = e.getAbsoluteX() - x;
            heightRect = height;
            break;
        case NINTY:
            x = e.getAbsoluteX();
            y = e.getAbsoluteY() + relativeDropTailPosition;
            width = e.width;
            height = -dropLength;

            widthRect = width;
            heightRect = e.getAbsoluteY() - y;
            break;
        case HUNDRED_AND_EIGHTY:
            x = e.getAbsoluteX() + relativeDropTailPosition;
            y = e.getAbsoluteY();
            width = -dropLength;
            height = e.height;

            widthRect = e.getAbsoluteX() - x;
            heightRect = height;
            break;
        case TWO_HUNDRED_AND_SEVENTY:
            x = e.getAbsoluteX();
            y = e.getAbsoluteY() - relativeDropTailPosition;
            width = e.width;
            height = dropLength;

            widthRect = width;
            heightRect = e.getAbsoluteY() - y;
            break;
        default:
            x = 0;
            y = 0;
            width = 0;
            height = 0;
            widthRect = 0;
            heightRect = 0;
        }

        Area drop = new Area(Window.getEllipse2D(x, y, width, height));
        Area cut = new Area(Window.getRectangle2D(x, y, widthRect, heightRect));
        drop.subtract(cut);
        
        // Ensure that droplet is only in the area of the module
        drop.intersect(e.getArea());
        
        areas.add(drop);
    }

    @Override
    public void visitRange(DropletRangePosition pos) {

        DrawableEdge e = edges.get(pos.edge);
        int x;
        int y;
        int width;
        int height;

        int nearestPosition;
        int dropLength;

        if (pos.defaultFlowDirection) {
            nearestPosition = (int) (pos.nearestPosition * factor);
            dropLength = (int) (pos.furthestPosition * factor) - nearestPosition;
        } else {
            nearestPosition = (int) ((pos.edge.getLength() - pos.furthestPosition) * factor);
            dropLength = (int) ((pos.edge.getLength() - pos.nearestPosition) * factor) - nearestPosition;
        }

        switch (e.rotation) {
        case ZERO:
            x = e.getAbsoluteX() + nearestPosition;
            y = e.getAbsoluteY();
            width = dropLength;
            height = e.height;
            break;
        case NINTY:
            x = e.getAbsoluteX();
            y = e.getAbsoluteY() - nearestPosition;
            width = e.width;
            height = -dropLength;
            break;
        case HUNDRED_AND_EIGHTY:
            x = e.getAbsoluteX() - nearestPosition;
            y = e.getAbsoluteY();
            width = -dropLength;
            height = e.height;
            break;
        case TWO_HUNDRED_AND_SEVENTY:
            x = e.getAbsoluteX();
            y = e.getAbsoluteY() + nearestPosition;
            width = e.width;
            height = dropLength;
            break;
        default:
            x = 0;
            y = 0;
            width = 0;
            height = 0;
        }

        areas.add(new Area(Window.getEllipse2D(x, y, width, height)));
    }

    @Override
    public void visitTail(DropletTailPosition pos) {

        if (pos.defaultFlowDirection) {
            showDropTail(pos.tailPosition, pos);
        } else {
            showDropHead(pos.edge.getLength() - pos.tailPosition, pos);
        }
    }

    private void showDropTail(double tailPosition, DropletPosition pos) {

        DrawableEdge e = edges.get(pos.edge);
        int x;
        int y;
        int width;
        int height;

        int xRect;
        int yRect;

        int dropLength = (int) (pos.getFullyContainedDropletLength() * factor);
        int dropTailPosition = (int) (tailPosition * factor);

        switch (e.rotation) {
        case ZERO:
            x = e.getAbsoluteX() + dropTailPosition;
            y = e.getAbsoluteY();
            width = dropLength;
            height = e.height;

            xRect = e.getAbsoluteX() + e.width;
            yRect = e.getAbsoluteY();
            break;
        case NINTY:
            x = e.getAbsoluteX();
            y = e.getAbsoluteY() - dropTailPosition;
            width = e.width;
            height = -dropLength;

            xRect = e.getAbsoluteX();
            yRect = e.getAbsoluteY() + e.height;
            break;
        case HUNDRED_AND_EIGHTY:
            x = e.getAbsoluteX() - dropTailPosition;
            y = e.getAbsoluteY();
            width = -dropLength;
            height = e.height;

            xRect = e.getAbsoluteX() + e.width;
            yRect = e.getAbsoluteY();
            break;
        case TWO_HUNDRED_AND_SEVENTY:
            x = e.getAbsoluteX();
            y = e.getAbsoluteY() + dropTailPosition;
            width = e.width;
            height = dropLength;

            xRect = e.getAbsoluteX();
            yRect = e.getAbsoluteY() + e.height;
            break;
        default:
            x = 0;
            y = 0;
            width = 0;
            height = 0;
            xRect = 0;
            yRect = 0;
        }

        Area drop = new Area(Window.getEllipse2D(x, y, width, height));
        Area cut = new Area(Window.getRectangle2D(xRect, yRect, width, height));
        drop.subtract(cut);
        areas.add(drop);
    }

    @Override
    public void visitEntire(DropletEntireChannelPosition pos) {

        DrawableEdge e = edges.get(pos.edge);
        areas.add(new Area(Window.getRectangle2D(e.getAbsoluteX(), e.getAbsoluteY(), e.width, e.height)));
    }

    @Override
    public void visitTrapped(DropletTrappedPosition pos) {

        DrawableEdge e = edges.get(pos.edge);
        areas.add(new Area(Window.getEllipse2D(e.getAbsoluteX(), e.getAbsoluteY(), e.width, e.height)));
    }
}
