/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator.gui.inout;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsEnvironment;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.geom.Area;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;

/**
 * This class allows for simple output of graphical objects on a window. It provides several methods to draw
 * points, lines, circles and text.
 */
public class Window {

    /** Constant for standard width */
    private static final int DEFAULT_WIDTH = 1000;

    /** Constant for height */
    private static final int DEFAULT_HEIGHT = 1100;

    /** X position of the next window that is openend */
    private static int xPosition = 0;

    protected Graphics g;

    private final RedrawHandler handler;

    private final boolean writeScreenshots;

    /**
     * Opens a new window and periodically calls {@link RedrawHandler#redraw(Window)} until the window is
     * closed.
     * 
     * @param handler
     *            a RedrawHandler instance that is used to periodically update the window contents.
     */
    public static void open(final RedrawHandler handler, boolean writeScreenshots) {

        new Thread() {

            public void run() {

                new Window(handler, writeScreenshots);
            }
        }.start();
    }

    private Window(RedrawHandler handler, boolean writeScreenshots) {

        this.handler = handler;
        this.writeScreenshots = writeScreenshots;
        if (GraphicsEnvironment.isHeadless()) {
            System.out.println("Graphical output will not work in a headless environment.");
            System.out.println("Please use -Djava.awt.headless=false VM parameter.");
        }
        AnimationFrame contents = new AnimationFrame();
        contents.setBounds(xPosition, 0, DEFAULT_WIDTH, DEFAULT_HEIGHT);
        xPosition += 100;
        contents.setVisible(true);
        contents.addWindowListener(new WindowAdapter() {

            public void windowClosing(WindowEvent evt) {

                Frame frame = (Frame) evt.getSource();
                frame.setVisible(false);
                frame.dispose();
            }
        });

        while (contents.isVisible()) {
            contents.repaint();
            try {
                Thread.sleep(20);
            } catch (InterruptedException e) {
            }
        }
    }

    /**
     * Static inner class for the content pane of the frame. Overrides the paint method and draws the image
     * object.
     */
    @SuppressWarnings("serial")
    private class AnimationFrame extends Frame {

        private BufferedImage image = null;
        private int width = 0;
        private int height = 0;
        private int counter = 1;

        @Override
        public void update(Graphics g) {

            paint(g);
        }

        @Override
        public void paint(Graphics g) {

            int contentWidth = g.getClipBounds().width;
            int contentHeight = g.getClipBounds().height;
            if (image == null || width != contentWidth || height != contentHeight) {
                width = contentWidth;
                height = contentHeight;
                image = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
            }
            Window.this.g = image.getGraphics();
            if (Window.this.g != null) {
                Window.this.g.fillRect(0, 0, contentWidth, contentHeight);
                handler.redraw(Window.this, contentWidth, contentHeight);
                Window.this.g.dispose();
                g.drawImage(image, 0, 0, null);

                if (writeScreenshots) {
                    try {
                        File outputfile = new File("visualization/vis" + String.format("%03d", counter++) + ".jpg");
                        ImageIO.write(image, "jpg", outputfile);
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
    }

    /**
     * Computes the ascent of the current font, ascent is the distance between the baseline and the top border
     * of the text.
     */
    private int getTextAscent() {

        FontMetrics fm = g.getFontMetrics();
        return fm.getAscent();
    }

    /**
     * Computes the width of a string.
     * 
     * @param text
     *            the string
     * @return the width of the string in pixel
     */
    public int getTextWidth(String text) {

        FontMetrics fm = g.getFontMetrics();
        return (int) Math.round(fm.getStringBounds(text, g).getWidth());
    }

    /**
     * Computes the height of text.
     * 
     * @return the height of text in pixel
     */
    public int getTextHeight() {

        FontMetrics fm = g.getFontMetrics();
        return fm.getHeight();
    }

    // Methods for drawing

    public void fillArea(Area a, Color c) {

        Graphics2D g2D = (Graphics2D) g;
        g2D.setColor(c);
        g2D.fill(a);
    }

    public void drawArea(Area a, Color c) {

        Graphics2D g2D = (Graphics2D) g;
        g2D.setColor(c);
        g2D.draw(a);
    }

    /**
     * Draws a rectangle on position (x, y) with width w and height h.
     * 
     * @param x
     *            the x-coordinate of the position
     * @param y
     *            the y-coordinate of the position
     * @param w
     *            the width of the rectangle
     * @param h
     *            the height of the rectangle
     */
    public void drawRectangle(int x, int y, int w, int h) {

        g.setColor(Color.black);
        g.drawRect(x, y, w, h);
    }

    /**
     * Draws the text left aligned at given position
     * 
     * @param text
     *            the text to draw
     * @param x
     *            the x-coordinate
     * @param y
     *            the y-coordinate
     */
    public void drawText(String text, int x, int y) {

        g.setColor(Color.black);
        g.setFont(new Font("TimesRoman", Font.PLAIN, 16));
        g.drawString(text, x, y + getTextAscent());
    }

    /**
     * Draws a rectangle on position (x, y) with width w and height h with given color.
     * 
     * @param x
     *            the x-coordinate of the position
     * @param y
     *            the y-coordinate of the position
     * @param w
     *            the width of the rectangle
     * @param h
     *            the height of the rectangle
     */
    public void drawRectangle(int x, int y, int w, int h, Color color) {

        if (w < 0) {
            x += w;
            w = -w;
        }
        if (h < 0) {
            y += h;
            h = -h;
        }

        g.setColor(color);
        g.drawRect(x, y, w, h);
    }

    /**
     * Draws the text left aligned at given position with given color
     * 
     * @param text
     *            the text to draw
     * @param x
     *            the x-coordinate
     * @param y
     *            the y-coordinate
     */
    public void drawText(String text, int x, int y, Color color) {

        g.setColor(color);
        g.drawString(text, x, y + getTextAscent());
    }

    /**
     * Fills a rectangle on position (x, y) with width w and height h with given color.
     * 
     * @param x
     *            the x-coordinate of the position
     * @param y
     *            the y-coordinate of the position
     * @param w
     *            the width of the rectangle
     * @param h
     *            the height of the rectangle
     */
    public void fillRectangle(int x, int y, int w, int h, Color color) {

        if (w < 0) {
            x += w;
            w = -w;
        }
        if (h < 0) {
            y += h;
            h = -h;
        }

        g.setColor(color);
        g.fillRect(x, y, w, h);
    }

    /**
     * Fills a circle at center (x, y) and radius r with given color.
     * 
     * @param x
     *            the x-coordinate of the center
     * @param y
     *            the y-coordinate of the center
     * @param r
     *            the radius of the circle
     */
    public void fillCircle(int x, int y, int w, int h, Color color) {

        if (w < 0) {
            x += w;
            w = -w;
        }
        if (h < 0) {
            y += h;
            h = -h;
        }

        g.setColor(color);
        g.fillOval(x, y, w, h);
    }

    public static Ellipse2D getEllipse2D(int x, int y, int w, int h) {

        if (w < 0) {
            x += w;
            w = -w;
        }
        if (h < 0) {
            y += h;
            h = -h;
        }

        return new Ellipse2D.Double(x, y, w, h);
    }

    public static Rectangle2D getRectangle2D(int x, int y, int w, int h) {

        if (w < 0) {
            x += w;
            w = -w;
        }
        if (h < 0) {
            y += h;
            h = -h;
        }

        return new Rectangle2D.Double(x, y, w, h);
    }

}
