/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator.state;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import at.jku.iic.microfluidics.droplet.basic.architecture.physical.CloggablePhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.TrappablePhysicalModule;
import at.jku.iic.microfluidics.droplet.electric.simulator.SuccessorStateCalculator;
import at.jku.iic.microfluidics.droplet.electric.simulator.SuccessorStateContainer;
import at.jku.iic.microfluidics.droplet.electric.simulator.event.Event;

/**
 * Represents the position of the droplet head.
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public class DropletHeadPosition extends DropletPosition {

    private static final double ALLOWED_TOLERANCE = 0.995;
    /**
     * The head position (in meters) of the droplet within the channel.
     */
    public final double headPosition;

    public DropletHeadPosition(PhysicalEdge edge, boolean defaultFlowDirection, double dropletVolume, double headPosition) {
        super(edge, defaultFlowDirection, dropletVolume);
        if(headPosition > edge.getLength()) {
            // Less than 0.1 per-mille error rate --> probably a double rounding error
            if(headPosition < edge.getLength() * 1.0001) {
                this.headPosition = edge.getLength();
            } else{
                throw new RuntimeException("This case should not occur!");
            }   
        } else {
            this.headPosition = headPosition;
        }
    }

    @Override
    public double getTimeDifferenceToNextEvent(Map<PhysicalEdge, Double> flowRates, double slipFactor) {

        return (edge.getLength() - headPosition) / getAbsoluteDropletVelocity(flowRates, slipFactor);
    }

    @Override
    public double getDropletLength() {

        return headPosition;
    }

    @Override
    public List<DropletPosition> moveDroplet(double timeDifference, Map<PhysicalEdge, Double> flowRates,
            Set<CloggablePhysicalChannel> cloggedChannels, Event event, double dropletVolumeShift, double slipFactor) {

        List<DropletPosition> newDropletPositions = new LinkedList<>();

        double newHeadPosition = headPosition + dropletVolumeShift / (edge.getHeight() * edge.getWidth());
        if (newHeadPosition < edge.getLength() * ALLOWED_TOLERANCE) {
            // In this case the head possible triggered the next event, but the droplet is not yet
            // fully contained in the channel.
//            assert newHeadPosition <= getFullyContainedDropletLength() || Math.abs(newHeadPosition - getFullyContainedDropletLength()) < newHeadPosition * 0.001;
            // Move droplet head with the velocity of the predecessor (therefore use the droplet volume)
            newDropletPositions.add(new DropletHeadPosition(edge, defaultFlowDirection, dropletVolume, newHeadPosition));
        } else {
            assert isTriggeringEvent(event) || newHeadPosition >= edge.getLength() * ALLOWED_TOLERANCE;
            SuccessorStateContainer succContainer = SuccessorStateCalculator.nextDropletState(this, flowRates);

            newDropletPositions.add(new DropletEntireChannelPosition(edge, defaultFlowDirection, dropletVolume));

            if (succContainer.getSuccessorEdge().isPresent()) {
                cloggedChannels.addAll(succContainer.getGetCloggedChannels());

                double overlapVolume = dropletVolumeShift - (edge.getLength() - headPosition) * (edge.getHeight() * edge.getWidth());
                double succHeadPosition = overlapVolume
                        / (succContainer.getSuccessorEdge().get().getHeight() * succContainer.getSuccessorEdge().get().getWidth());

                newDropletPositions.add(new DropletHeadPosition(succContainer.getSuccessorEdge().get(),
                        flowRates.get(succContainer.getSuccessorEdge().get()) >= 0, dropletVolume, succHeadPosition));
            } else if (edge instanceof TrappablePhysicalModule) {
                // When the droplet is contained in a trappable module, the droplet is trapped.
                newDropletPositions.add(new DropletTrappedPosition(edge, defaultFlowDirection, dropletVolume, false));
            }
            // Otherwise, the droplet exits the system
        }
        return newDropletPositions;
    }

    @Override
    public String toString() {

        return "Head; " + edge.name + "; " + defaultFlowDirection + "; headPosition=" + headPosition;
    }
    
    @Override
    public void accept(DropletPositionVisitor v) {
    
        v.visitHead(this);
    }
}
