/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator.state;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import at.jku.iic.microfluidics.droplet.basic.architecture.physical.CloggablePhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.electric.simulator.event.Event;

/**
 * Represents the position of a new injected droplet.
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public class DropletInjectionPosition extends DropletPosition {

    /**
     * The head position (in meters) of the droplet within the channel.
     */
    public final double headPosition;

    public DropletInjectionPosition(PhysicalEdge edge, boolean defaultFlowDirection, double dropletVolume, double headPosition) {
        super(edge, defaultFlowDirection, dropletVolume);
        this.headPosition = headPosition;

        double dropletLength = dropletVolume / (edge.getHeight() * edge.getWidth());
        if (dropletLength >= edge.getLength()) {
            throw new UnsupportedOperationException("It is a prerequisite that the first channel can contain the full droplet!");
        }
    }

    @Override
    public double getTimeDifferenceToNextEvent(Map<PhysicalEdge, Double> flowRates, double slipFactor) {

        return (getFullyContainedDropletLength() - headPosition) / getAbsoluteDropletVelocity(flowRates, slipFactor);
    }

    @Override
    public double getDropletLength() {

        return headPosition;
    }

    @Override
    public List<DropletPosition> moveDroplet(double timeDifference, Map<PhysicalEdge, Double> flowRates,
            Set<CloggablePhysicalChannel> cloggedChannels, Event event, double dropletVolumeShift, double slipFactor) {

        List<DropletPosition> newDropletPositions = new LinkedList<>();

        if (isTriggeringEvent(event)) {
            // The droplet is now fully contained in the channel
            newDropletPositions.add(new DropletRangePosition(edge, defaultFlowDirection, dropletVolume,
                    getFullyContainedDropletLength(), 0.0));
        } else {
            double distance = getAbsoluteDropletVelocity(flowRates, slipFactor) * timeDifference;
            newDropletPositions.add(new DropletInjectionPosition(edge, defaultFlowDirection, dropletVolume, headPosition + distance));
        }
        return newDropletPositions;
    }

    @Override
    public String toString() {

        return "Injection; " + edge.name + "; " + defaultFlowDirection + "; headPosition=" + headPosition;
    }
    
    @Override
    public void accept(DropletPositionVisitor v) {
    
        v.visitInjection(this);
    }

}
