/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator.state;

import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import at.jku.iic.microfluidics.droplet.basic.architecture.physical.CloggablePhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.electric.simulator.event.Event;
import at.jku.iic.microfluidics.droplet.electric.simulator.event.NewDropletPosition;

/**
 * Represents the position of a droplet.
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public abstract class DropletPosition {

    /**
     * The edge which contains (a part of) the droplet.
     */
    public final PhysicalEdge edge;

    /**
     * Represents the flow direction of the channel.
     */
    public final boolean defaultFlowDirection;

    /**
     * The volume of the droplet. This volume allows to determine the length of a droplet in a channel.
     */
    public final double dropletVolume;

    public DropletPosition(PhysicalEdge edge, boolean defaultFlowDirection, double dropletVolume) {
        super();
        Objects.requireNonNull(edge);
        this.edge = edge;
        this.defaultFlowDirection = defaultFlowDirection;
        this.dropletVolume = dropletVolume;
    }

    public abstract double getTimeDifferenceToNextEvent(Map<PhysicalEdge, Double> flowRates, double slipFactor);

    public double getFullyContainedDropletLength() {

        return dropletVolume / (edge.getHeight() * edge.getWidth());
    }

    public double getDropletResistance(double muC) {

        return getDropletResistance(edge, muC, getDropletLength());
    }

    public abstract double getDropletLength();

    public static double getDropletResistance(PhysicalEdge edge, double muC, double dropletLength) {

        return 3 * PhysicalEdge.getAFactor(edge.getWidth(), edge.getHeight()) * muC * dropletLength
                / (edge.getWidth() * Math.pow(edge.getHeight(), 3.0));
//        System.out.println("Droplet resistance manually changed!");
//        return 19 * PhysicalEdge.getAFactor(edge.getWidth(), edge.getHeight()) * muC * dropletLength
//                / (edge.getWidth() * Math.pow(edge.getHeight(), 3.0));
    }

    public abstract List<DropletPosition> moveDroplet(double timeDifference, Map<PhysicalEdge, Double> flowRates,
            Set<CloggablePhysicalChannel> cloggedChannels, Event event, double dropletVolumeShift, double slipFactor);

    public double getAbsoluteDropletVelocity(Map<PhysicalEdge, Double> flowRates, double slipFactor) {

        return getAbsoluteDropletVelocity(flowRates.get(edge), slipFactor);
    }
    
    public double getAbsoluteDropletVelocity(double flowRate, double slipFactor) {
        
        return Math.abs(flowRate / (edge.getHeight() * edge.getWidth())) * slipFactor;
    }

    protected boolean isTriggeringEvent(Event event) {

        return event instanceof NewDropletPosition && ((NewDropletPosition) event).triggeringDropletPosition == this;
    }

    public boolean hasDropletFlowDirectionChanged(Map<PhysicalEdge, Double> flowRates) {

         return hasDropletFlowDirectionChanged(flowRates.get(edge));
    }
    
    public boolean hasDropletFlowDirectionChanged(double flowRate) {
        
        return defaultFlowDirection ^ flowRate >= 0;
    }
    
    public abstract void accept(DropletPositionVisitor v);

}
