/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator.state;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import at.jku.iic.microfluidics.droplet.basic.architecture.physical.CloggablePhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.TrappablePhysicalModule;
import at.jku.iic.microfluidics.droplet.electric.simulator.SuccessorStateCalculator;
import at.jku.iic.microfluidics.droplet.electric.simulator.SuccessorStateContainer;
import at.jku.iic.microfluidics.droplet.electric.simulator.event.Event;

/**
 * Represents the position of a droplet, which is flowing in a channel.
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public class DropletRangePosition extends DropletPosition {

    /**
     * The furthest position (in meters) of the droplet within the channel.
     */
    public final double furthestPosition;

    /**
     * The nearest position (in meters) of the droplet w.r.t. the channel entrance.
     */
    public final double nearestPosition;

    public DropletRangePosition(PhysicalEdge edge, boolean defaultFlowDirection, double dropletVolume, double furthestPosition,
            double nearestPosition) {
        super(edge, defaultFlowDirection, dropletVolume);
        if (furthestPosition > edge.getLength()) {
            // Less than 0.1 per-mille error rate --> probably a double rounding error
            if (furthestPosition < edge.getLength() * 1.0001) {
                this.furthestPosition = edge.getLength();
            } else {
                throw new RuntimeException("This case should not occur!");
            }
        } else {
            this.furthestPosition = furthestPosition;
        }
        this.nearestPosition = nearestPosition;
    }

    @Override
    public double getTimeDifferenceToNextEvent(Map<PhysicalEdge, Double> flowRates, double slipFactor) {

        return (edge.getLength() - furthestPosition) / getAbsoluteDropletVelocity(flowRates, slipFactor);
    }

    @Override
    public double getDropletLength() {

        return furthestPosition - nearestPosition;
    }

    @Override
    public List<DropletPosition> moveDroplet(double timeDifference, Map<PhysicalEdge, Double> flowRates,
            Set<CloggablePhysicalChannel> cloggedChannels, Event event, double dropletVolumeShift, double slipFactor) {

        List<DropletPosition> newDropletPositions = new LinkedList<>();

        if (isTriggeringEvent(event)) {
            SuccessorStateContainer succContainer = SuccessorStateCalculator.nextDropletState(this, flowRates);

            if (!succContainer.getSuccessorEdge().isPresent() && edge instanceof TrappablePhysicalModule) {
                // When the droplet is contained in a trappable module, the droplet stays there.
                cloggedChannels.addAll(succContainer.getGetCloggedChannels());
                newDropletPositions.add(new DropletTrappedPosition(edge, defaultFlowDirection, dropletVolume, true));
            } else {
                newDropletPositions.add(new DropletTailPosition(edge, defaultFlowDirection, dropletVolume,
                        this.nearestPosition + getAbsoluteDropletVelocity(flowRates, slipFactor) * timeDifference));
                if (succContainer.getSuccessorEdge().isPresent()) {
                    cloggedChannels.addAll(succContainer.getGetCloggedChannels());
                    newDropletPositions.add(new DropletHeadPosition(succContainer.getSuccessorEdge().get(),
                            flowRates.get(succContainer.getSuccessorEdge().get()) >= 0, dropletVolume, 0.0));
                }
            }

        } else {
            newDropletPositions.add(new DropletRangePosition(edge, defaultFlowDirection, dropletVolume,
                    furthestPosition + getAbsoluteDropletVelocity(flowRates, slipFactor) * timeDifference,
                    nearestPosition + getAbsoluteDropletVelocity(flowRates, slipFactor) * timeDifference));
        }
        return newDropletPositions;
    }

    @Override
    public String toString() {

        return "Range; " + edge.name + "; " + defaultFlowDirection + "; furthestPosition=" + furthestPosition + "; nearestPosition"
                + nearestPosition;
    }

    @Override
    public void accept(DropletPositionVisitor v) {

        v.visitRange(this);
    }
}
