/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator.state;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import at.jku.iic.microfluidics.droplet.basic.architecture.physical.CloggablePhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.electric.simulator.SuccessorStateCalculator;
import at.jku.iic.microfluidics.droplet.electric.simulator.event.Event;

/**
 * Represents the position of the droplet tail.
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public class DropletTailPosition extends DropletPosition {

    /**
     * The tail position (in meters) of the droplet within the channel.
     */
    public final double tailPosition;

    public DropletTailPosition(PhysicalEdge edge, boolean defaultFlowDirection, double dropletVolume, double tailPosition) {
        super(edge, defaultFlowDirection, dropletVolume);
        this.tailPosition = tailPosition;
    }

    @Override
    public double getTimeDifferenceToNextEvent(Map<PhysicalEdge, Double> flowRates, double slipFactor) {

        return (edge.getLength() - tailPosition) / getAbsoluteDropletVelocity(flowRates, slipFactor);
    }

    @Override
    public double getDropletLength() {

        return edge.getLength() - tailPosition;
    }

    @Override
    public List<DropletPosition> moveDroplet(double timeDifference, Map<PhysicalEdge, Double> flowRates,
            Set<CloggablePhysicalChannel> cloggedChannels, Event event, double dropletVolumeShift, double slipFactor) {

        List<DropletPosition> newDropletPositions = new LinkedList<>();
        double distance = dropletVolumeShift / (edge.getHeight() * edge.getWidth());

        // The second condition checking whether the new tailPoisition is already outside the channel length
        // can evaluate to true, when an average of velocities is taken.
        if (isTriggeringEvent(event) || tailPosition + distance > edge.getLength()) {
            // After this event, the droplet is fully contained in the next channel
            // Remove unclogged channels
            cloggedChannels.removeAll(SuccessorStateCalculator.getUncloggedChannels(this, flowRates));
        } else {
            newDropletPositions.add(new DropletTailPosition(edge, defaultFlowDirection, dropletVolume, tailPosition + distance));
        }
        return newDropletPositions;
    }

    @Override
    public String toString() {

        return "Tail; " + edge.name + "; " + defaultFlowDirection + "; tailPosition=" + tailPosition;
    }

    @Override
    public void accept(DropletPositionVisitor v) {

        v.visitTail(this);
    }

}
