/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator.state;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import at.jku.iic.microfluidics.droplet.basic.architecture.physical.CloggablePhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalDropletInjectionTime;

/**
 * A container storing the droplet positions.
 * 
 * @author Andreas Grimmer &lt;andreas.grimmer@jku.at&gt;
 *
 */
public class PhysicalDropletState {

    private final LinkedList<DropletPosition> dropletPositions;
    private final String name;
    private final Set<CloggablePhysicalChannel> cloggedChannels;
    private final PhysicalDropletInjectionTime dropletInjectionTime;

    public PhysicalDropletState(LinkedList<DropletPosition> dropletPositions, String name, Set<CloggablePhysicalChannel> cloggedChannels, PhysicalDropletInjectionTime dropletInjectionTime) {

        this.dropletPositions = dropletPositions;
        this.name = name;
        this.cloggedChannels = cloggedChannels;
        this.dropletInjectionTime = dropletInjectionTime;
    }

    public List<DropletPosition> getDropletPositions() {

        return Collections.unmodifiableList(dropletPositions);
    }

    public String getName() {

        return name;
    }
    
    public PhysicalDropletInjectionTime getDropletInjectionTime() {
        
        return dropletInjectionTime;
    }
    
    public DropletPosition getDropletPositionOfTail() {
        
        return dropletPositions.getFirst();
    }
    
    public DropletPosition getDropletPositionOfHead() {

        return dropletPositions.getLast();
    }
    
    
    public Set<CloggablePhysicalChannel> getCloggedChannels() {
        
        return Collections.unmodifiableSet(cloggedChannels);
    }
    
    public static  Set<CloggablePhysicalChannel> getAllCloggedChannels(List<PhysicalDropletState> newDropletStates) {
        
        return newDropletStates.stream().map(r -> r.getCloggedChannels())
                .flatMap(Set::stream).collect(Collectors.toSet());
    }
    
    
    @Override
    public String toString() {
    
        StringBuilder sb = new StringBuilder();
        
        sb.append(name).append('\n');
        sb.append("Droplet positions:").append('\n');
        dropletPositions.stream().forEach(s -> sb.append('\t').append(s.toString()).append('\n'));
        sb.append("Clogged channels:").append('\n');
        cloggedChannels.stream().forEach(c -> sb.append('\t').append(c.toString()).append('\n'));
        return sb.toString();
    }


}
