/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.droplet.electric.simulator.state;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import at.jku.iic.microfluidics.droplet.basic.architecture.physical.CloggablePhysicalChannel;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.LaplacePressureParameter;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.TrappablePhysicalModule;
import at.jku.iic.microfluidics.droplet.electric.simulator.event.NewDropletPosition;

public class PhysicalSystemState {

    /**
     * Represents the current simulation time
     */
    public final double time;

    /**
     * The current states of the droplets at the corresponding time.
     */
    private final List<PhysicalDropletState> dropletStates;

    private final Map<PhysicalEdge, Double> flowRates;

    public PhysicalSystemState(double time, List<PhysicalDropletState> dropletStates, Map<PhysicalEdge, Double> flowRates) {
        super();
        this.time = time;
        this.dropletStates = dropletStates;
        this.flowRates = flowRates;
    }

    public Map<PhysicalEdge, Double> getFlowRates() {

        return Collections.unmodifiableMap(flowRates);
    }

    public double getFlowRate(PhysicalEdge edge) {

        return flowRates.get(edge);
    }

    public List<PhysicalDropletState> getDropletStates() {

        return Collections.unmodifiableList(dropletStates);
    }

    /**
     * Determines the next triggering droplet position.
     * 
     * @param flowRates
     *            the current flow rates in the channels.
     * @return
     */
    public Optional<NewDropletPosition> getNextTriggeringPosition(Map<PhysicalEdge, Double> flowRates, double slipFactor) {

        // Second determine the minimum time difference triggering the next event
        double minTimeDifference = Double.MAX_VALUE;
        DropletPosition triggeringPosition = null;
        for (PhysicalDropletState state : dropletStates) {
            for (DropletPosition pos : state.getDropletPositions()) {
                if (minTimeDifference > pos.getTimeDifferenceToNextEvent(flowRates, slipFactor)) {
                    minTimeDifference = pos.getTimeDifferenceToNextEvent(flowRates, slipFactor);
                    triggeringPosition = pos;
                }
            }
        }

        return triggeringPosition == null ? Optional.empty()
                : Optional.of(new NewDropletPosition(triggeringPosition, time + minTimeDifference));
    }

    public TrappablePhysicalModule getLaplacePressureEvent(PhysicalBiochip chip) {

        Set<CloggablePhysicalChannel> allCloggedChannels = PhysicalDropletState.getAllCloggedChannels(dropletStates);
        for (PhysicalDropletState state : dropletStates) {
            for (DropletPosition pos : state.getDropletPositions()) {
                if (pos instanceof DropletTrappedPosition) {
                    // This droplet is trapped.
                    // Check Laplace pressure
                    assert state.getDropletPositionOfHead().edge instanceof TrappablePhysicalModule;
                    TrappablePhysicalModule tpm = (TrappablePhysicalModule) state.getDropletPositionOfHead().edge;

                    for (LaplacePressureParameter lpp : tpm.getLaplacePressureParameters()) {

                        if (lpp.isPressureExceeding(flowRates, allCloggedChannels, chip)
                                && Collections.disjoint(getEdgesContainingADroplet(), lpp.getEmptyEdges())) {
                            return tpm;
                        }
                    }
                }
            }
        }
        return null;
    }

    private List<PhysicalEdge> getEdgesContainingADroplet() {

        return dropletStates.stream().map(e -> e.getDropletPositions()).flatMap(List::stream).map(e -> e.edge).collect(Collectors.toList());
    }

    @Override
    public String toString() {

        StringBuilder sb = new StringBuilder();
        sb.append("System state at ").append(time).append('\n');
        dropletStates.stream().forEach(a -> sb.append(a.toString()).append('\n'));
        return sb.toString();
    }
}
