package at.jku.iic.microfluidics.droplet.improvedrobustness;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Random;
import java.util.stream.Collectors;
import org.apache.commons.lang3.time.StopWatch;
import org.json.JSONException;
import at.jku.iic.microfluidics.droplet.improvedrobustness.metric.RobustnessMetric;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.Parameter;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.Parameters;
import at.jku.iic.microfluidics.droplet.improvedrobustness.simplex.FunctionSyntheticChip;
import at.jku.iic.microfluidics.droplet.improvedrobustness.singleparametervariation.SingleParameterVariation;
import at.jku.iic.microfluidics.droplet.improvedrobustness.testarchitectures.syntheticchip.SyntheticChip;
import org.apache.commons.math3.optim.InitialGuess;
import org.apache.commons.math3.optim.MaxEval;
import org.apache.commons.math3.optim.PointValuePair;
import org.apache.commons.math3.optim.nonlinear.scalar.GoalType;
import org.apache.commons.math3.optim.nonlinear.scalar.MultivariateFunctionMappingAdapter;
import org.apache.commons.math3.optim.nonlinear.scalar.ObjectiveFunction;
import org.apache.commons.math3.optim.nonlinear.scalar.noderiv.NelderMeadSimplex;
import org.apache.commons.math3.optim.nonlinear.scalar.noderiv.SimplexOptimizer;

public class ImprovedRobustness {

	public static void main(String[] args) throws JSONException, IOException {
		// chip name in corresponding resource folder: "syntheticChip.<chipNumber>.<design>.json"
		int chipNumber = 1; // either 1, 2 or 3
		String design = "initial";	//part of the file name of the chip so different designs can be analyzed

		System.out.println(String.format("Chip: %d, Design: %s", chipNumber, design));

		// start watch
		StopWatch watch = StopWatch.createStarted();

		// ### compute metric or improve robustness ###
		// just comment/uncomment the following methods to compute the desired function
		computeMetric_SyntheticChip(chipNumber, design);
		// improvementSingleParameterVariation_SyntheticChip(chipNumber, design);
		// improvementSimplex_SyntheticChip(chipNumber, design);

		// in order to compute the metric of an improved design just copy the generated file of the improvement process to the following resource folder and change the variable "design" accordingly:
		// src/main/resources/at/jku/iic/microfluidics/droplet/improvedrobustness/testarchitectures/syntheticchip/

		// end watch
		watch.stop();
		System.out.println(String.format("Elapsed Time = %.3f s", watch.getTime() / 1000.0));
	}

	public static void computeMetric_SyntheticChip(int chipNumber, String design) throws JSONException, IOException {
		// create synthetic chip
		SyntheticChip syntheticChip = new SyntheticChip(chipNumber, design);
		syntheticChip.channelNames = syntheticChip.getConsideredChannelNames();

		// create parameters and set standard deviation
		Parameters parameters = syntheticChip.getParametersForRobustnessAnalysis();
		double[] std = {0.005, 0.002, 0.002};
		parameters.forEach(p -> p.std = std[chipNumber - 1]);

		// define robustness metric
		RobustnessMetric metric = new RobustnessMetric(parameters, syntheticChip);
		Random randomGenerator = new Random(100);

		// compute robustness metric
		StopWatch watch = StopWatch.createStarted();
		System.out.println("Start metric computation:");
		double metricValue = metric.computeRobustnessMetric(randomGenerator);
		watch.stop();
		System.out.println(String.format("Elapsed Time = %.3f s", watch.getTime() / 1000.0));
		System.out.println("Final Value: " + metricValue);

		// save pSuccess values in file
		String content = "";
		for (Double temp : metric.pSuccessValues) {
			content += String.format("%.3e\n", temp);
		}
		Path filePath = Paths.get("pSuccessValues.csv");
		Files.writeString(filePath, content);
		System.out.println("File was saved to: " + filePath);
	}

	public static void improvementSingleParameterVariation_SyntheticChip(int chipNumber, String design) throws JSONException, IOException {
		// create synthetic chip
		SyntheticChip syntheticChip = new SyntheticChip(chipNumber, design);

		// create parameters
		syntheticChip.channelNames = syntheticChip.getConsideredChannelNames();
		Parameters parameters = syntheticChip.getParametersForParameterVariations();

		// improve
		System.out.println("Start Single Parameter Variation:");
		int[] nSensitive = {16, 8, 8}; // number of sensitive parameters for each chip
		SingleParameterVariation single = new SingleParameterVariation(parameters, syntheticChip);
		single.nIterationsMax = 1000;
		single.nConsiderSensitiveParametersMax = nSensitive[chipNumber - 1];
		single.startImprovement();

		System.out.println("nSimulations: " + single.nSimulations);

		// save improved design
		Path filePath = Paths.get("syntheticChip." + chipNumber + ".improved.single.json");
		Files.writeString(filePath, syntheticChip.createChip(parameters.getParameterMap()).toJSON().toString(4));
		System.out.println("File was saved to: " + filePath);
	}

	public static void improvementSimplex_SyntheticChip(int chipNumber, String design) throws JSONException, IOException {
		// create synthetic chip
		SyntheticChip syntheticChip = new SyntheticChip(chipNumber, design);

		// create parameters
		syntheticChip.channelNames = syntheticChip.getConsideredChannelNames();
		if (chipNumber == 3) syntheticChip.channelNames.remove("c18"); // TODO: c18 at chip 3, causes a heap error
		Parameters parameters = syntheticChip.getParametersForSimplex();

		// ### Sensitivity Analysis ###
		System.out.println("Sensitivity Analysis:");
		// compute minimal and maximal defects
		System.out.println("Compute minimal and maximal defects:");
		SingleParameterVariation single = new SingleParameterVariation(parameters, syntheticChip);
		if (!single.computeMinMaxDefects()) {
			System.out.println("Minimal and maximal defects could not be computed!");
			return;
		}

		// get only the most sensitive parameters
		int[] nSensitive = {16, 5, 3}; // number of sensitive parameters for each chip
		System.out.println("Find the " + nSensitive[chipNumber - 1] + " most sensitive parameters:");
		List<String> sensitiveParameterNames = SingleParameterVariation.findSensitiveParameters(parameters, nSensitive[chipNumber - 1]).stream().map(e -> e.name).collect(Collectors.toList());
		System.out.println(sensitiveParameterNames);

		// ### Simplex-Method ###
		double[] standardDeviations = {0.005, 0.002, 0.002}; // reasonable standard deviations for each chip number

		// get lower and upper bounds for all considered channel lengths
		int N = sensitiveParameterNames.size();
		System.out.println(N);
		double[] lowerBounds = new double[N];
		double[] upperBounds = new double[N];
		for (int i = 0; i < N; i++) {
			Parameter par = parameters.get(sensitiveParameterNames.get(i));
			System.out.println(par);
			lowerBounds[i] = par.lowerBound;
			upperBounds[i] = par.upperBound;
		}

		// cost function (bounded parameters)
		FunctionSyntheticChip costFunc = new FunctionSyntheticChip(chipNumber, design, standardDeviations[chipNumber - 1]);
		costFunc.channelNames = syntheticChip.channelNames;
		costFunc.parameterNames = sensitiveParameterNames;
		final MultivariateFunctionMappingAdapter wrapped = new MultivariateFunctionMappingAdapter(costFunc, lowerBounds, upperBounds);

		// simplex
		// the lower and upper steps array are needed to transform the steps correctly to the unbounded space
		double[] stepsLower = new double[N];
		double[] stepsUpper = new double[N];
		double[] initial = new double[N];
		for (int i = 0; i < N; i++) {
			Parameter par = parameters.get(sensitiveParameterNames.get(i));

			// get start point and steps for the simplex
			// both are still in bounded space and have to be transformed to unbounded values
			initial[i] = par.initialValue;
			stepsLower[i] = par.initialValue;
			stepsUpper[i] = par.maximalDefect;
		}
		// the steps have to be transformed correctly to the unbounded space
		double[] steps = new double[N];
		stepsLower = wrapped.boundedToUnbounded(stepsLower);
		stepsUpper = wrapped.boundedToUnbounded(stepsUpper);
		for (int i = 0; i < N; i++) {
			steps[i] = 0.9 * (stepsUpper[i] - stepsLower[i]);
		}

		SimplexOptimizer optimizer = new SimplexOptimizer(1e-2, 1e-3);
		NelderMeadSimplex simplex = new NelderMeadSimplex(steps);
		System.out.println(simplex);

		StopWatch watch = StopWatch.createStarted();
		System.out.println("Start Simplex improvement:");
		final PointValuePair optimum = optimizer.optimize(new MaxEval(1000), new ObjectiveFunction(wrapped), simplex, GoalType.MAXIMIZE, new InitialGuess(wrapped.boundedToUnbounded(initial)));
		watch.stop();
		System.out.println(String.format("Elapsed Time = %.3f s", watch.getTime() / 1000.0));

		final double[] bounded = wrapped.unboundedToBounded(optimum.getPoint());

		System.out.println(optimum.getValue());
		for (int i = 0; i < N; i++) {
			Parameter par = parameters.get(sensitiveParameterNames.get(i));
			par.initialValue = bounded[i];
			par.value = bounded[i];
			System.out.println(String.format("%s: %.3e", par.name, bounded[i]));
		}

		Path filePath = Paths.get("syntheticChip." + chipNumber + ".improved.simplex.json");
		Files.writeString(filePath, syntheticChip.createChip(parameters.getParameterMap()).toJSON().toString(4));
		System.out.println("File was saved to: " + filePath);
	}
}
