package at.jku.iic.microfluidics.droplet.improvedrobustness.metric;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import org.apache.commons.lang3.time.StopWatch;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.Parameter;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.ParameterMap;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.Parameters;

public class RobustnessMetric {
    Parameters parameters;
    SimulationFactory simulation;
    public int nBatchesMax = 50; // maximal number of batches (nSimulationsMax = nBatchesMax * nSimulationsPerBatch)
    public int nSimulationsPerBatch = 200; // number of simulations per batch
    public double errorTolerance = 0.02; // error tolerance to check the accuracy of the latest pSuccess value
    public int nConsiderForAccuracy = 500;
    int nSimulationsSuccess = 0; // number of successful simulations
    public List<Double> pSuccessValues = new ArrayList<Double>(nSimulationsPerBatch); // holds all pSuccess values

    public RobustnessMetric(Parameters parameters, SimulationFactory simulation) {
        this.parameters = parameters;
        this.simulation = simulation;
    }

    /**
     * Computes the value pSuccess for the robustness metric
     * 
     * @return the value of pSuccess
     */
    public void computeRobustnessMetric() {
        computeRobustnessMetric(new Random());
    }

    /**
     * Computes the value pSuccess for the robustness metric
     * 
     * @param randomGenerator random number generator
     * @return the value of pSuccess
     */
    public double computeRobustnessMetric(Random randomGenerator) {
        // initialize variables
        pSuccessValues.clear();

        // check if initial simulation is valid (otherwise the analysis does not work)
        ParameterMap initialParameters = parameters.getParameterMap();
        if (!simulation.createSimulation().simulateAndValidate(initialParameters)) {
            System.out.println("Initial simulation is not valid. Analysis not possible!");
            return 0.0;
        }

        // loop through the number of batches
        StopWatch watch = StopWatch.createStarted();
        Long timeIntermediate = 0L;
        for (int iBatch = 0; iBatch < nBatchesMax; iBatch++) {
            System.out.println(String.format("iBatch (nSimulationsPerBatch): %d (%d)", iBatch, nSimulationsPerBatch));

            // each entry in this list contains a parameterMap that are required for a simulation and the corresponding result
            List<SimulationResult> simulationResults = new ArrayList<>(nSimulationsPerBatch);

            // get parameters for each simulation (single threaded)
            // current batch consists of nBatch simulations
            int nValidParameters = 0;
            for (int iSim = 0; iSim < nSimulationsPerBatch; iSim++) {
                // prepare parameters
                boolean areParametersValid = true;
                for (Parameter parameter : parameters) {
                    // parameters which should not be included are set to their initial value (in case their value was different)
                    if (!parameter.includeInAnalysis) {
                        parameter.setValue(parameter.initialValue);
                        continue;
                    }

                    // set the value of the parameter accordingly to a gaussian distribution
                    parameter.setGaussianValue(randomGenerator);

                    // check if the parameter is valid
                    if (!parameter.isValueValid()) {
                        areParametersValid = false;
                        break;
                    }
                }

                // add parameters to list in case they are valid, otherwise add null (this will enforce a invalid simulation later)
                ParameterMap parameterMap = null;
                if (areParametersValid) {
                    parameterMap = parameters.getParameterMap();
                    nValidParameters++;
                }
                simulationResults.add(new SimulationResult(parameterMap));
            }

            System.out.println("Valid parameters added: " + nValidParameters);

            // simulate with the parameters obtained above (parallel)
            simulationResults.parallelStream().forEach(simulationResult -> {
                if (simulationResult.parameterMap == null) {
                    simulationResult.isSimulationValid = false;
                } else {
                    simulationResult.isSimulationValid = simulation.createSimulation().simulateAndValidate(simulationResult.parameterMap);
                }
            });

            // compute and add the pSuccess values (sequential)
            simulationResults.stream().forEach(simulationResult -> {
                if (simulationResult.isSimulationValid) {
                    nSimulationsSuccess++;
                }
                pSuccessValues.add(nSimulationsSuccess / (pSuccessValues.size() + 1.0));
            });

            // simulate with the parameters obtained above (sequential) and afterwards compute the pSuccess values (sequential)
            // for (SimulationResult simulationResult : simulationResults) {
            // simulationResult.isSimulationValid = simulation.simulateAndValidate(simulationResult.parameterMap);
            // if (simulationResult.isSimulationValid) {
            // nSimulationsSuccess++;
            // }
            // pSuccessValues.add(nSimulationsSuccess / (pSuccessValues.size() + 1.0));
            // }

            // print elapsed time
            // watch.split();
            long currentTime = watch.getTime();
            System.out.println(String.format("Time elapsed (delta) = %.3f (%.3f)", currentTime / 1000.0, (currentTime - timeIntermediate) / 1000.0));
            timeIntermediate = currentTime;

            // check if the accuracy of the latest pSuccess value is already sufficient
            // more precisely, check if the results from the latest nConsiderForAccuracy values are in a certain tolerance range
            if (isAccuracyReached(nConsiderForAccuracy)) {
                break;
            }

            System.out.println("latest pSuccessValue: " + pSuccessValues.get(pSuccessValues.size() - 1));
        }
        watch.stop();

        // return latest pSuccess value
        return pSuccessValues.get(pSuccessValues.size() - 1);
    }

    /**
     * Checks if the accuracy of last <code>nConsideredValues</code> values of pSuccess is sufficient
     * 
     * @return <code>true</code> if all goals are satisfied, <code>false</code> otherwise
     */
    private boolean isAccuracyReached(int nConsideredValues) {
        // checks if the number of pSuccess values is already high enough for the actual evaluation
        if (pSuccessValues.size() < nConsideredValues) {
            return false;
        }

        // check if the values are in the error tolerance
        double pSuccessReference = pSuccessValues.get(pSuccessValues.size() - 1);
        int start = pSuccessValues.size() - nConsideredValues;
        for (int i = start; i < pSuccessValues.size(); i++) {
            double delta = Math.abs(1.0 - pSuccessValues.get(i) / pSuccessReference);
            if (delta > errorTolerance) {
                return false;
            }
        }

        return true;
    }

    /**
     * Computes the value pSuccess for the robustness metric
     * 
     * @param randomGenerator random number generator
     * @return the value of pSuccess
     */
    public double computeRobustnessMetricAverage(Random randomGenerator, int nSimulations, int nAverageValues) {
        // initialize variables
        pSuccessValues.clear();

        // check if initial simulation is valid (otherwise the analysis does not work)
        ParameterMap initialParameters = parameters.getParameterMap();
        if (!simulation.createSimulation().simulateAndValidate(initialParameters)) {
            System.out.println("Initial simulation is not valid. Analysis not possible!");
            return 0.0;
        }

        System.out.println(String.format("nSimulations (nAverageValues): %d (%d)", nSimulations, nAverageValues));

        // each entry in this list contains a parameterMap that are required for a simulation and the corresponding result
        List<SimulationResult> simulationResults = new ArrayList<>(nSimulationsPerBatch);

        // get parameters for each simulation (single threaded)
        int nValidParameters = 0;
        for (int iSim = 0; iSim < nSimulations; iSim++) {
            // prepare parameters
            boolean areParametersValid = true;
            for (Parameter parameter : parameters) {
                // parameters which should not be included are set to their initial value (in case their value was different)
                if (!parameter.includeInAnalysis) {
                    parameter.setValue(parameter.initialValue);
                    continue;
                }

                // set the value of the parameter accordingly to a gaussian distribution
                parameter.setGaussianValue(randomGenerator);

                // check if the parameter is valid
                if (!parameter.isValueValid()) {
                    areParametersValid = false;
                    break;
                }
            }

            // add parameters to list in case they are valid, otherwise add null (this will enforce a invalid simulation later)
            ParameterMap parameterMap = null;
            if (areParametersValid) {
                parameterMap = parameters.getParameterMap();
                nValidParameters++;
            }
            simulationResults.add(new SimulationResult(parameterMap));
        }

        System.out.println("Valid parameters added: " + nValidParameters);

        // simulate with the parameters obtained above (parallel)
        simulationResults.parallelStream().forEach(simulationResult -> {
            if (simulationResult.parameterMap == null) {
                simulationResult.isSimulationValid = false;
            } else {
                simulationResult.isSimulationValid = simulation.createSimulation().simulateAndValidate(simulationResult.parameterMap);
            }
        });

        // compute and add the pSuccess values (sequential)
        simulationResults.stream().forEach(simulationResult -> {
            if (simulationResult.isSimulationValid) {
                nSimulationsSuccess++;
            }
            pSuccessValues.add(nSimulationsSuccess / (pSuccessValues.size() + 1.0));
        });

        //compute the average of the last nAverageValues values
        if(nAverageValues > pSuccessValues.size()){
            nAverageValues = pSuccessValues.size();
        }
        double pSuccessAverage = pSuccessValues.subList(pSuccessValues.size() - nAverageValues, pSuccessValues.size()).stream().mapToDouble(Double::doubleValue).sum() / (double) nAverageValues;

        System.out.println("pSuccessAverage: " + pSuccessAverage);

        // return pSuccessAverage
        return pSuccessAverage;
    }
}
