package at.jku.iic.microfluidics.droplet.improvedrobustness.parameter;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import org.json.JSONObject;

public class Parameter {
    public String name;

    // lower and upper bounds of the parameter (a robustness analysis will never search beyond these values for a solution)
    public double lowerBound;
    public double upperBound;

    // the minimal and maximal defect a parameter can have, while the simulation is still valid (the objectives are satisfied)
    public double minimalDefect;
    public double maximalDefect;

    // initial value and current value of the parameter
    public double initialValue;
    public double value;

    // parameters for the normal Gaussian distribution
    public double mean = 1.0;
    public double std = 0.1;

    // list of old initial values, in order to track the "improvements"
    public List<Double> initialValues = new ArrayList<>();

    // if true the parameter is included in robustness analysis
    public boolean includeInAnalysis;

    public Parameter(String name, double lowerBound, double upperBound, double initialValue, boolean includeInAnalysis) {
        this.name = name;
        this.lowerBound = lowerBound;
        this.upperBound = upperBound;
        this.initialValue = initialValue;
        this.value = initialValue;
        this.minimalDefect = initialValue;
        this.maximalDefect = initialValue;
        this.includeInAnalysis = includeInAnalysis;
    }

    public double getValue() {
        return value;
    }

    public String getName() {
        return name;
    }

    public void setValue(double value) {
        this.value = value;
    }

    public void setGaussianValue(Random random) {
        value = initialValue * (mean + std * random.nextGaussian());
    }

    public boolean isValueValid() {
        return lowerBound <= value && value <= upperBound;
    }

    @Override
    public String toString() {
        return String.format("name: %s, initial: %.3e, current: %.3e, bounds: [%.3e, %.3e], defectLimits: [%.3e, %.3e], mean: %.3e, std: %.3e", 
        name, initialValue, value, lowerBound, upperBound, minimalDefect, maximalDefect, mean, std);
    }

    public JSONObject toJSON() {
        JSONObject json = new JSONObject();
        json.put("name", name);
        json.put("lowerBound", lowerBound);
        json.put("upperBound", upperBound);
        json.put("minimalDefect", minimalDefect);
        json.put("maximalDefect", maximalDefect);
        json.put("initialValue", initialValue);
        json.put("value", value);
        json.put("mean", mean);
        json.put("std", std);
        json.put("includeInAnalysis", includeInAnalysis);

        return json;
    }

    public static Parameter fromJSON(JSONObject json) {
        String name = json.getString("name");
        double lowerBound = json.getDouble("lowerBound");
        double upperBound = json.getDouble("upperBound");
        double minimalDefect = json.getDouble("minimalDefect");
        double maximalDefect = json.getDouble("maximalDefect");
        double initialValue = json.getDouble("initialValue");
        double value = json.getDouble("value");
        double mean = json.getDouble("mean");
        double std = json.getDouble("std");
        boolean includeInAnalysis = json.getBoolean("includeInAnalysis");

        Parameter parameter = new Parameter(name, lowerBound, upperBound, initialValue, includeInAnalysis);
        parameter.value = value;
        parameter.minimalDefect = minimalDefect;
        parameter.maximalDefect = maximalDefect;
        parameter.mean = mean;
        parameter.std = std;

        return parameter;
    }
}
