package at.jku.iic.microfluidics.droplet.improvedrobustness.simplex;

import org.apache.commons.math3.analysis.MultivariateFunction;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import org.json.JSONException;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.improvedrobustness.metric.RobustnessMetric;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.Parameter;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.ParameterMap;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.Parameters;
import at.jku.iic.microfluidics.droplet.improvedrobustness.testarchitectures.syntheticchip.SyntheticChip;

public class FunctionSyntheticChip implements MultivariateFunction {
    int chipNumber;
    String design;
    double standardDeviation;
    private int iCostFuncEval = 0;
    public List<String> channelNames = new ArrayList<>();
    public List<String> parameterNames = new ArrayList<>();

    public FunctionSyntheticChip(int chipNumber, String design, double standardDeviation) {
        super();

        // set chip number
        this.chipNumber = chipNumber;

        // set the design (incorrect values lead to files that cannot be opened)
        this.design = design;

        this.standardDeviation = standardDeviation;
    }

    @Override
    public double value(double[] point) {
        return computeMetric_SyntheticChip(point);
    }


    public double computeMetric_SyntheticChip(double[] point) {
        iCostFuncEval++;
        System.out.println();
        System.out.println("CostFunctionEvaluation: " + iCostFuncEval);


        // ### Create initial chip from the point array ###
        // The order of the points must correspond to the list of the parameter names

        // create synthetic chip
        SyntheticChip syntheticChip = new SyntheticChip(chipNumber, design);

        // build parameter map (points change only the lengths)
        ParameterMap map = new ParameterMap();
        for (int i = 0; i < point.length; i++) {
            map.put(parameterNames.get(i), point[i]);
        }

        // get chip, with the updated parameters
        PhysicalBiochip chip;
        try {
            chip = syntheticChip.createChip(map);
        } catch (JSONException | IOException e) {
            e.printStackTrace();
            return 0.0;
        }

        // ### Compute robustness metric ###
        // Here the channels which are specified in the channel names are used
        // create parameters that should be included in the analysis (i.e., edges of the chip)
        // with the given standard deviation
        Parameters parameters = new Parameters();
        for (PhysicalEdge edge : chip.getEdges()) {
            // width
            Parameter parameterWidth = new Parameter(edge.name + ".w", 0.1 * edge.getWidth(), 2 * edge.getWidth(), edge.getWidth(), true);
            parameterWidth.std = standardDeviation;
            parameters.add(parameterWidth);

            // height
            Parameter parameterHeight = new Parameter(edge.name + ".h", 0.1 * edge.getHeight(), 2 * edge.getHeight(), edge.getHeight(), true);
            parameterHeight.std = standardDeviation;
            parameters.add(parameterHeight);

            // length
            Parameter parameterLength = new Parameter(edge.name + ".l", 0.1 * edge.getLength(), 2 * edge.getLength(), edge.getLength(), true);
            parameterLength.std = standardDeviation;
            parameters.add(parameterLength);
        }

        // define robustness metric
        RobustnessMetric metric = new RobustnessMetric(parameters, syntheticChip);
        metric.errorTolerance = 0.02;
        metric.nConsiderForAccuracy = 500;

        // metric.nBatchesMax = 1;
        // metric.nSimulationsPerBatch = 200;
        Random randomGenerator = new Random(1000);

        // compute robustness metric
        double metricValue = metric.computeRobustnessMetric(randomGenerator);
        // double metricValue = metric.computeRobustnessMetricAverage(randomGenerator, 400, 10);
        // metric.pSuccessValues.forEach(e -> System.out.println(e));

        return metricValue;
    }

}
