package at.jku.iic.microfluidics.droplet.improvedrobustness.singleparametervariation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.IntStream;
import org.apache.commons.lang3.time.StopWatch;
import at.jku.iic.microfluidics.droplet.improvedrobustness.metric.SimulationFactory;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.Parameter;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.ParameterMap;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.Parameters;

public class SingleParameterVariation {
	Parameters parameters;
	SimulationFactory simulation;
	public double deltaDefect = 1e-6;
	public double relativeDeltaTolerance = 0.01;
	public int nIterationsMax = 10;
	public boolean changeInsensitiveParameters = false;
	public int nConsiderSensitiveParametersMax = 8;
	public int nSimulations = 0;
	public int nIterations = 0;

	public SingleParameterVariation(Parameters parameters, SimulationFactory simulation) {
		this.parameters = parameters;
		this.simulation = simulation;
	}

	public boolean computeMinMaxDefects() {
		nSimulations = 0;

		// check if initial simulation is valid (otherwise the analysis does not work)
		if (!simulation.createSimulation().simulateAndValidate(parameters.getParameterMap())) {
			System.out.println("Initial simulation is not valid. Analysis not possible!");
			nSimulations++;
			return false;
		}

		// store initial parameter values (starting point for all parallel simulations)
		ParameterMap initialParameters = parameters.getParameterMap();

		// find limits of parameters (parallel) and sum up the number of simulations
		nSimulations += parameters.getParameters()
				.parallelStream()
				.mapToInt(parameter -> findLimits(parameter, initialParameters))
				.sum();

		// find limits of parameters (sequential) and sum up the number of simulations
		// for (Parameter parameter : parameters.getParameters()) {
		// System.out.println("Find defect limits of: " + parameter.name);
		// nSimulations += findLimits(parameter, initialParameters);
		// }

		return true;
	}

	public static List<Parameter> findSensitiveParameters(Parameters parameters, int nMostSensitiveParameters) {
		List<Parameter> sensitiveParameters = new ArrayList<>();

		Map<String, Double> parameterSensitivity = new HashMap<>();
		for (Parameter parameter : parameters) {
			// compute sensitivity value for the parameter
			// the closer the defect limits are to the initial value, the higher the sensitivity
			double initial = parameter.initialValue;
			double sensitivity = initial / (initial - parameter.minimalDefect) + initial / (parameter.maximalDefect - initial);

			parameterSensitivity.put(parameter.name, sensitivity);
		}

		// sort list (descending)
		List<Map.Entry<String, Double>> list = new ArrayList<>(parameterSensitivity.entrySet());
		list.sort((a, b) -> Double.compare(b.getValue(), a.getValue()));
		System.out.println(list);

		int end = (0 < nMostSensitiveParameters && nMostSensitiveParameters < list.size()) ? nMostSensitiveParameters : list.size();
		for (int i = 0; i < end; i++) {
			sensitiveParameters.add(parameters.get(list.get(i).getKey()));
		}

		return sensitiveParameters;
	}

	public List<Parameter> findSensitiveParameters(int nMostSensitiveParameters) {
		List<Parameter> sensitiveParameters = new ArrayList<>();
		// Parameters sensitiveParameters = new Parameters();
		nSimulations = 0;

		// check if initial simulation is valid (otherwise the analysis does not work)
		if (!simulation.createSimulation().simulateAndValidate(parameters.getParameterMap())) {
			System.out.println("Initial simulation is not valid. Analysis not possible!");
			nSimulations++;
			return sensitiveParameters;
		}

		// store initial parameter values (starting point for all parallel simulations)
		ParameterMap initialParameters = parameters.getParameterMap();

		// find limits of parameters (parallel) and sum up the number of simulations
		nSimulations += parameters.getParameters()
				.parallelStream()
				.mapToInt(parameter -> findLimits(parameter, initialParameters))
				.sum();

		Map<String, Double> parameterSensitivity = new HashMap<>();
		for (Parameter parameter : parameters) {
			// compute sensitivity value for the parameter
			// the closer the defect limits are to the initial value, the higher the sensitivity
			double initial = parameter.initialValue;
			double sensitivity = initial / (initial - parameter.minimalDefect) + initial / (parameter.maximalDefect - initial);

			parameterSensitivity.put(parameter.name, sensitivity);
		}

		// sort list (descending)
		List<Map.Entry<String, Double>> list = new ArrayList<>(parameterSensitivity.entrySet());
		list.sort((a, b) -> Double.compare(b.getValue(), a.getValue()));
		System.out.println(list);

		int end = (0 < nMostSensitiveParameters && nMostSensitiveParameters < list.size()) ? nMostSensitiveParameters : list.size();
		for (int i = 0; i < end; i++) {
			sensitiveParameters.add(parameters.get(list.get(i).getKey()));
		}

		return sensitiveParameters;
	}

	public Parameters startImprovement() {
		// initialize variables
		nSimulations = 0;
		nIterations = 0;

		// check if initial simulation is valid (otherwise the analysis does not work)
		if (!simulation.createSimulation().simulateAndValidate(parameters.getParameterMap())) {
			System.out.println("Initial simulation is not valid. Analysis not possible!");
			nSimulations++;
			return parameters;
		}

		StopWatch watch = StopWatch.createStarted();
		Long timeIntermediate = 0L;
		for (int iIteration = 0; iIteration < nIterationsMax; iIteration++) {
			System.out.println();
			System.out.println(String.format("iIteration: %d", iIteration));

			nIterations++; // used to obtain information about the computation, but basically the same as iIteration

			// get initial parameter values of this iteration (starting point for all parallel simulations)
			ParameterMap initialParameters = parameters.getParameterMap();

			// find limits of parameters (parallel) and sum up the number of simulations
			System.out.println("Find defect limits:");
			nSimulations += parameters.getParameters().parallelStream().mapToInt(parameter -> findLimits(parameter, initialParameters)).sum();

			// find limits of parameters (sequential) and sum up the number of simulations
			// for(Parameter parameter : parameters.getParameters()){
			// nSimulations += findLimits(parameter, initialParameters);
			// }

			// print elapsed time
			// watch.split();
			long currentTime = watch.getTime();
			System.out.println();
			System.out.println(String.format("Time elapsed (delta) = %.3f (%.3f)", currentTime / 1000.0, (currentTime - timeIntermediate) / 1000.0));
			timeIntermediate = currentTime;

			// find parameter which improves the robustness the most and modify its value
			System.out.println();
			System.out.println("Find sensitive parameters:");
			List<Parameter> mostSensitiveParameters = findSensitiveParameters(parameters, 0);
			int end = nConsiderSensitiveParametersMax > mostSensitiveParameters.size() ? mostSensitiveParameters.size() : nConsiderSensitiveParametersMax;
			boolean newImprovement = false;
			for (int i = 0; i < end; i++) {
				Parameter par = mostSensitiveParameters.get(i);

				// skip if parameter is insensitive (defect limits are the same as the boundary limits)
				if (par.maximalDefect == par.upperBound && par.minimalDefect == par.lowerBound) {
					System.out.println("Skip parameter: " + par.name);
					continue;
				}

				// if the two values are within a tolerance consider the next parameter, otherwise improve this parameter and start next iteration
				double actualValue = par.initialValue;
				double expectedValue = (par.maximalDefect + par.minimalDefect) / 2;
				double errorTolerance = 0.05;
				if (Math.abs((actualValue - par.minimalDefect) / (par.maximalDefect - par.minimalDefect) - 0.5) > errorTolerance) {
					System.out.println("Improve parameter: " + par.name);
					par.initialValue = expectedValue;
					par.value = par.initialValue;
					newImprovement = true;
					break;
				} else {
					System.out.println("Skip parameter: " + par.name);
				}
			}

			// end method if no new improvement can be made
			if (!newImprovement) {
				break;
			}
		}
		watch.stop();

		return parameters;
	}

	private int findLimits(Parameter parameter, ParameterMap initialParameters) {
		System.out.println("Start: " + parameter.name);
		// check if the parameter should be considered in the analysis
		if (!parameter.includeInAnalysis) {
			parameter.value = parameter.initialValue;
			return 0;
		}

		// make two copies of initialParameters map for the minimal and maximal value (due to parallel computing)
		ParameterMap initialParametersMinimalLimit = ParameterMap.makeCopy(initialParameters);
		ParameterMap initialParametersMaximalLimit = ParameterMap.makeCopy(initialParameters);

		// find minimal and maximal limit (parallel), and return the required number of simulations
		int sum = IntStream.of(0, 1).parallel().map(i -> {
			if (i == 0) {
				return findMinimalLimit(parameter, initialParametersMinimalLimit);
			} else {
				return findMaximalLimit(parameter, initialParametersMaximalLimit);
			}
		}).sum();

		System.out.println("End: " + parameter.name);

		return sum;
	}

	private int findMinimalLimit(Parameter parameter, ParameterMap initialParameters) {
		// initialize variables
		int simulationCount = 0;

		// first check the lower boundary (if the simulation is valid, then return immediately)
		// also eliminates the problem, that the binary search does not return the lower bound value
		initialParameters.put(parameter.name, parameter.lowerBound);
		simulationCount++;
		if (simulation.createSimulation().simulateAndValidate(initialParameters)) {
			// set minimalDefect and reset the parameter value to the initialValue
			parameter.minimalDefect = parameter.lowerBound;
			parameter.value = parameter.initialValue;
			return simulationCount;
		}

		// binary search
		double left = parameter.lowerBound;
		double right = parameter.initialValue;
		double result = right;

		while (left <= right) {
			// set the value of the parameter in the middle of the interval [left, right]
			double middle = left + (right - left) / 2;
			initialParameters.put(parameter.name, middle);
			// System.out.println(String.format("Compute Minimal Limit: [%.3e, %.3e, %.3e]", left, middle, right));

			// if simulation is valid continue search in the interval [left, middle], otherwise in [middle, right]
			// the addition and subtraction of deltaDefect ensures, that the search terminates
			simulationCount++;
			if (simulation.createSimulation().simulateAndValidate(initialParameters)) {
				result = middle;
				right = middle - deltaDefect;
			} else {
				left = middle + deltaDefect;
			}
		}

		// set minimalDefect and reset the parameter value to the initialValue
		parameter.minimalDefect = result;
		parameter.value = parameter.initialValue;

		return simulationCount;
	}

	private int findMaximalLimit(Parameter parameter, ParameterMap initialParameters) {
		// initialize variables
		int simulationCount = 0;

		// first check the upper boundary (if the simulation is valid, then return immediately)
		// also eliminates the problem, that the binary search does not return the upper bound value
		initialParameters.put(parameter.name, parameter.upperBound);
		simulationCount++;
		if (simulation.createSimulation().simulateAndValidate(initialParameters)) {
			// set minimalDefect and reset the parameter value to the initialValue
			parameter.maximalDefect = parameter.upperBound;
			parameter.value = parameter.initialValue;
			return simulationCount;
		}

		// binary search
		double left = parameter.initialValue;
		double right = parameter.upperBound;
		double result = left;

		while (left <= right) {
			// set the value of the parameter in the middle of the interval [left, right]
			double middle = left + (right - left) / 2;
			initialParameters.put(parameter.name, middle);
			// System.out.println(String.format("Compute Maximal Limit: [%.3e, %.3e, %.3e]", left, middle, right));

			// if simulation is valid continue search in the interval [middle, right], otherwise in [left, middle]
			// the addition and subtraction of deltaDefect (1 micrometer) ensures, that the search terminates
			simulationCount++;
			if (simulation.createSimulation().simulateAndValidate(initialParameters)) {
				result = middle;
				left = middle + deltaDefect;
			} else {
				right = middle - deltaDefect;
			}
		}

		// set maximalDefect and reset the parameter value to the initialValue
		parameter.maximalDefect = result;
		parameter.value = parameter.initialValue;

		return simulationCount;
	}
}
