package at.jku.iic.microfluidics.droplet.improvedrobustness.testarchitectures.syntheticchip;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import at.jku.iic.microfluidics.droplet.basic.architecture.EndPoint;
import at.jku.iic.microfluidics.droplet.basic.architecture.Path;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalBiochip;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalEdge;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.PhysicalPump;
import at.jku.iic.microfluidics.droplet.basic.architecture.physical.VolumetricFlowRatePump;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalDropletInjectionSequence;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalDropletInjectionTime;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalHeaderInjectionTime;
import at.jku.iic.microfluidics.droplet.basic.injection.physical.PhysicalPayloadInjectionTime;
import at.jku.iic.microfluidics.droplet.electric.simulator.PhysicalSimulator;
import at.jku.iic.microfluidics.droplet.electric.simulator.PhysicalSystemStatePostProcessor;
import at.jku.iic.microfluidics.droplet.electric.simulator.state.PhysicalSystemState;
import at.jku.iic.microfluidics.droplet.improvedrobustness.metric.Simulation;
import at.jku.iic.microfluidics.droplet.improvedrobustness.metric.SimulationFactory;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.ParameterMap;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.Parameter;
import at.jku.iic.microfluidics.droplet.improvedrobustness.parameter.Parameters;

public class SyntheticChip implements Simulation, SimulationFactory {
	int chipNumber;
	String design;
	PhysicalBiochip chip;
	List<PhysicalSystemState> states;
	Experiments experiments;
	public List<String> channelNames;

	/**
	 * 
	 * @param chipNumber number of chip
	 * @param design name of the design of the chip (possible values are for example "initial", "improved.single", "improved.multi", etc.). Basically a part of the json file name.
	 */
	public SyntheticChip(int chipNumber, String design) {
		// set chip number
		this.chipNumber = chipNumber;

		// set the design (incorrect values lead to files that cannot be opened)
		this.design = design;
	}

	public List<String> getConsideredChannelNames() throws JSONException, IOException {
		List<String> channelNames = new ArrayList<>();

		// create chip
		PhysicalBiochip chip = PhysicalBiochip.fromJSON(getJsonChip());

		// get all channels which are no modules or bypass channels (i.e, name starts with c)
		for (PhysicalEdge edge : chip) {
			if (edge.name.startsWith("c")) {
				channelNames.add(edge.name);
			}
		}

		// remove channels which are part of a bifurcation with a bypass channel
		for (PhysicalEdge bypassChannel : chip) {
			// check each bypass channel
			if (bypassChannel.name.startsWith("bp")) {
				// loop over edges connected to the src point of the bypass channel
				List<PhysicalEdge> srcEdges = new ArrayList<>();
				srcEdges.addAll(bypassChannel.src.getPredecessors());
				srcEdges.addAll(bypassChannel.src.getSuccessors());
				for (PhysicalEdge srcEdge : srcEdges) {
					// loop over edges connected to the dst point of the bypass channel
					List<PhysicalEdge> dstEdges = new ArrayList<>();
					dstEdges.addAll(bypassChannel.dst.getPredecessors());
					dstEdges.addAll(bypassChannel.dst.getSuccessors());
					for (PhysicalEdge dstEdge : dstEdges) {
						// do not consider the bypass channel
						if (srcEdge == bypassChannel || dstEdge == bypassChannel) {
							continue;
						}

						// get endpoint for src and dst edge (i.e., endpoint which is not connected to the bypass)
						EndPoint<PhysicalEdge> srcEdgeEndPoint = srcEdge.src == bypassChannel.src ? srcEdge.dst : srcEdge.src;
						EndPoint<PhysicalEdge> dstEdgeEndPoint = dstEdge.src == bypassChannel.dst ? dstEdge.dst : dstEdge.src;

						// remove channels from list if they share the same endpoint (does nothing if channels were already removed in previous steps)
						if (srcEdgeEndPoint == dstEdgeEndPoint) {
							channelNames.remove(srcEdge.name);
							channelNames.remove(dstEdge.name);
						}
					}
				}
			}
		}

		// remove channels which are connected to the pump
		// only remove when a single channel is connected to the src or dst point of the pump
		// this may happens when multiple channels are connected to "ground" (the input channel of the network should always be a single channel)
		PhysicalPump pump = chip.getPumps().get(0);
		List<PhysicalEdge> srcEdges = new ArrayList<>();
		srcEdges.addAll(pump.src.getPredecessors());
		srcEdges.addAll(pump.src.getSuccessors());
		if (srcEdges.size() == 1) {
			channelNames.remove(srcEdges.get(0).name);
		}
		List<PhysicalEdge> dstEdges = new ArrayList<>();
		dstEdges.addAll(pump.sink.getPredecessors());
		dstEdges.addAll(pump.sink.getSuccessors());
		if (dstEdges.size() == 1) {
			channelNames.remove(dstEdges.get(0).name);
		}

		return channelNames;
	}

	@Override
	public Simulation createSimulation() {
		return new SyntheticChip(chipNumber, design);
	}

	public JSONObject getJsonChip() throws JSONException, IOException {
		return new JSONObject(new String(SyntheticChip.class.getResourceAsStream("syntheticChip." + chipNumber + "." + design + ".json").readAllBytes(), StandardCharsets.UTF_8));
	}

	public JSONObject getJsonExperiments() throws JSONException, IOException {
		return new JSONObject(new String(SyntheticChip.class.getResourceAsStream("syntheticChip." + chipNumber + ".experiments.json").readAllBytes(), StandardCharsets.UTF_8));
	}

	public JSONObject getJsonParameters() throws JSONException, IOException {
		return new JSONObject(new String(SyntheticChip.class.getResourceAsStream("syntheticChip." + chipNumber + "." + design + ".parameters.json").readAllBytes(), StandardCharsets.UTF_8));
	}



	public PhysicalBiochip createChip(ParameterMap parameters) throws JSONException, IOException {
		// read initial design from json
		JSONObject jsonChip = getJsonChip();

		// set the parameters of the json file accordingly
		JSONArray edgesJson = jsonChip.getJSONArray("edges");
		for (int i = 0; i < edgesJson.length(); i++) {
			// get edge
			JSONObject edge = edgesJson.getJSONObject(i);
			String name = edge.getString("name");

			// change width (parameter has suffix .w)
			Double width = parameters.get(name + ".w");
			if (width != null) {
				edge.put("width", width);
			}

			// change height (parameter has suffix .h)
			Double height = parameters.get(name + ".h");
			if (height != null) {
				edge.put("height", height);
			}

			// change length (parameter has suffix .l)
			Double length = parameters.get(name + ".l");
			if (length != null) {
				edge.put("length", length);
			}
		}

		// create chip
		return PhysicalBiochip.fromJSON(jsonChip);
	}

	@Override
	public boolean simulateAndValidate(ParameterMap parameters) {
		// read design from json and corresponding experiments
		try {
			chip = createChip(parameters);
			experiments = Experiments.fromJSON(getJsonExperiments());
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}

		VolumetricFlowRatePump pump = (VolumetricFlowRatePump) chip.getPumps().get(0);

		// test the chip with the following experiments
		for (Experiment experiment : experiments.getExperiments()) {
			// prepare simulation
			List<PhysicalDropletInjectionTime> dropletInjections = new LinkedList<>();
			for (int i = 0; i < experiment.injectionOrder.size(); i++) {
				// get droplet name
				String dropletName = experiment.injectionOrder.get(i);

				// get injection time
				double injectionTime = experiment.dropletDistances.get(i);

				// inject droplet and check if it is a payload or header droplet
				if (dropletName.startsWith("p")) {
					dropletInjections.add(new PhysicalPayloadInjectionTime(pump, injectionTime, chip.payloadDropletVolume));
				} else {
					dropletInjections.add(new PhysicalHeaderInjectionTime(pump, injectionTime, chip.headerDropletVolume));
				}
			}

			// simulate
			states = null;
			try {
				PhysicalSimulator simulator = new PhysicalSimulator(chip, new PhysicalDropletInjectionSequence(dropletInjections));
				states = simulator.simulate(true);
			} catch (Exception e) {
				// e.printStackTrace();
				states = null;
			}

			// check if simulation was valid
			if (states == null) {
				return false;
			}

			// check the paths of the droplets
			PhysicalSystemStatePostProcessor post = new PhysicalSystemStatePostProcessor(states);
			for (DropletPath desiredPath : experiment.getDropletPaths()) {
				Path<PhysicalEdge> actualDropletPath = post.getPathOfDroplet(desiredPath.dropletName);
				// System.out.println("\tDesiredDropletPath (" + desiredPath.dropletName + "):" + desiredPath.path);
				// System.out.println("\tActualDropletPath (" + desiredPath.dropletName + "): " + actualDropletPath.getEdges().stream().map(e -> e.name).collect(Collectors.toList()));
				if (!desiredPath.isSameAs(actualDropletPath)) {
					return false;
				}
			}
		}

		return true;
	}

	public Parameters getParametersForRobustnessAnalysis() throws JSONException, IOException {
		// get chip
		PhysicalBiochip chip = PhysicalBiochip.fromJSON(getJsonChip());

		// create parameters that should be included in the analysis (i.e., edges of the chip)
		Parameters parameters = new Parameters();
		for (PhysicalEdge edge : chip.getEdges()) {
			// width
			Parameter parameterWidth = new Parameter(edge.name + ".w", 0.1 * edge.getWidth(), 2 * edge.getWidth(), edge.getWidth(), true);

			// height
			Parameter parameterHeight = new Parameter(edge.name + ".h", 0.1 * edge.getHeight(), 2 * edge.getHeight(), edge.getHeight(), true);

			// length
			Parameter parameterLength = new Parameter(edge.name + ".l", 0.1 * edge.getLength(), 2 * edge.getLength(), edge.getLength(), true);

			// add to parameters
			parameters.add(parameterWidth);
			parameters.add(parameterHeight);
			parameters.add(parameterLength);
		}

		return parameters;
	}

	public Parameters getParametersForParameterVariations() throws JSONException, IOException {
		// get chip
		PhysicalBiochip chip = PhysicalBiochip.fromJSON(getJsonChip());

		Parameters parameters = new Parameters();
		for (PhysicalEdge edge : chip.getEdges()) {
			// only consider listed edges
			if (!channelNames.contains(edge.name)) {
				continue;
			}

			// length
			Parameter parameterLength = new Parameter(edge.name + ".l", 0.5 * edge.getLength(), 2.0 * edge.getLength(), edge.getLength(), true);

			// add to parameters
			parameters.add(parameterLength);
		}

		return parameters;
	}

	public Parameters getParametersForSimplex() throws JSONException, IOException {
		// get chip
		PhysicalBiochip chip = PhysicalBiochip.fromJSON(getJsonChip());

		Parameters parameters = new Parameters();
		for (PhysicalEdge edge : chip.getEdges()) {
			// only consider listed edges
			if (!channelNames.contains(edge.name)) {
				continue;
			}

			// length
			Parameter parameterLength = new Parameter(edge.name + ".l", 0.5 * edge.getLength(), 2.0 * edge.getLength(), edge.getLength(), true);

			// add to parameters
			parameters.add(parameterLength);
		}

		return parameters;
	}
}
