/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.utils.dumper;

import java.awt.Color;
import java.io.PrintWriter;

/**
 * Dumper implementation using the graph modeling language (GML).<br>
 * <p>
 * For a detailed specification of the GML visit <a href=
 * "http://www.fim.uni-passau.de/fileadmin/files/lehrstuhl/brandenburg/projekte/gml/gml-technical-report.pdf"
 * >GML: A portable Graph File Format</a> (July 2012).
 * </p>
 * <p>
 * This implementation dumps a call graph to the provided writer using the graph modeling language.<br>
 * GML looks like following:<br>
 * <code>
 * graph [<br>
 * &nbsp;&nbsp;comment "this is a graph"<br>
 * &nbsp;&nbsp;directed 1<br>
 * &nbsp;&nbsp;node [<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;id 0<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;label "Node 0"<br>
 * &nbsp;&nbsp;]<br>
 * &nbsp;&nbsp;node [<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;id 1<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;label "Node 1"<br>
 * &nbsp;&nbsp;]<br>
 * &nbsp;&nbsp;edge [<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;source 0<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;target 1<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;label "Edge from Node 0 to Node 1"<br>
 * &nbsp;&nbsp;]<br>
 * ]<br>
 * </code>
 * </p>
 * 
 * @author Florian Angerer, November 2011
 * @version 1.3
 * @since 1.0
 * 
 */
public class GMLDumper implements Dumper {
    
    private final String creator;

    public GMLDumper() {

        this.creator = "dumper";
    }

    public GMLDumper(String creator) {

        this.creator = creator;
    }
    
    protected String ln() {
        return System.lineSeparator();
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.jku.ssw.kemroiec.callGraph.dumper.CallGraphDumper_v2#generateHeader()
     */
    @Override
    public void generateHeader(PrintWriter w) {
        w.append("Creator \"").append(creator).append('"').append(ln()).append("Version \"1.0\"").append(ln())
                .append("graph [").append(ln()).append("directed 1").append(ln()).append("label \"\"").append(ln());
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.jku.ssw.kemroiec.callGraph.dumper.CallGraphDumper_v2#generateTrailer()
     */
    @Override
    public void generateTrailer(PrintWriter w) {
        w.append(ln()).append(']');
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.jku.ssw.kemroiec.callGraph.dumper.CallGraphDumper_v2#generateNode(at.jku.ssw.kemroiec.
     * callGraph.CGNode2, java.lang.String)
     */
    @Override
    public void generateNode(PrintWriter w, String n, String nodeID) {
    	n = n.replaceAll("\"", "&quot;");
    	n = n.replaceAll("&", "&amp;");
    	
        w.append("node [").append(ln()).append("id ").append(nodeID).append("").append(ln()).append("label \"").append(n)
                .append("\"").append(ln()).append("]").append(ln());
    }
    
    
	@Override
    public void generateNode(PrintWriter w, String n, String nodeID,
			Color nodeColor) {
		n = n.replaceAll("\"", "&quot;");
    	n = n.replaceAll("&", "&amp;");
		
		w.append("node [").append(ln()).append("id ").append(nodeID).append("").append(ln()).append("label \"").append(n)
		.append("\"").append(ln()).append("graphics").append(ln()).append("[").append(ln()).append(" x 0.0 ").append(ln()).append(" y 0.0 ").append(ln()).append(" w 30.0 ").append(ln()).append(" h 30.0 ").append(ln()).append(" type \"rectangle\" ").append(ln()).append(" fill \"#" + 
				Integer.toHexString(nodeColor.getRGB() & 0x00FFFFFF) + "\"").append(ln()).append(" outline \"#000000\"").append(ln()).append("]")
		.append("").append(ln()).append("]").append(ln());
	}

	@Override
	public void generateEdge(PrintWriter w, String src, String description,
			String dest) {
		w.append("edge [").append(ln()).append("source ").append(src).append("").append(ln()).append("target ").append(dest)
        .append("").append(ln()).append("label \"").append(description).append("\"]").append(ln());
		
	}

	@Override
	public void generateEdge(PrintWriter w, String src, String description,
			String dest, Color color, boolean dashed) {
		w.append("edge [").append(ln()).append("source ").append(src).append("").append(ln()).append("target ").append(dest)
		.append("").append(ln()).append("graphics").append(ln()).append("[").append(ln()).append(" fill \"#" + 
				Integer.toHexString(color.getRGB() & 0x00FFFFFF) + "\"").append(ln()).append("");
		if(dashed) w.append("style \"dashed\"");
		w.append("targetArrow \"standard\"");
		w.append("").append(ln()).append("]");
        w.append("").append(ln()).append("label \"").append(description).append("\"]").append(ln());
	}
}
