/*******************************************************************************
 * Copyright (C) 2018 Andreas Grimmer
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package at.jku.iic.microfluidics.utils.dumper;

import java.io.PrintWriter;

/**
 * Abstract class implementing common features to dump a graph to a {@link PrintWriter} in a
 * certain format {@link Dumper}.<br>
 * <p>
 * Dumping a graph to a stream is always done in a similar way. This dumper assumes following
 * structure of a textual description of a graph:<br>
 * <ol>
 * <li>A header is used defining some general properties of the graph.</li>
 * <li>Then, a set of nodes and edges is specified.</li>
 * <li>At the end, some trailer must be written (e.g. just a closing brace or something similar).</li>
 * </ol>
 * </p>
 * <p>
 * {@link Dumper} implements methods which must exactly generate these parts. 
 * This abstract class implements basic features like generating unique IDs for nodes and
 * eliminating redundant edges.
 * </p>
 * <p>
 * <b>Note:</b><br>
 * Don't forget to call method {@link #dump()} to actually dump a graph.<br>
 * You can use one instance of this class several times but it will always dump the same (perhaps
 * changed) graph to the same stream.
 * </p>
 * 
 * 
 * @author Florian Angerer, July 2012
 * @version 1.3
 * @since 1.3
 * 
 */
public abstract class GraphDumper {
	
	/** switch to disable outputting of calls to unknown/predefined POUs */
    protected boolean includeUnknown = true;

    /**
     * switch to enable elimination of redundant edges (e.g. calls to the same POU several times)
     */
    protected boolean eliminateRedundantEdges = true;

    /** the print writer to write the generated code to */
    protected final PrintWriter writer;

    /** just a counter used to generate unique IDs */
    private int nodeId = 0;
    
    protected final Dumper dumper;

    /**
     * Constructor initializing the writer and the dumper.<br>
     * To dump the graph, use abstract method {@link #dump()};
     * 
     * @param writer
     *            the writer where to dump the graph in textual format to
     * @param dumper
     *            the dumper specifying the format 
     */
    protected GraphDumper(PrintWriter writer, Dumper dumper) {

        if (writer == null) {
            throw new IllegalArgumentException("writer must not be null");
        }
        if(dumper == null) {
        	throw new IllegalArgumentException("dumer must not be null");
        }

        this.writer = writer;
        this.dumper = dumper;
    }

    /**
     * Dumps a graph to {@link #writer}.
     */
    public abstract void dump();
    
    /**
     * Generates a new unique node ID.
     * 
     * @return A fresh and unique node ID.
     */
    protected String genNodeID() {

        return Integer.toString(nodeId++);
    }
   

    /**
     * @return the includeUnknown
     */
    public boolean isIncludeUnknown() {

        return includeUnknown;
    }

    /**
     * @param includeUnknown
     *            the includeUnknown to set
     */
    public void setIncludeUnknown(boolean includeUnknown) {

        this.includeUnknown = includeUnknown;
    }

    /**
     * Returns flag {@link #eliminateRedundantEdges}.
     * 
     * @see #eliminateRedundantEdges
     * @return value of flag {@link #eliminateRedundantEdges}
     */
    public boolean isEliminateRedundantEdges() {

        return eliminateRedundantEdges;
    }

    /**
     * Sets flag {@link #eliminateRedundantEdges}.
     * 
     * @see #eliminateRedundantEdges
     * @param eliminateRedundantEdges
     *            the Boolean value to set
     */
    public void setEliminateRedundantEdges(boolean eliminateRedundantEdges) {

        this.eliminateRedundantEdges = eliminateRedundantEdges;
    }

}
